/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import javax.net.ssl.*;
import com.sun.net.ssl.*;

import j4d.open.*;

class ConnectionImpl implements Connection
{
	// IP or DNS name of the host this instance is connected to
	private String					m_strHost = null;
	
	private Properties				m_propsInfo = null;
	
	// A connection to 4D DB via 4D Open For Java
	private opConnection			m_opConnection = null;
	
	// An object that will handle all communication with the DB.
	private opProcess				m_opProcess = null;
	
	private static final int			DEFAULT_PORT = 19813;
	
	private DatabaseMetaDataImpl	m_dbMD = null;
	
	/**
	*	Here is the thing - 4D allows only one selection per table and obviously I need to be able
	*	to use a particular table in as many ResultSets as I want. For that purpose, each ResultSet should register
	*	all tables that it is going to use. Before actually using a table, ResultSet (or any other table user) should
	*	register it. After all tables are registered, ResultSet should restore selection for those tables
	*	that were used by somebody else and lock on its opProcess
	*	and do its work. ResultSet must recreate all its selections because they may have been changed
	*	by other ResultSets. After ResultSet is done using tables it should unregister them.
	*
	*	Find out on what level "One selection per table" rule works - on opConnection or opProcess.
	*	I think it is on opConnection level, but if it is on opProcess level then table registration should be done 
	*	on opProcess level too.
	*
	*
	*	m_htTablesRegister = { ( "TableName", TableRegisterer ), ... }
	*		where "TableName" is a key and TableRegisterer is a value.
	*/
	private Hashtable				m_htTablesRegister = new Hashtable ( );
	
	// Is used to synchronize table registration
	private Object					m_objTableRegisterSync = new Object ( );
	
	/**
	*	m_htTableUsers = { ( "TableName", TableRegistererThatWasLastToUseTheTable ), ... }
	*/
	private Hashtable				m_htTableUsers = new Hashtable ( );
	
	/**
		Synchronization lock for safe access to opProcess.
	*/
	private Object						m_objDataAccessLock = new Object ( );
	
	/**
		"Connection Is Closed" flag. Once m_bClosed is set to true,
		all methods will throw "Connection is closed" exception.
	*/
	private boolean						m_bClosed = false;

	private ConnectionKeeper			m_connectionKeeper = null;
	
	/**
	*	Contains references to all statements created by this connection.
	*/
	private Vector						m_vctrStatements = new Vector ( );
	
	/**
	*	A unique key to generate unique selection names.
	*/
	private int							m_nSelectionID = 0;
	/**
	*	A unique key to generate unique set names.
	*/
	private int							m_nSetID = 0;
	
	private boolean						m_bAutoCommit = true;
	private boolean						m_bTransactionInProgress = false;
	private int							m_nTransactionIsolation = TRANSACTION_READ_UNCOMMITTED;
	
	/* A flag that indicates usage of optimazed methods in versions >= 2004 */
	private boolean						m_bUseOptimization = false;
	/**
	*	Constructor.
	*
	*	strURL should contain either ip address or dns name of the 4D server
	*	with optional port number. Port in the URL has higher priority then 
	*	"port" property.
	*/
	ConnectionImpl ( String strURL, Properties propsInfo ) throws SQLException
	{
		if ( strURL. indexOf ( ":" ) >= 0 )
		{
			m_strHost = strURL. substring ( 0, strURL. indexOf ( ":" ) );
			if ( strURL. indexOf ( ":" ) + 1 < strURL. length ( ) )
				// There are characters after ':'
				propsInfo. setProperty ( "port", strURL. substring ( strURL. indexOf ( ":" ) + 1 ) );
		}
		else
			m_strHost = strURL;

		m_propsInfo = propsInfo;
		
		int		nPort = DEFAULT_PORT;
		String	strPort = propsInfo. getProperty ( "port" );
		if ( strPort != null )
			try
			{
				nPort = Integer. parseInt ( strPort );
				if ( nPort <= 0 )
					throw new SQLException (
										"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
										"Invalid port - " + strPort );
			}
			catch ( NumberFormatException nfE )
			{
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
									"Invalid port - " + strPort );
			}
			
		String	strUserName = propsInfo. getProperty ( "user" );
		if ( strUserName == null )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: User name is missing" );
		String	strUserPassword = propsInfo. getProperty ( "password" );
		if ( strUserPassword == null )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: User password is missing" );
		
		String	strSSL = propsInfo. getProperty ( "ssl" );	
		try
		{
			if ( strSSL != null && strSSL. equals ( "true" ) )
				m_opConnection = new opDriverManager ( ). getConnection ( m_strHost, nPort, null );
			else
				m_opConnection = new opDriverManager ( ). getConnection ( m_strHost, nPort );
			m_opConnection. setSleepValue ( 0 );
		}
		catch ( opException opE )
		{
			throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
							opE. GetErrorMsg ( ) +
							"\nUnable to connect to " +
							m_strHost +
							" on port " +
							nPort );
		}
		catch ( NoSuchAlgorithmException nsaE )
		{
			throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
							nsaE. getMessage ( ) +
							"\nUnable to connect to " + m_strHost + " on port " + nPort );
		}
		catch ( KeyManagementException kmE )
		{
			throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
							kmE. getMessage ( ) +
							"\nUnable to connect to " + m_strHost + " on port " + nPort );
		}
		catch ( IOException ioE )
		{
			throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
							ioE. getMessage ( ) +
							"\nUnable to connect to " + m_strHost + " on port " + nPort );
		}

		String	strLocalHostIP = "";
		try
		{
			strLocalHostIP = InetAddress. getLocalHost ( ). getHostAddress ( );
			strLocalHostIP = " " + strLocalHostIP;
		}
		catch ( UnknownHostException uhE )
		{ ; }
		catch ( SecurityException sE )
		{ ; }
		
		try
		{
			m_opProcess = m_opConnection. startProcess (
												"4D_JDBC", // station
												strUserName,
												strUserPassword,
												"JDBC Connection" + strLocalHostIP );
		}
		catch ( opException opE )
		{
			throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::ConnectionImpl: " +
							opE. GetErrorMsg ( ) +
							"\nUnable to login" );
		}
		
		String		strOptimization = propsInfo. getProperty ( "com.4d.jdbc.useOptimization" );
		if ( strOptimization != null && strOptimization. toLowerCase ( ). equals ( "true" ) )
			m_bUseOptimization = true;
			
		m_dbMD = new DatabaseMetaDataImpl ( m_opProcess, m_objDataAccessLock, this );

		m_connectionKeeper = new ConnectionKeeper ( m_objDataAccessLock );
		m_connectionKeeper. add ( m_opProcess );
	}
	
	/**
	*	Clears all warnings reported for this Connection object.
	*/
	public void clearWarnings ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		return;
	}

	/**
	*	Releases a Connection's database and JDBC resources  immediately instead of waiting for  them to be automatically released.
	*/
	public void close ( ) throws SQLException
	{
		if ( m_bClosed )
			return;
			
		// TODO: commit current (if any) transaction
		m_connectionKeeper. stopAll ( );
		
		StringBuffer	strbException = new StringBuffer ( "" );

		// Close all Statements that were opened using this connection
		for ( int i = 0; i < m_vctrStatements. size ( ); i++ )
			try
			{
				( ( StatementImpl ) m_vctrStatements. elementAt ( i ) ). close ( );
			}
			catch ( SQLException sqlE )
			{
				// Instead of throwing an exception lets try also to close a connection.
				strbException. append (
								"com.fourd.jdbc.ConnectionImpl::close: Failed to close statement # " +
								+ i + ". " + sqlE. getMessage ( ) );
			}
		
		if ( m_opProcess != null )
			try
			{
				m_opConnection. stopProcess ( m_opProcess );
			}
			catch ( opException opE )
			{
				// Instead of throwing an exception lets try also to close a connection.
				/*strbException. append (
								"com.fourd.jdbc.ConnectionImpl::close: Failed to stop process. " +
								opE. GetErrorMsg ( ) );*/
			}
		
		if ( m_opConnection != null )
			try
			{
				m_opConnection. CloseConnection ( );
			}
			catch ( opException opE )
			{
				/*strbException. append (
								"\ncom.fourd.jdbc.ConnectionImpl::close: Failed to close connection. " +
								opE. GetErrorMsg ( ) );*/
			}
		
		if ( strbException. length ( ) > 0 )
			throw new SQLException ( strbException. toString ( ) );
			
		m_bClosed = true;
	}
	
	/**
	*	Makes all changes made since the previous  commit/rollback permanent and releases any database locks  currently held by the Connection.
	*/
	public void commit ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
		
		if (	m_nTransactionIsolation == TRANSACTION_NONE ||
				!m_bTransactionInProgress )
			// No need to commit - there is no transaction
			return;
			
		// Lets commit current transaction
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				m_bTransactionInProgress = false;
				m_opProcess. ValidateTransaction ( );
// System. out. println ( "Committing transaction" );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::commit: " +
									"Failed to commit transaction.\n" +
									opE. getMessage ( ) );
			}
		}
	} // end of public void commit ( ) throws SQLException

	/**
	*	Creates a Statement object for sending  SQL statements to the database.
	*/
	public Statement createStatement ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		ensureTransaction ( );
		
		StatementImpl	sImpl = new StatementImpl ( this, m_opConnection, m_opProcess, m_dbMD, m_objDataAccessLock );
		m_vctrStatements. addElement ( sImpl );
		
		return sImpl;
	}

	/**
	*	Creates a Statement object that will generate  ResultSet objects with the given type and concurrency.
	*/
	public Statement createStatement ( int nResultSetType, int nResultSetConcurrency ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		// TODO: implement properly and completely
		
		return createStatement ( );
	}

	/**
	*	Gets the current auto-commit state.
	*/
	public boolean getAutoCommit ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		return m_bAutoCommit;
	}

	/**
	*	Returns the Connection's current catalog name.
	*/
	public String getCatalog ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "getCatalog" );
		
		return null;
	}

	/**
	*	Gets the metadata regarding this connection's database.
	*/
	public DatabaseMetaData getMetaData ( ) throws SQLException
	{
		return m_dbMD;
	}

	/**
	*	Gets this Connection's current transaction isolation level.
	*/
	public int getTransactionIsolation ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		return m_nTransactionIsolation;
	}

	/**
	*	Gets the type map object associated with this connection.
	*/
	public Map getTypeMap ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "getTypeMap" );
		
		return null;
	}

	/**
	*	Returns the first warning reported by calls on this Connection.
	*/
	public SQLWarning getWarnings ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		return null;
	}

	/**
	*	Tests to see if a Connection is closed.
	*/
	public boolean isClosed ( ) throws SQLException
	{
		return m_bClosed;
	}

	/**
	*	Tests to see if the connection is in read-only mode.
	*/
	public boolean isReadOnly ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "isReadOnly" );
		
		return true;
	}

	/**
	*	Converts the given SQL statement into the system's native SQL grammar.
	*/
	public String nativeSQL ( String strSQL ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		return strSQL;
	}

	/**
	*	Creates a CallableStatement object for calling  database stored procedures.
	*/
	public CallableStatement prepareCall ( String strSQL ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
		
		if ( strSQL == null )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::prepareCall: " +
								"SQL string can not be null" );

		ensureTransaction ( );
		
		CallableStatementImpl		cStatementImlp = new CallableStatementImpl (
																	this,
																	m_opConnection,
																	m_opProcess,
																	m_dbMD,
																	m_objDataAccessLock,
																	strSQL );
							
		return cStatementImlp;
	}

	/**
	*	Creates a CallableStatement object that will generate  ResultSet objects with the given type and concurrency.
	*/
	public CallableStatement prepareCall ( String strSQL, int nResultSetType, int nResultSetConcurrency ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "prepareCall" );
		
		return null;
	}

	/**
	*	Creates a PreparedStatement object for sending  parameterized SQL statements to the database.
	*/
	public PreparedStatement prepareStatement ( String strSQL ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		if ( strSQL == null )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::prepareStatement: " +
								"SQL string can not be null" );

		ensureTransaction ( );
		
		PreparedStatementImpl		pStatementImlp = new PreparedStatementImpl (
																	this,
																	m_opConnection,
																	m_opProcess,
																	m_dbMD,
																	m_objDataAccessLock,
																	strSQL );
							
		return pStatementImlp;
	}

	/**
	*	Creates a PreparedStatement object that will generate  ResultSet objects with the given type and concurrency.
	*/
	public PreparedStatement prepareStatement ( String strSQL, int nResultSetType, int nResultSetConcurrency ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		//throwUnimplementedException ( "prepareStatement" );
		
		//TODO: implement properly
		return prepareStatement ( strSQL );
	}

	/**
	*	Drops all changes made since the previous  commit/rollback and releases any database locks currently held  by this Connection.
	*/
	public void rollback ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		if (	m_nTransactionIsolation == TRANSACTION_NONE ||
				!m_bTransactionInProgress )
			// There is no transaction in progress, no need to rollback
			return;
			
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				m_bTransactionInProgress = false;
				m_opProcess. CancelTransaction ( );
// System. out. println ( "Cancelling transaction" );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::rollback: " +
									"Failed to rollback transaction.\n" +
									opE. getMessage ( ) );
			}
		}
	} // end of public void rollback ( ) throws SQLException

	/**
	*	Sets this connection's auto-commit mode.
	*/
	public void setAutoCommit ( boolean bAutoCommit ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
		
		if (	m_nTransactionIsolation == TRANSACTION_NONE &&
				!bAutoCommit )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::setAutoCommit: " +
								"Can't disable auto-commit because current transaction " +
								"isolation level is TRANSACTION_NONE");
			
		m_bAutoCommit = bAutoCommit;
		if ( bAutoCommit && m_bTransactionInProgress )
			/*
				If the auto commit is being enabled and there is a transaction in progress
				then I have to commit that transaction.
			*/
			commit ( );
	}

	/**
	*	Sets a catalog name in order to select a subspace of this Connection's database in which to work.
	*/
	public void setCatalog ( String strCatalog ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "setCatalog" );
		
		return;
	}

	/**
	*	Puts this connection in read-only mode as a hint to enable   database optimizations.
	*/
	public void setReadOnly ( boolean bReadOnly ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "setReadOnly" );
		
		return;
	}

	/**
	*	Attempts to change the transaction  isolation level to the one given.
	*/
	public void setTransactionIsolation ( int nLevel ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
		
		if ( nLevel == TRANSACTION_NONE )
			// Commit current transaction (if any)
			commit ( );
		else if ( nLevel == TRANSACTION_READ_UNCOMMITTED )
		{
			/*
				If nLevel was TRANSACTION_NONE then nothing should change because
				I simly enable previously disabled transactoins.
				If nLevel was TRANSACTION_READ_UNCOMMITTED then obviously nothing
				chanegs too. So this is an empty 'else if' block basically.
			*/
		}
		else
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::setTransactionIsolation: " +
								"Unsupported transaction isolation level - " + nLevel );
								
		m_nTransactionIsolation = nLevel;
	}

	/**
	*	Installs the given type map as the type map for  this connection.
	*/
	public void setTypeMap ( Map map ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl:: Connection is closed" );
			
		throwUnimplementedException ( "setTypeMap" );
			
		return;
	}

	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.ConnectionImpl::" +
							strMethodName +
							": Method is not implemented" );
	}
	
// !!!!!!!!!!!!!!!!!!!! ------------------ Methods that are not part of the Connection interface ------------------ !!!!!!!!!!!!!!!!!!!! 

	private void s_e_p_a_r_a_t_o_r__________________________1 ( )
	{
		;
	}
	
	/**
	*	Returns array of table names that were requested to be registered and last time
	*	were used by a TableRegisterer other then objTableRegisterer.
	*	Tables in the returned array should be "re-selected".
	*
	*	Always returns non-null array
	*/
	synchronized String [ ] registerTables ( String [ ] szstrTableNames, Object objTableRegisterer ) throws SQLException
	{
		if ( szstrTableNames == null )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl::registerTables: Table names must not be null" );
			
		// Lets make sure that no tables are being registered twice
		String [ ]		szstrAlreadyRegistered = extractRegisteredTables ( szstrTableNames, objTableRegisterer );
		if ( szstrAlreadyRegistered. length > 0 )
			throw new SQLException (
								"com.fourd.jdbc.ConnectionImpl::registerTables: Table " +
								szstrAlreadyRegistered [ 0 ] +
								" is already registered." );
			
		long		nWaiting = 0;
		while ( countRegisteredTables ( szstrTableNames ) > 0 )
		{
			try
			{
				synchronized ( m_objTableRegisterSync )
				{
					m_objTableRegisterSync. wait ( 10000 );
				}
				nWaiting += 10000;
			}
			catch ( InterruptedException iE )
			{
				;
			}
			
			if ( countRegisteredTables ( szstrTableNames ) == 0 ) // All our tables are unregistered now
				break;
				
			System. err. println (
							"com.fourd.jdbc.ConnectionImpl::registerTables: Waiting " +
							nWaiting / 1000 + " seconds to register tables." );
		}
		
		Vector		vctrTablesToReselect = new Vector ( );
		Object		objLastRegisterer = null;
		// Register tables
		for ( int i = 0; i < szstrTableNames. length; i++ )
		{
			m_htTablesRegister. put ( szstrTableNames [ i ]. toUpperCase ( ), objTableRegisterer );
			
			objLastRegisterer = m_htTableUsers. put ( szstrTableNames [ i ]. toUpperCase ( ), objTableRegisterer );
			if ( objLastRegisterer != null && objLastRegisterer != objTableRegisterer ) // A table registerer other then objTableRegisterer was the last one to access table szstrTableNames [ i ]
			{
				vctrTablesToReselect. addElement ( szstrTableNames [ i ] );
			}
		}
			
		return ( String [ ] ) vctrTablesToReselect. toArray ( new String [ vctrTablesToReselect. size ( ) ] );
	}
	
	synchronized void unregisterTables ( String [ ] szstrTableNames, Object objTableRegisterer ) throws SQLException
	{
		if ( szstrTableNames == null )
			throw new SQLException ( "com.fourd.jdbc.ConnectionImpl::unregisterTables: Table names must not be null" );
			
		for ( int i = 0; i < szstrTableNames. length; i++ )
		{
			if ( !m_htTablesRegister. containsKey ( szstrTableNames [ i ]. toUpperCase ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::unregisterTables: Attempt to unregister " +
									"a non-registered table - " + szstrTableNames [ i ] );
			
			
			if ( objTableRegisterer != m_htTablesRegister. remove ( szstrTableNames [ i ]. toUpperCase ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::unregisterTables: Table " +
									szstrTableNames [ i ] +
									" is being unregistered not by the original registerer" );
		}
		
		// Let know all those who are waiting to register tables
		synchronized ( m_objTableRegisterSync )
		{
			m_objTableRegisterSync. notifyAll ( );
		}
	}
	
	/**
	*	Always returns non-null array
	*
	*	Returns all those tables from szstrTableNames that are already registered for ResultSet rs.
	*	Returns 0-length array if none of the tables from szstrTableNames are registered for ResultSet rs.
	*/
	private String [ ] extractRegisteredTables ( String [ ] szstrTableNames, Object objTableRegisterer )
	{
		Vector		vctrResult = new Vector ( );
		for ( int i = 0; i < szstrTableNames. length; i++ )
			if ( m_htTablesRegister. get ( szstrTableNames [ i ]. toUpperCase ( ) ) == objTableRegisterer )
				vctrResult. addElement ( szstrTableNames [ i ]. toUpperCase ( ) );
				
		return ( String [ ] ) vctrResult. toArray ( new String [ vctrResult. size ( ) ] );
	}
	
	/**
	*	Returns the amount of tables from szstrTableNames that are registered.
	*/
	private int countRegisteredTables ( String [ ] szstrTableNames )
	{
		int			nResult = 0;
		for ( int i = 0; i < szstrTableNames. length; i++ )
			if ( m_htTablesRegister. containsKey ( szstrTableNames [ i ]. toUpperCase ( ) ) )
				nResult++;
				
		return nResult;
	}
	
	synchronized String generateSelectionName ( )
	{
		m_nSelectionID++;
		
		return "jdbcselname" + m_nSelectionID;
	}

	synchronized String generateSetName ( )
	{
		m_nSetID++;
		
		return "jdbcsetname" + m_nSetID;
	}
	
	/**
	*	Starts new transaction if necessary.
	*/
	private void ensureTransaction ( ) throws SQLException
	{
		if (	m_nTransactionIsolation == TRANSACTION_NONE ||
				m_bAutoCommit ||
				m_bTransactionInProgress )
			// No need to start a new transaction
			return;
			
		/*
			Auto-commit is disabled and there is no transaction in progress,
			thus I need to start a new one.
		*/
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				m_opProcess. StartTransaction ( );
// System. out. println ( "Starting transaction" );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.ConnectionImpl::ensureTransaction: " +
									"Failed to start transaction.\n" +
									opE. getMessage ( ) );
			}
			m_bTransactionInProgress = true;
		}
	} // end of private void ensureTransaction ( )
	
	String getHost ( )
	{
		return m_strHost;
	}
	
	String getUserName ( )
	{
		return m_propsInfo. getProperty ( "user" );
	}
	
	opProcess getopProcess ( )
	{
		return m_opProcess;
	}
	
	Object getDataAccessLock ( )
	{
		return m_objDataAccessLock;
	}
	
	boolean isUsingOptimization ( )
	{
		return m_bUseOptimization;
	}
}