/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

/**
*	Wraps all 4D functionality that does not map to JDBC.
*/
public final class FDExtensions
{
	private opProcess			m_opProcess = null;
	private Object				m_objDataAccessLock = null;
	
	private ConnectionImpl		m_connectionImpl = null;
	
	/**
	*	Constructs FDExtensions object from a given connection.
	*	Note, that Connection object must be generated by 4D JDBC driver.
	*	Otherwise, an SQLException is thrown.
	*	For more information on each particular method, please see
	*	4D documentation.
	*/
	public FDExtensions ( Connection conn ) throws SQLException
	{
		if ( conn == null )
			throw new SQLException (
								"com.fourd.jdbc.FDExtensions::FDExtensions: " +
								"Connection can not be null" );
								
		if ( !( conn instanceof com.fourd.jdbc.ConnectionImpl ) )
			throw new SQLException (
								"com.fourd.jdbc.FDExtensions::FDExtensions: " +
								"Incompatible connection class; 4D connection is required" );
								
		m_connectionImpl = ( ConnectionImpl ) conn;
		m_opProcess = m_connectionImpl. getopProcess ( );
		m_objDataAccessLock = m_connectionImpl. getDataAccessLock ( );
	}
	
	/**
	*	For more information see 4D documentation on semaphores.
	*/
	public boolean setSemaphore ( String strName ) throws SQLException
	{
		if ( strName == null )
			throw new SQLException (
								"com.fourd.jdbc.FDExtensions::setSemaphore: " +
								"Semaphore name can not be null" );
		
		boolean			bResult = false;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				bResult = m_opProcess. SetSemaphore ( strName );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::setSemaphore: " +
									"Failed to set semaphore.\n" +
									opE. getMessage ( ) );
			}
		}
		
		return bResult;
	} // end of public setSemaphore ( String srtName ) throws SQLException

	/**
	*	For more information see 4D documentation on semaphores.
	*/
	public void clearSemaphore ( String strName ) throws SQLException
	{
		if ( strName == null )
			throw new SQLException (
								"com.fourd.jdbc.FDExtensions::clearSemaphore: " +
								"Semaphore name can not be null" );
		
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				m_opProcess. ClearSemaphore ( strName );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::clearSemaphore: " +
									"Failed to clear semaphore.\n" +
									opE. getMessage ( ) );
			}
		}
	} // end of public void clearSemaphore ( String strName ) throws SQLException
	
	/*
		Blocks until a given semaphore lock is obtained or a specified amount of time
		elapsed.
		String		strName - semaphore name
		int			nWaitTime - waits indefinitely if nWaitTime <= 0, otherwise
			waits no more than nWaitTime seconds.
			
		Returns true if a semaphore lock was obtained within a given period of
		time, otherwise returns false.
		
		Throws an exception on error.
	*/
	boolean waitForSemaphore ( String strName, int nWaitTime ) throws SQLException
	{
		if ( strName == null )
			throw new SQLException (
								"com.fourd.jdbc.FDExtensions::waitForSemaphore: " +
								"Semaphore name can not be null" );

		long		nStart = System. currentTimeMillis ( );
		while ( setSemaphore ( strName ) )
		{
			if ( nWaitTime > 0 && nWaitTime < ( System. currentTimeMillis ( ) - nStart ) / 1000 )
				return false;
				
//System. out. println ( "Waiting for " + ( ( System. currentTimeMillis ( ) - nStart ) / 1000 ) + " seconds" );
		}
		
		return true;
	}
	
	/**
	*	Returns number of user processes currently running.
	*	This number excludes the processes managed by 4D Server - 
	*	User Interface, Client Manager, Cache Manager, Web Server, and Index Builder.
	*/
	public short getUserProcessCount ( ) throws SQLException
	{
		short		nResult = 0;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				nResult = m_opProcess. CountUserProcesses ( );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::getUserProcessCount: " +
									"Failed to count user processes.\n" +
									opE. getMessage ( ) );
			}
		}
		
		return nResult;
	} // end of public short getUserProcessCount ( ) throws SQLException

	/**
	*	Returns number of users currently connected to the server.
	*/
	public short getUserCount ( ) throws SQLException
	{
		short		nResult = 0;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				nResult = m_opProcess. CountUsers ( );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::getUserCount: " +
									"Failed to count users.\n" +
									opE. getMessage ( ) );
			}
		}
		
		return nResult;
	} // end of public short getUserCount ( ) throws SQLException
	
	/**
	*	Immediately saves the data buffers to the disk.
	*/
	public void flushBuffers ( ) throws SQLException
	{
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				m_opProcess. FlushBuffers ( );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::flushBuffers: " +
									"Failed to flush buffers.\n" +
									opE. getMessage ( ) );
			}
		}
	} // end of public void flushBuffers ( ) throws SQLException

	/**
	*	Returns the date of the 4D Server machine.
	*/
	public GregorianCalendar getServerDate ( ) throws SQLException
	{
		GregorianCalendar		gcResult = null;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				gcResult = m_opProcess. GetDate ( );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::getServerDate: " +
									"Failed to get date.\n" +
									opE. getMessage ( ) );
			}
		}
		
		return gcResult;
	} // end of public GregorianCalendar getServerDate ( ) throws SQLException

	/**
	*	Returns the time of the 4D Server machine.
	*/
	public GregorianCalendar getServerTime ( ) throws SQLException
	{
		GregorianCalendar		gcResult = null;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				gcResult = m_opProcess. GetTime ( );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.FDExtensions::getServerTime: " +
									"Failed to get time.\n" +
									opE. getMessage ( ) );
			}
		}
		
		return gcResult;
	} // end of public GregorianCalendar getServerTime ( ) throws SQLException
}