/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

/**
*	Represents the actual data of the result set produced by SELECT command.
*	
*	Data is prefetched on a row basis. If the data requested is not in the cache then
*		it is requested from the server, returned to client and put into cache.
*	Current implementation fetches rows one by one but it is definitely possible to fetch rows in blocks.
*	Well, if not fetch in blocks then at least pretend that they are fetched in blocks.
*
*	Current implementation uses Hashtable to store rows. Each row is represented by Object [ ].
*
*
*	This as well may be the part of the ResultSet class instead of being a separate class, but ResultSet has
*	enough methods of its own.
*/
class RSDataCache
{
	// A reference to the connection that generated instance of this class
	private ConnectionImpl			m_connectionImpl = null;
	
	/**
	*	A reference to the statement that generated this data cache.
	*/
	private StatementImpl			m_statementImpl = null;

	private ResultSetImpl				m_rsImpl = null;
	
	/*
		Storage for retrieved rows. Each row can be identified by its 0-based index represented by String.
		For example, to extract row number 12 one would write:
			opData [ ]		szopdRow = ( opData [ ] ) m_htData. get ( "11" );
	*/
	private Hashtable				m_htData = new Hashtable ( );
	
	/*
		SQLCommand instance that produced ResultSet that uses instance of this class
	*/
	private SQLCommand 			m_sqlCommand = null;
	
	/*
		Meta data about a table whose data is represented in the instance if this class.
		Basically a FROM table.
		Later I may need to use array of SQLTableMetaData to handle joins.
		This is a DB version of meta data, not the one produced while parsing the query.
	*/
	private SQLTableMetaData		m_sqltDBMD = null;
	
	private opProcess				m_opProcess = null;
	
	/**
		Should always synchronize on m_objDataAccessLock before accessing m_opProcess.
		Can't synchronize on m_opProcess itself.
	*/
	private Object						m_objDataAccessLock = null;
	
	private DatabaseMetaDataImpl		m_dbMD = null;
	
	// Helpers
	/*
		Instead of creating helper 4D Open for Java objects each time we need to access opProcess, lets create them
		only once and store them in class members.
	*/
	
	/*
		FROM and JOIN table names are used as keys. With each table name we associate an opFieldArray
		that we will use to fetch rows for this table.
	*/
	private Hashtable				m_htFieldArrayHelpers = new Hashtable ( );
	
	/*
		FROM and JOIN table names are used as keys. With each table name we associate an int[]
		that contains indexes of the columns in the SELECT clause of that table.
	*/
	private Hashtable				m_htIndexArrayHelpers = new Hashtable ( );
	
	/*
		JOIN table names are used as keys. Each element is an array of Strings that contains
		names of columns that are not used directly in SELECT clause but rather in JOIN clause.
		Strings in array are in proper order to correspond to elements of m_htFieldArrayHelpers.
		
		m_htJOINedColumnNamesHelper = 
		{
			...
			< "TableName", String [ ] = { ..., "ColumnName", ... }>,
			...
		}
	*/
	private Hashtable				m_htJOINedColumnNamesHelper = new Hashtable ( );
	
	/*
		JOIN table names are used as keys. Each element is a hashtable.
		1-based row indexes (represented by strings) are keys in each sub-hashtable.
		With each row index an array is associated. Each array contains values for
		columns with names as in arrays in m_htJOINedColumnNamesHelper.

		m_htJOINedValues = 
		{
			...
			< "TableName", Hashtable =
									{
										...
										< "RowNumber", opData [ ] >,
										...
									} >,
			...
		}
	*/
	private Hashtable				m_htJOINedValues = new Hashtable ( );
	
	private int						m_nSELECTColumnCount = 0;

	RSDataCache (
			SQLCommand sqlCommand,
			SQLTableMetaData sqltDBMD,
			ConnectionImpl connectionImpl,
			StatementImpl statementImpl,
			ResultSetImpl rsImpl,
			opProcess opP,
			Object objDataAccessLock,
			DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		m_sqlCommand = sqlCommand;
		m_sqltDBMD = sqltDBMD;
		m_connectionImpl = connectionImpl;
		m_statementImpl = statementImpl;
		m_rsImpl = rsImpl;
		m_opProcess = opP;
		m_objDataAccessLock = objDataAccessLock;
		m_dbMD = dbMD;
		
		m_nSELECTColumnCount = sqlCommand. getSELECTColumnCount ( );
		
		prepareHelpers ( );
	}
	
	/**
	*	The only use is for non-empty read-only (disconnected) result sets.
	*/
	RSDataCache ( opData [ ] [ ] szopData )
	{
		m_nSELECTColumnCount = szopData [ 0 ]. length;
		
		int					nRowCount = szopData. length;
		for ( int i = 1; i <= nRowCount; i++ )
			m_htData. put ( String. valueOf ( i ), szopData [ i - 1 ] );
	}
	
	/**
	*	Each time we need to fetch row data, we need to use opFieldArray. Instead of creating them each time,
	*	we will create them only once and reuse when possible.
	*	For each FROM and JOIN table we also need to know indexes of its columns in the SELECT clause so that we can cache
	*	the data properly. Again, we will calculate them only once and then reuse when possible.
	*
	*	For example, we need to handle SELECT t1.c1, t2.k1, t1.c4, t2.k2 FROM ...
	*	We need to save the facts that :
	*		- for table t1 we only need to fetch columns c1 and c4. 
	*		- index of the column c1 is 0 and index of column c4 is 2		// { 0, 2 }
	*		- for table t2 we only need to fetch columns k1 and k2. 
	*		- index of the column k1 is 1 and index of column k2 is 3		// { 1, 3 }
	*/
	private void prepareHelpers ( ) throws SQLException
	{
		SQLTableMetaData [ ]			szsqltMD = m_sqlCommand. getFROMTables ( );
		for ( int i = 0; i < szsqltMD. length; i ++ ) // although we know that there is only one FROM table
			prepareHelpers ( szsqltMD [ i ] );

		szsqltMD = m_sqlCommand. getJOINTables ( );
		for ( int i = 0; i < szsqltMD. length; i ++ )
			prepareHelpers ( szsqltMD [ i ] );
	} // end of private void prepareHelpers ( ) throws SQLException
	
	/**
	*	Prepares helpers for a particular table.
	*/
	private void prepareHelpers ( SQLTableMetaData sqltMD ) throws SQLException
	{
		if ( sqltMD == null )
			throw new SQLException ( "com.fourd.jdbc.RSDataCache::prepareHelpers: Table can not be null" );
			
		SQLTableMetaData 		sqltDBMD = m_dbMD. getTable ( sqltMD. getDBName ( ) );
		m_htJOINedValues. put ( sqltMD. getDBName ( ). toUpperCase ( ), new Hashtable ( ) ); // prepare storage for each table
		SQLColumnMetaData [ ]		szsqlcMD = m_sqlCommand. getSELECTColumns ( sqltMD. getDBName ( ) );
		
		// Array of columns that are not used in SELECT clause but are used for JOINs
		String [ ]					szstrHiddenJOINedToColumns = m_sqlCommand. getColumnsJOINed (
																						sqltMD. getDBName ( ),
																						false );
	
		m_htJOINedColumnNamesHelper. put (
										sqltMD. getDBName ( ). toUpperCase ( ),
										szstrHiddenJOINedToColumns );
/*System. out. println (
				"Table " + sqltMD. getDBName ( ) + " uses " +
				szstrHiddenJOINedToColumns. length + " hidden column(s) for JOIN(s)" );*/
		
		opFieldArray				opfaHelper = new opFieldArray (
														szsqlcMD. length +
														szstrHiddenJOINedToColumns. length );
		opfaHelper. mTargetTable = sqltDBMD. getNumber ( );
		for ( int i = 0; i < szsqlcMD. length; i++ )
		{
			opfaHelper. mFieldArray [ i ] = new opField (
												sqltDBMD. getNumber ( ),
												sqltDBMD. getColumnNumber ( szsqlcMD [ i ]. getDBName ( ) ) );
																
			// Lets initialize field types
			opfaHelper. mFieldArray [ i ]. mFieldType = sqltDBMD. getColumnMetaData (
																	szsqlcMD [ i ]. getDBName ( ) ).
																				getopType ( );
		}
		for ( int i = 0; i < szstrHiddenJOINedToColumns. length; i++ )
		{
			opfaHelper. mFieldArray [ i + szsqlcMD. length ] = new opField (
															sqltDBMD. getNumber ( ),
															sqltDBMD. getColumnNumber (
																		szstrHiddenJOINedToColumns [ i ] ) );
																
			// Lets initialize field types
			opfaHelper. mFieldArray [ i + szsqlcMD. length ]. mFieldType = sqltDBMD. getColumnMetaData (
																	szstrHiddenJOINedToColumns [ i ] ).
																		getopType ( );
		}
		
		m_htFieldArrayHelpers. put (
								sqltMD. getDBName ( ). toUpperCase ( ),
								opfaHelper );
		m_htIndexArrayHelpers. put (
								sqltMD. getDBName ( ). toUpperCase ( ),
								m_sqlCommand. getSELECTColumnIndexes ( sqltMD. getDBName ( ) ) );
	} // end of private void prepareHelpers ( SQLTableMetaData sqltMD ) throws SQLException
	
	/**
	*	Removes row from the cache. Updates index of previously cached rows by
	*	decreasing it by 1. Lets say I have rows 1, 2, 3, 4, 5, 8, and 9 cached. Row
	*	#3 is being deleted. Then I have to decrease index for rows 4, 5, 8, and 9.
	*	There will be the following rows cached after the method is done then:
	*		1, 2, 3, 4, 7, and 8. Note that #3 is not the old #3, it is the old #4.
	*/
	void deleteRow ( int nRow )
	{
		// Delete row from main cache
		deleteRow ( nRow, m_htData );

		// Delete row from JOINed cache.
		/*
			TODO: possible optimization. Delete row from joined cache only if 
			it was in fact deleted from main cache.
		*/
		Enumeration			enum = m_htJOINedValues. elements ( );
		while ( enum. hasMoreElements ( ) )
			deleteRow ( nRow, ( Hashtable ) enum. nextElement ( ) );
	} // end of void deleteRow ( int nRow )
	
	private void deleteRow ( int nRow, Hashtable ht )
	{
		ht. remove ( String. valueOf ( nRow ) );
		
		// Reduce row numbers by 1 for all those that are greater then nRow.
		int				nRowCount = m_rsImpl. getRowCount ( );
		Object			objRow = null;
		String			strRowNumber = null;
		for ( int i = nRow + 1; i <= nRowCount; i++ )
		{
			strRowNumber = String. valueOf ( i );
			if ( ( objRow = ht. get ( strRowNumber ) ) == null )
				continue;
				
			ht. put (
					String. valueOf ( i - 1 ),
					ht. remove ( strRowNumber ) );
		}
	}
	
	/**
	*	Removes row from the cache but DOES NOT change order numbers for other rows
	*	(like deleteRow does). Thus, if I have rows 1, 2, 3, 4, 5, 8, and 9 cached and
	*	row #3 is being deleted, then there will be the following rows cached after
	*	the method is done:
	*		1, 2, 4, 5, 8, and 9.
	*	So row #3 is gone and nothing else is changed.
	*	Mthod does nothing if the row does not exist.
	*/
	void removeRow ( int nRow )
	{
		// Remove row from main cache
		m_htData. remove ( String. valueOf ( nRow ) );

		// Remove row from JOINed cache.
		/*
			TODO: possible optimization. Remove row from joined cache only if 
			it was in fact removed from main cache.
		*/
		Enumeration			enum = m_htJOINedValues. elements ( );
		while ( enum. hasMoreElements ( ) )
			( ( Hashtable ) enum. nextElement ( ) ). remove ( String. valueOf ( nRow ) );
	} // end of void removeRow ( int nRow )
	
	boolean rowIsCached ( int nRow )
	{
		return m_htData. get ( String. valueOf ( nRow ) ) != null;
	}
	
	/**
	*	I can't use the fetch size from the resultset as is because deleted rows create "holes"
	*	in the current selection. Thus, if I try to fetch a block with a hole in it, an exception
	*	will be thrown by 4d open for java. So what this method does is it takes nCurrentRow
	*	and figures out the size of a largest block of rows without holes in it so that nCurrentRow
	*	is the first row of that block and the size of such a block does not exceed current fetch
	*	size of the resultset.
	*	Also, some of the rows may already be prefetched and updated with new values. This is fine for
	*	the main FROM table but bad for updated values from JOINed columns. Updated values from JOINed
	*	columns may be replaced by old values when the whole block is fetched. Thus already prefetched
	*	rows must not be refetched.
	*	The task is accomplished by calculating nCurrentRow's original index in the original selection
	*	and then comparing that original index to the orininal index of the next row. If they differ exactly
	*	by one then there are no holes between current and the next row. Operation is then repeated
	*	for second and third row, etc... If the difference between the original indexes of two neighbour rows
	*	is more then 1 then there is a hole and thus method stops and returns obtained block size. Also, 
	*	if the next row is alredy prefetched then method also stops.
	*
	*	TODO:Possible optimization. Row can be refetched if there were no changes to it. For example:
	*	ResultSet contains 10 rows, fetch size is 5. I read values from rows #4 and #6. They are thus cached.
	*	Now I want to walk through rows and get one value from each row. Here is how block fetching
	*	will look : 
	*		1. Rows #1-#3
	*		2. Row #5
	*		3. Rows #7-#10
	*	3 fetches in total. Whereas, it would be only two fetches - #1-#5 and #6-#10 if I knew that 
	*	"unupdated" rows can be safely refetched.
	*
	*	TODO: method should keep in mind newly created rows.
	*/
	private int calculateFetchBlockSize ( int nCurrentRow ) throws SQLException
	{
		int			nFetchSize = m_rsImpl. getFetchSize ( );
		int			nRowCount = m_rsImpl. getOriginalRowCount ( );
		if ( nCurrentRow + nFetchSize > nRowCount + 1 )
			nFetchSize = nRowCount - nCurrentRow + 1;
			
		// So the fetch block contains no more than nFetchSize rows. Lets see if there are any holes
		
		int		nResult = 1;
		int		nCurrentOriginalIndex = m_rsImpl. mapRowToSelection ( nCurrentRow );
		int		nNextOriginalIndex = -1;
		for ( int i = 0; i < nFetchSize - 1; i++ )
		{
			nNextOriginalIndex = m_rsImpl. mapRowToSelection ( nCurrentRow + 1 );
			if (
				( nCurrentOriginalIndex + 1 != nNextOriginalIndex ) || // a hole in the selection
				rowIsCached ( nCurrentRow + 1 ) ) // row is already cached, stop; must not refetch
				return nResult;
				
			nResult++;
			nCurrentRow++;
			nCurrentOriginalIndex++;
		}
		
// System. out. println ( "RSDataCache::calculateFetchBlockSize: " + nResult );
		return nResult;
	}
	
	/**
	*	Returns value of a designated column in the current row of the result set.
	*
	*	Current implementation assumes that there is only one table to select from, there were no joins.
	*	nColumnQueryIndex is 0-based. nCurrentRow is 1-based.
	*	
	*	Throws an SQLException if anything goes wrong.
	*
	*
	*	TODO: there is a bug in the 4D Open for Java, the amount of rows in the empty selection is not
	*	0 but 1. I should handle this situation properly.
	*/
	Object getValue ( int nCurrentRow, int nColumnQueryIndex ) throws SQLException
	{
		if ( nColumnQueryIndex < 0 || nColumnQueryIndex >= m_nSELECTColumnCount )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::getValue: Column index " +
								nColumnQueryIndex + " is out of range" );
		
		opData [ ]			szopdRow = ( opData [ ] ) m_htData. get ( String. valueOf ( nCurrentRow ) );
		if ( szopdRow != null )
		{
			// Row is already pre-fetched and cached
			return convertToJavaObject ( szopdRow [ nColumnQueryIndex ] );
		}
		else
		{
			try
			{
				// Request tables for exclusive use
				m_rsImpl. registerTables ( );
					fetchValue ( nCurrentRow, nColumnQueryIndex );
                // Release the tables
				m_rsImpl. unregisterTables ( );
				
				// Now that the data is fetched and cached, let call getValue again.
				return getValue ( nCurrentRow, nColumnQueryIndex );
			}
			catch ( SQLException sqlE )
			{	
				m_rsImpl. unregisterTables ( );
				
sqlE. printStackTrace ( );
				throw new SQLException (
									"com.fourd.jdbc.RSDataCache::getValue: " + sqlE. getMessage ( ) +
									"\nFailed to fetch row data." );
			}
		}
	} // end of Object getValue  ...
	
	/**
	*	Fetches the value for a given row and colunm from database.
	*	Actually, it fetches not only one value but also the whole block of rows from main
	*	table and all tables joined to it. It also caches fetched data.
	*
	*	This method assumes that a call to registerTables ( ) was
	*	already made and the unregisterTables ( ) will be called afterwords.
	*
	*	nColumnQueryIndex is 0-based. nCurrentRow is 1-based.
	*/
	/*
		Here is what I need to do to implement JOINs.
		
DONE		1. Fetch and cache data range for FROM table
			2. Add tables with cached data to the FetchedVector
			3. Get all tables joined directly to the tables in the FetchedVector
				4. For each found table compose a WHERE request using join condition and fetch the data for that table
				5. Using join conditions find and cache fetched data.
			6. If there are more joined tables then GoTo step 2 else END.
	*/
	private opData fetchValue ( int nCurrentRow, int nColumnQueryIndex ) throws SQLException
	{
		// First, lets fetch the data for the main FROM table and cache it.
		// Need to fetch a block of rows
			
		// Lets calculate the fetch block size
		// int			nFetchSize = m_rsImpl. getFetchSize ( );
		// int			nRowCount = m_rsImpl. getRowCount ( );
				
		// Row indexes that are used in calls to opProcess. RangeSelectionToArray (...) are 0-based
		int		nMin = nCurrentRow - 1;
		int		nFetchBlockSize = calculateFetchBlockSize ( nCurrentRow );
		int		nMax = nMin + nFetchBlockSize - 1;

		// Lets extract opFieldArray helper for FROM table
		String				strFROMTable = m_sqltDBMD. getDBName ( );
		opFieldArray		opfaFROMHelper = ( opFieldArray ) m_htFieldArrayHelpers. get (
																			strFROMTable. toUpperCase ( ) );
		
		opDataArray [ ]		szopdaRows = new opDataArray [ opfaFROMHelper. mFieldArray. length ];
		for ( int i = 0; i < opfaFROMHelper. mFieldArray. length; i++ )
		{
			szopdaRows [ i ] = new opDataArray ( nMax - nMin + 1 );
			for ( int j = 0; j < szopdaRows [ i ]. mDataArray. length; j++ )
				szopdaRows [ i ]. mDataArray [ j ]. mType = ( short ) opfaFROMHelper.
																		mFieldArray [ i ].
																					mFieldType;
		}
			
		try
		{
			/*
				Before actually doing anything with opProcess, we have to acquire a sync lock
				because there may be multiple threads trying to access opProcess object simultaneously.
			*/
			synchronized ( m_objDataAccessLock )
			{
				int		nFirst = m_rsImpl. mapRowToSelection ( nCurrentRow ) - 1;
				m_opProcess. RangeSelectionToArray (
												szopdaRows,
												opfaFROMHelper,
												nFirst, // nMin,
												nFirst + nFetchBlockSize - 1 );
			}
				
			// Lets extract column index helper for FROM table
			int [ ]				sznColumnIndexes = ( int [ ] ) m_htIndexArrayHelpers. get (
																			strFROMTable. toUpperCase ( ) );
			opData [ ]			szopdRow = cacheRows (
													nCurrentRow,
													szopdaRows,
													sznColumnIndexes,
													( String [ ] ) m_htJOINedColumnNamesHelper. get (
																			strFROMTable. toUpperCase ( ) ),
													( Hashtable ) m_htJOINedValues. get (
																			strFROMTable. toUpperCase ( ) ) );
			
			fetchJoinedValues (
							nCurrentRow,
							nMax - nMin + 1 );

			// TODO: return proper value for column from joind table
			return szopdRow [ nColumnQueryIndex ];
		}
		catch ( opException opE )
		{
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::getValue: " + opE. GetErrorMsg ( ) +
								"\nFailed to fetch row data." );
		}
	} // end of Object fetchValue ( int nCurrentRow, int nColumnQueryIndex ) throws SQLException
	
	/**
	*	Fetches  and caches block of rows for joined tables. Assumes that data for the main FROM table is
	*	already fetched. This method assumes that a call to registerTables ( ) was
	*	already made and the unregisterTables ( ) will be called afterwords.
	*	
	*	int nFirstRowIndex	- 1-based row index of the first row in the block in the main table.
	*	int nBlockSize		- the amount of rows to fetch.
	*/
	private void fetchJoinedValues (
							int nFirstRowIndex,
							int nBlockSize ) throws SQLException
	{
		Vector				vctrFetchedTables = new Vector ( ); // Names of the tables which data is already fetched
		String				strFROMTableName = m_sqltDBMD. getDBName ( );
		vctrFetchedTables. addElement ( strFROMTableName. toUpperCase ( ) ); // FROM table

		Vector					vctrJOINTables = new Vector ( );
		SQLTableMetaData [ ]	szsqltMD = m_sqlCommand. getJOINTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
			vctrJOINTables. addElement ( szsqltMD [ i ] );
		
		SQLTableMetaData			sqltMD = null;
		SQLTableMetaData			sqltMDJoinedTo = null;
		while ( vctrJOINTables. size ( ) > 0 ) // SQL command was validated earlier and so we do not have to worry about endless loop
		{
			for ( int i = 0; i < vctrJOINTables. size ( ); i++ )
			{
				sqltMD = ( SQLTableMetaData ) vctrJOINTables. elementAt ( i );
				sqltMDJoinedTo = m_sqlCommand. getTable ( sqltMD. getTableAliasJoinedTo ( strFROMTableName ) );
				if ( !vctrFetchedTables. contains ( sqltMDJoinedTo. getDBName ( ). toUpperCase ( ) ) )
					continue;
					
// System. out. println ( "Should fetch data for joined table " + sqltMD. getDBName ( ) );
				fetchJoinedValues (
								sqltMD,
								sqltMDJoinedTo,
								nFirstRowIndex,
								nBlockSize );
							
				vctrFetchedTables. addElement ( sqltMD. getDBName ( ). toUpperCase ( ) );
				vctrJOINTables. removeElement ( sqltMD );
				
				break;
			}
		} // end of while ( vctrJOINTables. size ( ) > 0 )
	}
	
	/**
	*	Fetches and caches block of rows for joined table sqltMD. It assumes that all necessary
	*	rows are already fetched for all tables that sqltMD is joined to.
	*	This method assumes that a call to registerTables ( ) was
	*	already made and the unregisterTables ( ) will be called afterwords.
	*
	*	sqltMDJoinedTo		- a table, to which sqltMD is joined directly.
	*	int nFirstRowIndex	- 1-based index of the first row in the block of the main FROM table
	*	int nBlockSize		- amount of rows to fetch
	*/
	private void fetchJoinedValues (
							SQLTableMetaData sqltMD,
							SQLTableMetaData sqltMDJoinedTo,
							int nFirstRowIndex,
							int nBlockSize ) throws SQLException
	{
		if ( sqltMD == null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::fetchJoinedValues: " +
								"Table can not be null");
								
		/*
				1. Extract first join condition.
				2. Extract column values from table sqltMDJoinedTo for a given block of rows for join condition.
				3. Restore server side selection for sqltMD to its original state (AllRecords).
				4. Reduce selection in sqltMD to contain only values from join condition.
				5. Fetch the whole selection for sqltMD.
				6. For each row in the block locate and cache the joined data using full join condition.
		*/
		
		// 1.
		JoinCondition [ ]		szjConditions = sqltMD. getJOINConditions ( );
		// Name of the first column we are joined to
		String				strJoinedToColumnName = szjConditions [ 0 ]. getLeftColumn ( ). getDBName ( );
		// Name of the colunm being joined
		String				strColumnNameBeingJoined = szjConditions [ 0 ]. getRightColumn ( ). getDBName ( );
		
		/*
			Now, using table name sqltMDJoinedTo. getDBName ( ) and column name strJoinedToColumnName
			we need to extract already prefetched values for a block of rows.
			NOTE: some of the elements of szopdJoinValues array may be null.
		*/
		// 2.
		opData [ ]				szopdJoinValues = getopValues (
															sqltMDJoinedTo. getDBName ( ),
															strJoinedToColumnName,
															nFirstRowIndex,
															nBlockSize );
		
		// Lets prepare for 4.
		Vector					vctrSearch = new Vector ( );
		SQLTableMetaData		sqltDBMD = m_dbMD. getTable ( sqltMD. getDBName ( ) );
		opSearch				opS = null;
		byte					btLogicalOperator = opConstants. NONE;
		int						nTableNumber = sqltDBMD. getNumber ( );
		int						nColumnNumber = sqltDBMD. getColumnNumber ( strColumnNameBeingJoined );
		for ( int i = 0; i < szopdJoinValues. length; i++ )
		{
			if ( szopdJoinValues [ i ] == null )
				continue;
				
			opS = new opSearch (
							btLogicalOperator,
							nTableNumber,
							nColumnNumber,
							opConstants. EQUAL,
							szopdJoinValues [ i ] );
			vctrSearch. addElement ( opS );
			
			btLogicalOperator = opConstants. OR;
		}
		
		opFieldArray		opfaHelper = ( opFieldArray ) m_htFieldArrayHelpers. get (
																			sqltMD. getDBName ( ). toUpperCase ( ) );
		int [ ]				sznColumnIndexes = ( int [ ] ) m_htIndexArrayHelpers. get (
																			sqltMD. getDBName ( ). toUpperCase ( ) );
		String [ ]			szstrColumnNames = ( String [ ] ) m_htJOINedColumnNamesHelper. get (
																			sqltMD. getDBName ( ). toUpperCase ( ) );
		Hashtable			htValues = ( Hashtable ) m_htJOINedValues. get (
																			sqltMD. getDBName ( ). toUpperCase ( ) );
																			
		if ( vctrSearch. size ( ) == 0 )
		{
			/*
				All opData elements are null and thus there is nothing to join to.
				All I need to do is cache null values.
			*/
			opDataArray [ ]			szopdaNulls = createRowBlock (
																opfaHelper. mFieldArray. length,
																nBlockSize,
																true,
																opfaHelper );
			cacheRows (
					nFirstRowIndex,
					szopdaNulls,
					sznColumnIndexes,
					szstrColumnNames,
					htValues );
			
			return;
		}
		
		opSearchArray		opsArray = new opSearchArray ( vctrSearch. size ( ) );
		for ( int i = 0; i < vctrSearch. size ( ); i++ )
			opsArray. mSearchArray [ i ] = ( opSearch ) vctrSearch. elementAt ( i );
		
		// 3. & 4. & 5.
		opDataArray [ ]			szopdaValuesBeingJoined = null;
		synchronized ( m_objDataAccessLock )
		{
			try
			{
				// AllRecords is not necessary because Search should use the whole table
				// 3.
				// m_opProcess. AllRecords ( new opTable ( nTableNumber ) );
				
				// 4. TODO: find out how many search conditions I may pass at once. May be I'll have to fetch rows in blocks?
				opSelection			opsSearch = m_opProcess. Search ( opsArray );
				if ( opsSearch. mRecordsInSelection == 0 )
				{
					// There is nothing to join. Cache nulls.
					opDataArray [ ]			szopdaNulls = createRowBlock (
																		opfaHelper. mFieldArray. length,
																		nBlockSize,
																		true,
																		opfaHelper );
					cacheRows (
							nFirstRowIndex,
							szopdaNulls,
							sznColumnIndexes,
							szstrColumnNames,
							htValues );

					return;
				}
				
				// 5.
				szopdaValuesBeingJoined = createRowBlock (
													opfaHelper. mFieldArray. length,
													opsSearch. mRecordsInSelection,
													false,
													opfaHelper );
				// Lets set types of the data retrieved.
				
				
				m_opProcess. SelectionToArray (
											szopdaValuesBeingJoined,
											opfaHelper );
			}
			catch ( opException opE )
			{
				throw new SQLException (
									"com.fourd.jdbc.RSDataCache::fetchJoinedValues:" +
									"Failed to select joined data.\n" +
									opE. GetErrorMsg ( ) );
			}
		} // end of synchronized ( m_objDataAccessLock )
		
		// Lets prepare for 6.
		// First of all, lets prepare a block of rows with already prefetched values to which I'm joining.
		// Also, lets find indexes of the columns that are being joined in just fetched block of rows.
		
		// szopdValuesJoinedTo [ i ] contains opData [ ] that represents values for column #i
		opData [ ] [ ]		szopdValuesJoinedTo = new opData [ szjConditions. length ] [ ];
		
		// sznIndexesBeingJoined [ i ] contains index of the column in szopdaValuesBeingJoined joined to szopdValuesJoinedTo [ i ]
		int [ ]				sznIndexesBeingJoined = new int [ szjConditions. length ];
		
		for ( int i = 0; i < szjConditions. length; i++ )
		{
			szopdValuesJoinedTo [ i ] = getopValues (
												sqltMDJoinedTo. getDBName ( ),
												szjConditions [ i ]. getLeftColumn ( ). getDBName ( ),
												nFirstRowIndex,
												nBlockSize );
			sznIndexesBeingJoined [ i ] = getFetchIndex (
													sqltMD.getDBName ( ),
													szjConditions [ i ]. getRightColumn ( ). getDBName ( ) );
		}
		
		joinValues (
				szopdValuesJoinedTo,
				szopdaValuesBeingJoined,
				sznIndexesBeingJoined,
				nFirstRowIndex,
				sqltMD );
	} // end of private void fetchJoinedValues (
	
	/**
	*	Joins and caches row block from one table to the row block from another.
	*
	*	szopdValuesJoinedTo			- szopdValuesJoinedTo [ i ] contains opData [ ] that represents values for column #i
	*	szopdaValuesBeingJoined		- values being joined to szopdValuesJoinedTo
	*	sznIndexesBeingJoined		- columns in szopdaValuesBeingJoined may be in different order than in szopdValuesJoinedTo,
	*									sznIndexesBeingJoined contains indexes of columns in szopdaValuesBeingJoined that
	*									correspond to colunms in szopdValuesJoinedTo
	*/
	private void joinValues (
							opData [ ] [ ] szopdValuesJoinedTo,
							opDataArray [ ] szopdaValuesBeingJoined,
							int [ ] sznIndexesBeingJoined,
							int nFirstRowIndex,
							SQLTableMetaData sqltMDBeingJoined ) throws SQLException
	{
		// For each row in szopdValuesJoinedTo lets find and join values from szopdaValuesBeingJoined
		
		int					nColumnCountJoinedTo = szopdValuesJoinedTo. length;
		int					nRowCount = szopdValuesJoinedTo [ 0 ]. length;
		opData [ ]			szopDataRowJoinedTo = new opData [ nColumnCountJoinedTo ]; // we will reuse this array for each row
		
		// Final resulting joined values. This is what we will cache.
		opDataArray [ ]		szopdaResult = new opDataArray [ szopdaValuesBeingJoined. length ];
		for ( int i = 0; i < szopdaResult. length; i++ )
			szopdaResult [ i ] = new opDataArray ( nRowCount );

		opData [ ]			szopdRowJoined = null;
		int					nColumnCountBeingJoined = szopdaValuesBeingJoined. length;
		int					j = 0;
		for ( int i = 0; i < nRowCount; i++ )
		{
			// Lets prepare row #i
			for ( j = 0; j < nColumnCountJoinedTo; j++ )
				szopDataRowJoinedTo [ j ] = szopdValuesJoinedTo [ j ] [ i ];
				
			szopdRowJoined = joinRow (
									szopDataRowJoinedTo,
									szopdaValuesBeingJoined,
									sznIndexesBeingJoined );
				
			for ( j = 0; j < nColumnCountBeingJoined; j++ )
				szopdaResult [ j ]. mDataArray [ i ] = szopdRowJoined [ j ];
		}
		
		String				strTableName = sqltMDBeingJoined. getDBName ( ). toUpperCase ( );
		int [ ]				sznColumnIndexes = ( int [ ] ) m_htIndexArrayHelpers. get ( strTableName );
		String [ ]			szstrJOINedColumnNames = ( String [ ] ) m_htJOINedColumnNamesHelper. get ( strTableName );
		Hashtable			htValues = ( Hashtable ) m_htJOINedValues. get ( strTableName );
		cacheRows (
				nFirstRowIndex,
				szopdaResult,
				sznColumnIndexes,
				szstrJOINedColumnNames,
				htValues );
	} // end of private void joinValues ( ...
	
	/**
	*	Joins one row from one table to the row in another table. This method does not cache joined row,
	*	it just retuns it as a result.
	*
	*/
	private opData [ ] joinRow (
						opData [ ] szopDataRowJoinedTo,
						opDataArray [ ] szopdaValuesBeingJoined,
						int [ ] sznIndexesBeingJoined ) throws SQLException
	{
		int				nColumnCountBeingJoined = szopdaValuesBeingJoined. length;
		opData [ ]		szopdResult = new opData [ nColumnCountBeingJoined ];
		if ( szopDataRowJoinedTo [ 0 ] == null )
		{
			/*
				Row being joined to contains null data and thus there is nothing to join to.
				Lets return nulls.
			*/
			for ( int i = 0; i < nColumnCountBeingJoined; i++ )
				szopdResult [ i ] = null;
				
			return szopdResult;
		}
		
		// Lets try to find matching row for szopDataRowJoinedTo in szopdaValuesBeingJoined
		int				nRowCount = szopdaValuesBeingJoined [ 0 ]. mDataArray. length;
		int				nColumnCountJoinedTo = szopDataRowJoinedTo. length;
		int				j = 0;
		opData [ ]		szopdRowBeingJoined = new opData [ nColumnCountJoinedTo ]; // will reuse this array for every row
		for ( int i = 0; i < nRowCount; i++ )
		{
			// Lets prepare row #i and then compare it to szopDataRowJoinedTo
			for ( j = 0; j < nColumnCountJoinedTo; j++ )
				szopdRowBeingJoined [ j ] = szopdaValuesBeingJoined [ sznIndexesBeingJoined [ j ] ]. mDataArray [ i ];
				
			if ( !rowsAreEqual ( szopDataRowJoinedTo, szopdRowBeingJoined ) )
				continue;
				
			// Found the row!
			for ( j = 0; j < nColumnCountBeingJoined; j++ )
				szopdResult [ j ] = szopdaValuesBeingJoined [ j ]. mDataArray [ i ];
				
			return szopdResult;
		}
		
		// We did not find the row. Lets return nulls.
		for ( int i = 0; i < nColumnCountBeingJoined; i++ )
			szopdResult [ i ] = null;
			
		return szopdResult;
	} // end of private void joinRow (
	
	/**
	*	Returns true if rows contain the same elements, false - otherwise.
	*/
	private boolean rowsAreEqual ( opData [ ] szopdLeft, opData [ ] szopdRight ) throws SQLException
	{
		for ( int i = 0; i < szopdLeft. length; i++ )
			if ( !opDataAreEqual ( szopdLeft [ i ], szopdRight [ i ] ) )
				return false;
				
		return true;
	}
	
	private boolean opDataAreEqual ( opData opdLeft, opData opdRight ) throws SQLException
	{
		if ( opdLeft == null || opdRight == null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::opDataAreEqual: " +
								"Elements compared can not be null." );
								
		if ( opdLeft. mType != opdRight. mType )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::opDataAreEqual: " +
								"Elements compared can not be of different types. " +
								"Left type is " + opdLeft. mType + " right type is " + opdRight. mType );
								
		if ( opdLeft. mType == opConstants. ALPHANUMERIC )
			return opdLeft. mString. toUpperCase ( ). equals ( opdRight. mString. toUpperCase ( ) );
		else if ( opdLeft. mType == opConstants. BOOLEAN )
			return opdLeft. mBoolean == opdRight. mBoolean;
		else if ( opdLeft. mType == opConstants. INTEGER )
			return opdLeft. mInteger == opdRight. mInteger;
		else if ( opdLeft. mType == opConstants. LONGINTEGER )
			return opdLeft. mLongInteger == opdRight. mLongInteger;
		else if ( opdLeft. mType == opConstants. REAL )
			return opdLeft. mReal == opdRight. mReal;
		else if ( opdLeft. mType == opConstants. TEXT )
			return opdLeft. mString. toUpperCase ( ). equals ( opdRight. mString. toUpperCase ( ) );
		else if ( opdLeft. mType == opConstants. TIME )
			return opdLeft. mTime. getTime ( ). getTime ( ) == opdRight. mTime. getTime ( ). getTime ( );
		else if ( opdLeft. mType == opConstants. DATE )
			return opdLeft. mDate. getTime ( ). getTime ( ) == opdRight. mDate. getTime ( ). getTime ( );
		else
			// TODO: implement comparison for all other types.
			return false;
	}
	
	private opData [ ] getopValues (
							String strTableName,
							String strColumnName,
							int nFirstRowIndex,
							int nBlockSize ) throws SQLException
	{
		opData [ ]		opdaResult = new opData [ nBlockSize ];
		/*
			First of all, I need to figure out where the column data is stored - in the SELECT cache or in JOIN cache.
			In other words, is strColumnName used in SELECT clause or not.
		*/
		int				nColumnIndex = -1;
		Hashtable		htCache = null;
		if ( columnIsJOINedOnly (
						strTableName,
						strColumnName ) )
		{
			nColumnIndex = getJOINedOnlyColumnIndex (
												strTableName,
												strColumnName );
			// Column's data is in the JOINed cache. Lets extract table's cache.
			htCache = ( Hashtable ) m_htJOINedValues. get ( strTableName. toUpperCase ( ) );
		}
		else
		{
			// Column's data is in the main SELECT cache
			nColumnIndex = m_sqlCommand. getSELECTColumnIndex ( strTableName, strColumnName );
			htCache = m_htData;
		}
		
		opData [ ]		szopData = null;
		for ( int i = 0; i < nBlockSize; i++ )
		{
			szopData = ( opData [ ] ) htCache. get ( String. valueOf ( nFirstRowIndex + i ) );
			opdaResult [ i ] = szopData [ nColumnIndex ];
		}
		
		return opdaResult;
	}
	
	/**
	*	Returns true if a given column is not used in SELECT clause but only in JOIN.
	*/
	private boolean columnIsJOINedOnly (
								String strTableName,
								String strColumnName )
	{
		String [ ]		szstrColumnNames = ( String [ ] ) m_htJOINedColumnNamesHelper. get (
																strTableName. toUpperCase ( ) );
		for ( int i = 0; i < szstrColumnNames. length; i++ )
			if ( szstrColumnNames [ i ]. toUpperCase ( ). equals (
													strColumnName. toUpperCase ( ) ) )
				return true;
				
		return false;
	}
	
	private int getJOINedOnlyColumnIndex (
								String strTableName,
								String strColumnName ) throws SQLException
	{
		String [ ]		szstrColumnNames = ( String [ ] ) m_htJOINedColumnNamesHelper. get (
																strTableName. toUpperCase ( ) );
		for ( int i = 0; i < szstrColumnNames. length; i++ )
			if ( szstrColumnNames [ i ]. toUpperCase ( ). equals (
													strColumnName. toUpperCase ( ) ) )
				return i;
				
		throw new SQLException (
							"com.fourd.jdbc.RSDataCache::getJOINedOnlyColumnIndex: " +
							"Column not found" );
	}
	
	/**
	*	Data for a table is always fetched in the same order;
	*	returns the index of a given column in the fetched block of rows.
	*/
	private int getFetchIndex ( String strTableName, String strColumnName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::getFetchIndex: " +
								"Table name can not be null");
		if ( strColumnName == null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::getFetchIndex: " +
								"Column name can not be null");
								
		int				nColumnNumber = m_dbMD. getTable ( strTableName ). getColumnNumber ( strColumnName );
		opFieldArray	opfaHelper = ( opFieldArray ) m_htFieldArrayHelpers. get ( strTableName. toUpperCase ( ) );
		int				nColumnCount = opfaHelper. mFieldArray. length;
		for ( int i = 0; i < nColumnCount; i++ )
			if ( opfaHelper. mFieldArray [ i ]. mFieldNumber == nColumnNumber )
				return i;
				
		throw new SQLException (
							"com.fourd.jdbc.RSDataCache::getFetchIndex: " +
							"Column " + strTableName + "." + strColumnName +
							" was not found");
	}
	
	/**
	*	Releases all data in the cache by setting it to null.
	*/
	void clear ( )
	{
		/*
			TODO: test and see how effective this is. Should I, maybe, call System. gc ( ); ? Or, should I
			set each array that represents one row to null before emptying hashtable?
		*/
		m_htData. clear ( );
		
		Enumeration			enum = m_htJOINedValues. elements ( );
		while ( enum. hasMoreElements ( ) )
			( ( Hashtable ) enum. nextElement ( ) ). clear ( );
	}
	
	/**
	*	Caches a sequential blcok of rows represented by szopdaRows. The index of the first row is nFirstRowIndex
	*	nFirstRowIndex is 1-based here.
	*	szopdaRows [ i ] contains values for column #i
	*
	*	Returns a reference to the first cached row in the block.
	*/
	/*private Object [ ] cacheRows ( int nFirstRowIndex, opDataArray [ ] szopdaRows ) throws SQLException
	{
		int				nRowCount = szopdaRows [ 0 ]. mDataArray. length;
		int				nColumnCount = szopdaRows. length;
		Object [ ]		szobjRow = null;
		Object [ ]		szobjResult = null;
		for ( int i = 0; i < nRowCount; i++ )
		{	
			szobjRow = new Object [ nColumnCount ];
			for ( int j = 0; j < nColumnCount; j++ )
				szobjRow [ j ] = convertToJavaObject ( szopdaRows [ j ]. mDataArray [ i ] );

			if ( szobjResult == null )
				szobjResult = szobjRow;
				
			m_htData. put ( String. valueOf ( nFirstRowIndex + i ), szobjRow );
		}
		
		return szobjResult;
	}
	*/
	
	/**
	*	Caches a sequential block of rows represented by szopdaRows. The index of the first row is nFirstRowIndex
	*	nFirstRowIndex is 1-based here.
	*	szopdaRows [ i ] contains values for column # sznColumnIndexes [ i ]
	*		(1) Last several elements of each row in szopdaRows may contain values for columns that are not used in SELECT 
	*		clause but rather in JOIN clauses. We will cache those values separately.
	*
	*	Returns a reference to the first cached row in the block.
	*/
	private opData [ ] cacheRows (
							int nFirstRowIndex,
							opDataArray [ ] szopdaRows,
							int [ ] sznColumnIndexes,
							String [ ] szstrJOINedColumnNames,
							Hashtable htValues ) throws SQLException
	{
		int				nRowCount = szopdaRows [ 0 ]. mDataArray. length;
		int				nColumnCount = sznColumnIndexes. length; // NOT szopdaRows. length; because see comment (1)
		int				nSELECTColumnCount = m_sqlCommand. getSELECTColumnCount ( );
		opData [ ]		szopdRow = null;
		opData [ ]		szopdResult = null;
		boolean			bRowExisted = false;
		for ( int i = 0; i < nRowCount; i++ )
		{
			szopdRow = ( opData [ ] ) m_htData. get ( String. valueOf ( nFirstRowIndex + i ) );
			if ( szopdRow == null )
			{
				bRowExisted = false;
				szopdRow = new opData [ nSELECTColumnCount ];
				for ( int k = 0; k < nSELECTColumnCount; k++ )
					szopdRow [ k ] = null;
			}
			else
				bRowExisted = true;
				
			for ( int j = 0; j < nColumnCount; j++ )
				szopdRow [ sznColumnIndexes [ j ] ] = szopdaRows [ j ]. mDataArray [ i ];

			if ( szopdResult == null )
				szopdResult = szopdRow;
				
			if ( !bRowExisted )
				m_htData. put ( String. valueOf ( nFirstRowIndex + i ), szopdRow );
		} // end of for ( int i = 0; i < nRowCount; i++ )
		
		cacheJoinedValues (
					nFirstRowIndex,
					szopdaRows,
					szstrJOINedColumnNames,
					htValues );
							
		return szopdResult;
	}
	
	/**
	*	Caches values for those columns that are not used in SELECT clause but rather in JOIN clause.
	*/
	private void cacheJoinedValues (
							int nFirstRowIndex,
							opDataArray [ ] szopdaRows,
							String [ ] szstrJOINedColumnNames,
							Hashtable htValues ) throws SQLException
	{
		if ( szstrJOINedColumnNames. length == 0 )
			// There are no JOINs columns to cache
			return;
			
		// Lets deal with data used for JOINs only and not for SELECT

		int				nRowCount = szopdaRows [ 0 ]. mDataArray. length;
		int				nOffset = szopdaRows. length - szstrJOINedColumnNames. length; // index of the first JOINs value
		int				nTotalColumnCount = szopdaRows. length;
		int				nColumnCount = szstrJOINedColumnNames. length;
		opData [ ]		szopData = null;
		for ( int i = 0; i < nRowCount; i++ )
		{
			szopData = new opData [ nColumnCount ];
			for ( int j = nOffset; j < nTotalColumnCount; j++ )
				szopData [ j - nOffset ] = szopdaRows [ j ]. mDataArray [ i ];
			
			/*
				TODO: should I deal with the case when particular row is already cached?
				Currently I replace cached row with the new one.
			*/
			htValues. put ( String. valueOf ( nFirstRowIndex + i ), szopData );
		}
	} // end of private void cacheJoinedValues (
	
	private Object convertToJavaObject ( opData opD ) throws SQLException
	{
		if ( opD == null )
			return null;
			
		// TODO: possible location for memory optimization
		if ( opD. mType == opConstants. ALPHANUMERIC )
			return new String ( opD. mString );
		else if ( opD. mType == opConstants. BOOLEAN )
			return new Boolean ( opD. mBoolean );
		else if ( opD. mType == opConstants. INTEGER )
			return new Short ( opD. mInteger );
		else if ( opD. mType == opConstants. LONGINTEGER )
			return new Integer ( opD. mLongInteger );
		else if ( opD. mType == opConstants. REAL )
			return new Double ( opD. mReal );
		else if ( opD. mType == opConstants. TEXT )
			return new String ( opD. mString );
		else if ( opD. mType == opConstants. TIME )
			return new java.sql.Time ( opD. mTime. getTime ( ). getTime ( ) );
		else if ( opD. mType == opConstants. DATE )
		{
			if ( opD. mDate == null )
				return null;
			
			return new java.sql.Date ( opD. mDate. getTime ( ). getTime ( ) );
		}
		else if ( opD. mType == opConstants. BLOB )
			if ( opD. mBlob. length == opD. mBlobSize )
				return opD. mBlob;
			else
			{
				byte [ ]		szbtResult = new byte [ opD. mBlobSize ];
				System. arraycopy (
								opD. mBlob, 0,
								szbtResult, 0,
								opD. mBlobSize );
									
				return szbtResult;
			}
		else if ( opD. mType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - PICTURE" );
		else if ( opD. mType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( opD. mType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
							"Unknown data type - " + opD. mType );
	}
	
	/**
	*	Creates a set of rows in 4D Open style:
	*	Each element of opDataArray [ ] represents a column. Thus, 
	*		opDataArray. length == nColumnCount
	*		opDataArray. mDataArray [ i ]. length == nRowCount
	*
	*	Each "field" value in returned array is set to null if bSetToNull is true:
	*		opDataArray [ i ]. mDataArray [ j ] == null
	*
	*	opfaHelper is used to set types of data objects
	*
	*	This is just a helper method.
	*/
	private opDataArray [ ] createRowBlock (
									int nColumnCount,
									int nRowCount,
									boolean bSetToNull,
									opFieldArray opfaHelper ) throws SQLException
	{
		if ( nColumnCount <= 0 )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::createEmptyRowBlock:" +
								"There should be at least one column" );
		if ( nRowCount <= 0 )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::createEmptyRowBlock:" +
								"There should be at least one row" );
								
		opDataArray [ ]		szopdaResult = new opDataArray [ nColumnCount ];
		for ( int i = 0; i < nColumnCount; i++ )
		{
			szopdaResult [ i ] = new opDataArray ( nRowCount );
			
			for ( int j = 0; j < nRowCount; j++ )
			{
				szopdaResult [ i ]. mDataArray [ j ]. mType = ( short )  opfaHelper.
																				mFieldArray [ i ].
																						mFieldType;
				if ( bSetToNull )
					szopdaResult [ i ]. mDataArray [ j ] = null;
			}
		}
		
		return szopdaResult;
	}
	
	/**
	*	nRow is 1-based
	*/
	void update ( int nRow, RowCache rCache ) throws SQLException
	{
		if ( rCache == null )
			throw new SQLException ( "com.fourd.jdbc.RSDataCache::update: Row cache can not be null." );
			
		opData [ ]			szopdRow = ( opData [ ] ) m_htData. get ( String. valueOf ( nRow ) );
		if ( szopdRow == null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::update: " +
								"Can not update, row is not cached." );
		/**
		*	For each column in SELECT clause, try to extract its updated value from
		*	the RowCache object. If there is an updated value then find the original value
		*	in the cache and replace it with the updated one.
		*/
		
		int						nSELECTCColumnCount = m_sqlCommand. getSELECTColumnCount ( );
		Object					objValue = null;
		SQLColumnMetaData		sqlcMD = null;
		String					strTableName = null;
		SQLTableMetaData		sqltMD = null;
		for ( int i = 0; i < nSELECTCColumnCount; i++ )
		{
			if ( !rCache. columnIsCached ( i ) )
				continue; // value for this column is not cached, thus nothing to update
				
			objValue = rCache. get ( i );
			sqlcMD = m_sqlCommand. getSELECTColumn ( i ); // column meta data from query
			strTableName = sqlcMD. getDBTableName ( ); // name or alias of a table in a query
			if ( strTableName == null ) // main FROM table
				strTableName = m_sqltDBMD. getDBName ( );
			sqltMD = m_sqlCommand. getTable ( strTableName ); // table meta data from query
			sqltMD = m_dbMD. getTable ( sqltMD. getDBName ( ) ); // table meta data from DB
			sqlcMD = sqltMD. getColumnMetaData ( sqlcMD. getDBName ( ) ); // column meta data from DB
			szopdRow [ i ] = SQLTools. createopData (
													objValue,
													sqlcMD. getopType ( ) );
		}
	} // end of void update ( int nRow, RowCache rCache ) throws SQLException
	
	/**
	*	nRow is 1-based
	*/
	void insert ( int nRow, RowCache rCache ) throws SQLException
	{
		if ( rCache == null )
			throw new SQLException ( "com.fourd.jdbc.RSDataCache::insert: Row cache can not be null." );
			
		opData [ ]			szopdRow = ( opData [ ] ) m_htData. get ( String. valueOf ( nRow ) );
		if ( szopdRow != null )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::insert: " +
								"Duplicate row number, can not insert row." );
								
		int						nSELECTCColumnCount = m_sqlCommand. getSELECTColumnCount ( );
		szopdRow = new opData [ nSELECTCColumnCount ];
		Object					objValue = null;
		SQLColumnMetaData		sqlcMD = null;
		String					strTableName = null;
		SQLTableMetaData		sqltMD = null;
		for ( int i = 0; i < nSELECTCColumnCount; i++ )
		{
			if ( !rCache. columnIsCached ( i ) )
			{
				szopdRow [ i ] = null;
				
				continue; // no value for this column
			}
				
			objValue = rCache. get ( i );
			sqlcMD = m_sqlCommand. getSELECTColumn ( i ); // column meta data from query
			strTableName = sqlcMD. getDBTableName ( ); // name or alias of a table in a query
			if ( strTableName == null ) // main FROM table
				strTableName = m_sqltDBMD. getDBName ( );
			sqltMD = m_sqlCommand. getTable ( strTableName ); // table meta data from query
			sqltMD = m_dbMD. getTable ( sqltMD. getDBName ( ) ); // table meta data from DB
			sqlcMD = sqltMD. getColumnMetaData ( sqlcMD. getDBName ( ) ); // column meta data from DB
			szopdRow [ i ] = SQLTools. createopData (
													objValue,
													sqlcMD. getopType ( ) );
		}
		
		m_htData. put ( String. valueOf ( nRow ), szopdRow );
	} // void insert ( int nRow, RowCache rCache ) throws SQLException
}