/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

/**
*	TODO: break this class into hierarchy of smaller classes. 
*	One class is responsible for navigation, other for getXXX, other 
*	for updateXXX, other for my custom methods, etc...
*/

package com.fourd.jdbc;

import com.fourd.jdbc.rs.*;
	
import java.sql.*;
import java.util.*;
import java.io.*;
import java.math.*;
import java.net.*;
	
import j4d.open.*;
	
class ResultSetImpl extends RSNavigator
{
	// A reference to the connection that generated instance of this class
	private ConnectionImpl			m_connectionImpl = null;

	/*
		SQLCommand instance that produced this ResultSet
	*/
	private SQLCommand 			m_sqlCommand = null;
	
	/*
		Meta data about a table whose data is represented in the instance if this class.
		Later I may need to use array of SQLTableMetaData to handle joins.
		This is a DB version of meta data, not the one produced while parsing the query.
		
		Right now this variable represents a single FROM table.
	*/
	private SQLTableMetaData		m_sqltDBMD = null;
	/**
		Amount of rows in the current selection of the "main" table. 
		"Main" table is a table with which all others are joined. Later, I may need
		to use an array of ints to store amount of rows in each selection for each
		joined table.
	*/
	private int						m_nRowCount = 0;
	
	/**
	*	Contains current amount of original rows in this resultset.
	*	Original rows are rows that were present when result set was created.
	*	m_nOriginalRowCount can only get smaller, when rows are deleted.
	*	It can't get more than its initial value.
	*	New rows are always created at the end of the result set. Thus,
	*	this variable helps to distinguish between old original and 
	*	new rows. I need to distinguish them because they are deleted and
	*	updated differently.
	*/
	private int						m_nOriginalRowCount = 0;
	
	/**
	*	A reference to the statement that generated this result set.
	*/
	private StatementImpl			m_statementImpl = null;
	
	/**
	*	Represents selection in the main table. I may need to use array of 
	*	opSelection to support joins.
	*/
	private opSelection			m_opSelection = null;
	
	private opProcess				m_opProcess = null;
	
	/**
	*	Actual ResultSet's data. Currently it contains data for a single FROM table referenced by m_sqltDBMD
	*/
	private RSDataCache			m_rsDataCache = null;
	
	/**
		Should always synchronize on m_objDataAccessLock before accessing m_opProcess.
		Can't synchronize on m_opProcess itself.
	*/
	private Object						m_objDataAccessLock = null;
	
	private DatabaseMetaDataImpl		m_dbMD = null;
	
	private ResultSetMetaDataImpl		m_RSMDImpl = null;

	// Navigation handling
	private static final int			POSITION_BEFORE_FIRST = -1;
	private static final int			POSITION_AFTER_LAST = -2;
	private boolean						m_bCursorOnDeletedRow = false;
	private boolean						m_bCursorOnInsertedRow = false;
	// m_nCurrentRow is 1-based
	private int					m_nCurrentRow = POSITION_BEFORE_FIRST;
	
	/**
	*	Contains 1-based index of all deleted rows withing this resultset in respect
	*	to the row count at the moment of deletion. Lets say, original selection
	*	have three rows : #1, #2, and #3. I delete row #2 and thus m_vctrDeletedRows will
	*	contain String "2". Row #1 will remain the same and row #3 will become #2. Now,
	*	I again delete row #2. This time, I again will put string "2" into the 
	*	vector m_vctrDeletedRows even though row #2 was actually #3 in the original
	*	selection. Index of a deleted row is stored as String. Indexes are
	*	stored in the order they were deleted, this order is important.
	*/
	private Vector						m_vctrDeletedRows = new Vector ( );
	
	/**
	*	Contains string representation of 4d record numbers for new records. 
	*	Record numbers are stored in the same order that records are created.
	*/
	private Vector						m_vctrNewRows = new Vector ( );
	
	// Cache for new or updated row.
	private RowCache					m_rCache = null;
	
	/**
		"ResultSet Is Closed" flag. Once m_bClosed is set to true,
		all methods will throw "ResultSet is closed" exception.
	*/
	private boolean						m_bClosed = false;

	// Helpers.
	/**
	*	Each time we need to do something with opProcess we either need an opTable or opSelection reference.
	*	Instead of creating them each time we need them, we will just use one reference as a class member.
	*/
	private opTable					m_opTableHelper = new opTable ( );
	
	/**
	*	Names of the tables used in the query.
	*	Initialize this array once and reuse it each time I need to register or unregister them.
	*/
	private String [ ]				m_szstrTableNames = null;
	
	/**
	*	Rows are fetched not one by one but in blocks. The amount of rows in the block.
	*/
	private static final int				DEFAULT_FETCH_SIZE = 3000;
	private int						m_nFetchSize = DEFAULT_FETCH_SIZE;
	
	private int						m_nConcurrency = CONCUR_UPDATABLE; // default
	
	private FDExtensions			m_fdExtensions = null;
		
	ResultSetImpl (
			SQLCommand sqlCommand,
			SQLTableMetaData sqltDBMD,
			opSelection opSelection,
			StatementImpl statementImlp,
			opProcess opP,
			ConnectionImpl connectionImpl,
			Object objDataAccessLock,
			DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		m_sqlCommand = sqlCommand;
		m_sqltDBMD = sqltDBMD;
		m_opSelection = opSelection;
		m_nRowCount = m_opSelection. mRecordsInSelection;
		m_nOriginalRowCount = m_nRowCount;
		m_statementImpl = statementImlp;
		m_opProcess = opP;
		m_connectionImpl = connectionImpl;
		m_objDataAccessLock = objDataAccessLock;
		m_dbMD = dbMD;
		
		prepareHelpers ( );
		
		m_rsDataCache = new RSDataCache (
										m_sqlCommand,
										m_sqltDBMD,
										m_connectionImpl,
										m_statementImpl,
										this,
										m_opProcess,
										m_objDataAccessLock,
										m_dbMD );
										
		m_fdExtensions = new FDExtensions ( m_connectionImpl );
	}
	
	/**
	*	The only use of this constructor is to create read-only local ("disconnected")
	*	resultsets.
	*
	*	szopData [ i ] - row #i
	*	szopData [ i ] [ j ] - row #i, column #j
	*/
	ResultSetImpl ( SQLCommand sqlCommand, opData [ ] [ ] szopData )
	{
		m_sqlCommand = sqlCommand;
		m_nRowCount = szopData. length;
		m_rsDataCache = new RSDataCache ( szopData );
		m_nConcurrency = CONCUR_READ_ONLY;
	}

	/**
	*	Moves the cursor to the given row number in this ResultSet object.
	*/
	public boolean absolute ( int nRow) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		// Should work even if m_nRowCount == 0

		if ( nRow == 0 )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::absolute: Row index can not be 0" );
			
		resetAllEdits ( );
		
		if ( nRow > 0 )
		{
			if ( nRow > m_nRowCount )
			{
				// Place the cursor after the last row
				m_nCurrentRow = POSITION_AFTER_LAST;
				return false;
			}
			
			m_nCurrentRow = nRow;
		}
		else // nRow < 0
		{
			if ( nRow < ( -1 ) * m_nRowCount )
			{
				// Place the cursor before the first row
				m_nCurrentRow = POSITION_BEFORE_FIRST;
				return false;
			}
			
			m_nCurrentRow = m_nRowCount + nRow + 1; // which is m_nRowCount - ( -1 ) * nRow + 1
		}
		
		return true;
	}
	
	/**
	*	Moves the cursor to the end of this ResultSet object, just after the last row. 
	*/
	public void afterLast ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		m_nCurrentRow = POSITION_AFTER_LAST;
		resetAllEdits ( );
	}
	

	/**
	*	Moves the cursor to the front of this ResultSet object, just before the first row. 
	*/
	public void beforeFirst ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		m_nCurrentRow = POSITION_BEFORE_FIRST;
		resetAllEdits ( );
	}
	

	/**
	*	Cancels the updates made to the current row in this ResultSet object. 
	*/
	public void cancelRowUpdates ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );
			
		deleteRowCache ( );
	}
	

	/**
	*	Clears all warnings reported on this ResultSet object.
	*/
	public void clearWarnings ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		// TODO: Implement properly
		
		return;
	}
	

	/**
	*	Releases this ResultSet object's database and JDBC resources immediately instead of waiting for this to happen when it is automatically closed. 
	*/
	public void close ( ) throws SQLException
	{
		if ( m_bClosed )
			return;
			// throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		/*
			TODO:
					1. Clean the data cache
					2. Release current selection(s) by setting the amount of rows in it(them) to 0.
					3. Release all named selections, if there are any.
		*/
		
		m_rsDataCache. clear ( );
		
		if ( m_opProcess == null )
		{
			// Disconnected result set
			m_bClosed = true;
			return;
		}
		
		// TODO: opProcess. ClearNamedSelection
		
		// Lets reduce selection to 0 rows
		registerTables ( );
			synchronized ( m_objDataAccessLock )
			{
				try
				{
					m_opSelection. mRecordNumber = 0;
					m_opProcess. ReduceSelection ( m_opSelection );
				}
				catch ( opException opE )
				{
					unregisterTables ( );
					
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::close: " + opE. GetErrorMsg ( ) +
										"/nUnable to release selection.");
				}
			}
		unregisterTables ( );
		
		m_bClosed = true;
			
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Deletes the current row from this ResultSet object and from the underlying database. 
	*/
	public void deleteRow ( ) throws SQLException
	{
		/*
			Two choices - delete row completely or create a "hole", blank row.
			"Hole" is easy to implement but it does not look like a clean solution.
			I'd rather prefer to delete row completely. More work but looks much cleaner.
			The only problem is, which row should be made the current after the
			deletion of one? May be keep cursor at the deleted row? All read/write
			attempts would throw an exception until cursor is removed from the 
			deleted row. Once the cursor is moved - deleted row is no longer 
			accesible. To delete all rows one would have to write something like this:
				resultSet. first ( );
				while ( !resultSet. isAfterLast ( ) )
				{
					resultSet. deleteRow ( );
					resultSet. next ( );
				}
		*/
		/*
		*	TODO: Current implementation does not delete rows from joined tables.
		*
		*	It actually should not, although there my be cases when it is needed to 
		*	delete a row not only from the main FROM table but also from all JOINed
		*	tables. This should be handled via some kind of parameter passed to a 
		*	Connection object. It may also be handled on ResultSet level. Do it later.
		*/
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );
		
		if ( m_bCursorOnInsertedRow )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::deleteRow: " +
								"Cursor is on the insert row" );
			
		validateCurrentRow ( );
			
		/*
				1. Set deletion flag to true ( POSITION_AT_DELETED )
				2. Delete row from the database
				3. Delete row from cache
					3.1 Delete cached row
					3.2 Order number for all rows after deleted one should be
						decreased by 1. Currently all cached rows are stored in the
						hash table and such "re-ordering" may be slow. 
		*/
		
		// 1.
		if ( m_bCursorOnDeletedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::deleteRow: Row is already deleted" );
		
		// 2.
		/*
			Current selection DOES change when I delete a record from a table. It becomes empty,
			which is apparently a bug in 4D Open, as I was told.
			Thus, I need a brand new object to register tables with. This will make sure
			that whoever accesses changed table next (including this resultset) will
			have to update its current selection of all the tables used by this resultset.
		*/
		Object		objLock = new Object ( );
		registerTables ( objLock, false );
			synchronized ( m_objDataAccessLock )
			{
				try
				{
					if ( m_nCurrentRow <= m_nOriginalRowCount )
					{ // delete original row that was present when result set was created
						m_opSelection. mCurrentRecord = mapRowToSelection ( m_nCurrentRow ) - 1; // m_nCurrentRow - 1;
						m_opProcess. GotoSelectedRecord ( m_opSelection );
						/*
							A call to RecordsInSelection not only returns amount of rows in 
							a selection but also returns mRecordNumber for current row in selection.
							I need mRecordNumber in order to delete a row because this value is
							used in opProcess. DeleteRecord ( ). Unfortunately, mRecordNumber has
							nothing to do with row order number. mRecordNumber is just some number
							assigned to a row.
						*/
						m_opProcess. RecordsInSelection ( m_opSelection );
					
						m_opTableHelper. mRecordNumber = m_opSelection. mRecordNumber;
						m_opTableHelper. mTableNumber = m_sqltDBMD. getNumber ( );
						m_opProcess. DeleteRecord ( m_opTableHelper );
						m_nOriginalRowCount--;
					
						storeDeletedRowIndex ( m_nCurrentRow );
					}
					else
					{ // delete a new row that was created during this resultset lifetime
						/*
							TODO: optimization. There is no need to restore original selection for 
									this 'else' branch.
						*/
						/*
							This will make sure that the record I'm going to delete,
							is within the current selection and thus is available for deletion.
						*/
						m_opTableHelper. mTableNumber = m_sqltDBMD. getNumber ( );
						m_opProcess. AllRecords ( m_opTableHelper );
					
						int			nNewRecordIndex = m_nCurrentRow - m_nOriginalRowCount - 1;
						m_opTableHelper. mRecordNumber = Integer. valueOf (
															( String ) m_vctrNewRows. elementAt (
																nNewRecordIndex ) ). intValue ( );
						m_opProcess. GotoRecord ( m_opTableHelper ); // absolutely necessay
						m_opProcess. DeleteRecord ( m_opTableHelper );
						
						m_vctrNewRows. remove ( nNewRecordIndex );
					}
				}
				catch ( opException opE )
				{
					unregisterTables ( objLock );
					
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::deleteRow: " + opE. GetErrorMsg ( ) +
										"\nUnable to delete current row.");
				}
			}
		unregisterTables ( objLock );
		
		// 3.
		m_rsDataCache. deleteRow ( m_nCurrentRow );
		m_bCursorOnDeletedRow = true;
		m_nRowCount--;
	} // end of public void deleteRow ( ) throws SQLException

	/**
	*	Maps the given ResultSet column name to its ResultSet column index. 
	*/
	public int findColumn ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		return m_sqlCommand. getSELECTColumnIndex ( strColumnName ) + 1;
	}
	
	/**
	*	Moves the cursor to the first row in this ResultSet object. 
	*/
	public boolean first ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		resetAllEdits ( );
		
		if ( m_nRowCount == 0 )
			return false;
			
		m_nCurrentRow = 1;
		
		// throwUnimplementedException ( "" );
		return true;
	}
	
	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as an Array object in the Java programming language. 
	*/
	public Array getArray ( int i ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getArray" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as an Array object in the Java programming language. 
	*/
	public Array getArray ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getArray" );
		
		return null;
	}
	

	/**
	*	
	*	Gets the value of the designated column in the current row of this ResultSet object as a stream of ASCII characters. 
	*/
	public InputStream getAsciiStream ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getAsciiStream" );
		
		return null;
	}
	
	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a stream of ASCII characters. 
	*/
	public InputStream getAsciiStream ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getAsciiStream" );
		
		return null;
	}
	
	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.math.BigDecimal with full precision. 
	*/
	public BigDecimal getBigDecimal ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBigDecimal" );
		
		return null;
	}
	
	/**
	*	Deprecated.
	*/
	public BigDecimal getBigDecimal ( int nColumnIndex, int nScale) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBigDecimal" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.math.BigDecimal with full precision. 
	*/
	public BigDecimal getBigDecimal ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBigDecimal" );
		
		return null;
	}
	
	/**
	*	Deprecated.
	*/
	public BigDecimal getBigDecimal ( String strColumnName, int nScale) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBigDecimal" );
		
		return null;
	}
	

	/**
	*	Gets the value of a column in the current row as a stream of 
Gets the value of the designated column in the current row of this ResultSet object as a binary stream of uninterpreted bytes. 
	*/
	public InputStream getBinaryStream ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBinaryStream" );
		
		return null;
	}
	
	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a stream of uninterpreted byte s. 
	*/
	public InputStream getBinaryStream ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBinaryStream" );
		
		return null;
	}
	
	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Blob object in the Java programming language. 
	*/
	public Blob getBlob ( int i ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBlob" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Blob object in the Java programming language. 
	*/
	public Blob getBlob ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getBlob" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a boolean in the Java programming language. 
	*/
	public boolean getBoolean ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Boolean ) SQLTools. convert ( objValue, Boolean. class ) ). booleanValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a boolean in the Java programming language. 
	*/
	public boolean getBoolean ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Boolean ) SQLTools. convert ( objValue, Boolean. class ) ). booleanValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a byte in the Java programming language. 
	*/
	public byte getByte ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Byte ) SQLTools. convert ( objValue, Byte. class ) ). byteValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a byte in the Java programming language. 
	*/
	public byte getByte ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Byte ) SQLTools. convert ( objValue, Byte. class ) ). byteValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a byte array in the Java programming language. 
	*/
	public byte [ ] getBytes ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( byte [ ] ) SQLTools. convert ( objValue, byte [ ]. class );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a byte array in the Java programming language. 
	*/
	public byte [ ] getBytes ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( byte [ ] ) SQLTools. convert ( objValue, byte [ ]. class );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.io.Reader object. 
	*/
	public Reader getCharacterStream ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getCharacterStream" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.io.Reader object. 
	*/
	public Reader getCharacterStream ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getCharacterStream" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Clob object in the Java programming language. 
	*/
	public Clob getClob ( int i ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getClob" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Clob object in the Java programming language. 
	*/
	public Clob getClob ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getClob" );
		
		return null;
	}
	

	/**
	*	Returns the concurrency mode of this ResultSet object. 
	*/
	public int getConcurrency ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getConcurrency" );
		
		return -1;
	}
	

	/**
	*	Gets the name of the SQL cursor used by this ResultSet object. 
	*/
	public String getCursorName ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getCursorName" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Date object in the Java programming language. 
	*/
	public java.sql.Date getDate ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( java.sql.Date ) SQLTools. convert ( objValue, java.sql.Date. class );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Date object in the Java programming language. 
	*/
	public java.sql.Date getDate ( int nColumnIndex, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getDate" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Date object in the Java programming language. 
	*/
	public java.sql.Date getDate ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( java.sql.Date ) SQLTools. convert ( objValue, java.sql.Date. class );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Date object in the Java programming language. 
	*/
	public java.sql.Date getDate ( String strColumnName, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getDate" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a double in the Java programming language. 
	*/
	public double getDouble ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Double ) SQLTools. convert ( objValue, Double. class ) ). doubleValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a double in the Java programming language. 
	*/
	public double getDouble ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Double ) SQLTools. convert ( objValue, Double. class ) ). doubleValue ( );
	}
	

	/**
	*	Returns the fetch direction for this ResultSet object. 
	*/
	public int getFetchDirection ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getFetchDirection" );
		
		return -1;
	}
	

	/**
	*	Returns the fetch size for this ResultSet object. 
	*/
	public int getFetchSize ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		return m_nFetchSize;
		
		//throwUnimplementedException ( "" );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a float in the Java programming language. 
	*/
	public float getFloat ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Float ) SQLTools. convert ( objValue, Float. class ) ). floatValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a float in the Java programming language. 
	*/
	public float getFloat ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Float ) SQLTools. convert ( objValue, Float. class ) ). floatValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as an int in the Java programming language. 
	*/
	public int getInt ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Integer ) SQLTools. convert ( objValue, Integer. class ) ). intValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as an int in the Java programming language. 
	*/
	public int getInt ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Integer ) SQLTools. convert ( objValue, Integer. class ) ). intValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a long in the Java programming language. 
	*/
	public long getLong ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Long ) SQLTools. convert ( objValue, Long. class ) ). intValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a long in the Java programming language. 
	*/
	public long getLong ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Long ) SQLTools. convert ( objValue, Long. class ) ). intValue ( );
	}
	

	/**
	*	Retrieves the number, types and properties of this ResultSet object's columns. 
	*/
	public ResultSetMetaData getMetaData ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		if ( m_RSMDImpl == null )
			m_RSMDImpl = new ResultSetMetaDataImpl ( m_sqlCommand, m_dbMD );
			
		return m_RSMDImpl;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as an Object in the Java programming language. 
	*/
	public Object getObject ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateCurrentRow ( );
		
		if ( m_rCache != null )
			/*
				The current row is being updated but the update
				has not been commited yet or current row is insert row.
			*/
			return m_rCache. get ( nColumnIndex - 1 );
		else
			// return value from the cache (fetch it first, if necessary)
			return m_rsDataCache. getValue ( m_nCurrentRow, nColumnIndex - 1 );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as an Object in the Java programming language. 
	*/
	public Object getObject ( int i, Map map ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getObject" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as an Object in the Java programming language. 
	*/
	public Object getObject ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateCurrentRow ( );
		
		int		nColumnQueryIndex = m_sqlCommand. getSELECTColumnIndex ( strColumnName );
		
		if ( m_rCache != null )
			/*
				The current row is being updated but the update
				has not been commited yet or current row is insert row.
			*/
			return m_rCache. get ( nColumnQueryIndex );
		else
			return m_rsDataCache. getValue ( m_nCurrentRow, nColumnQueryIndex );

		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as an Object in the Java programming language. 
	*/
	public Object getObject ( String strColumnName, Map map ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getObject" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Ref object in the Java programming language. 
	*/
	public Ref getRef ( int i ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getRef" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a Ref object in the Java programming language. 
	*/
	public Ref getRef ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getRef" );
		
		return null;
	}
	

	/**
	*	Retrieves the current row number. 
	*/
	public int getRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		if (	m_nRowCount == 0 || m_bCursorOnDeletedRow || m_bCursorOnInsertedRow ||
				m_nCurrentRow == POSITION_BEFORE_FIRST ||
				m_nCurrentRow == POSITION_AFTER_LAST )
			return 0;
			
		// throwUnimplementedException ( "" );
		return m_nCurrentRow;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a short in the Java programming language. 
	*/
	public short getShort ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( ( Short ) SQLTools. convert ( objValue, Short. class ) ). shortValue ( );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a short in the Java programming language. 
	*/
	public short getShort ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( ( Short ) SQLTools. convert ( objValue, Short. class ) ). shortValue ( );
	}
	

	/**
	*	Returns the Statement object that produced this ResultSet object. 
	*/
	public Statement getStatement ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		return m_statementImpl;
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a String in the Java programming language. 
	*/
	public String getString ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( String ) SQLTools. convert ( objValue, String. class );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a String in the Java programming language. 
	*/
	public String getString ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( String ) SQLTools. convert ( objValue, String. class );
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Time object in the Java programming language. 
	*/
	public Time getTime ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( nColumnIndex );
		return ( Time ) SQLTools. convert ( objValue, Time. class );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Time object in the Java programming language. 
	*/
	public Time getTime ( int nColumnIndex, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTime" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Time object in the Java programming language. 
	*/
	public Time getTime ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		Object		objValue = getObject ( strColumnName );
		return ( Time ) SQLTools. convert ( objValue, Time. class );
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Time object in the Java programming language. 
	*/
	public Time getTime ( String strColumnName, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTime" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Timestamp object in the Java programming language. 
	*/
	public Timestamp getTimestamp ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTimestamp" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Timestamp object in the Java programming language. 
	*/
	public Timestamp getTimestamp ( int nColumnIndex, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTimestamp" );
		
		return null;
	}
	

	/**
	*	Gets the value of the designated column in the current row of this ResultSet object as a java.sql.Timestamp object. 
	*/
	public Timestamp getTimestamp ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTimestamp" );
		
		return null;
	}
	

	/**
	*	Returns the value of the designated column in the current row of this ResultSet object as a java.sql.Timestamp object in the Java programming language. 
	*/
	public Timestamp getTimestamp ( String strColumnName, Calendar cal ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getTimestamp" );
		
		return null;
	}
	

	/**
	*	Returns the type of this ResultSet object. 
	*/
	public int getType ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getType" );
		
		return -1;
	}
	

	/**
	*	Deprecated. use getCharacterStream in place of getUnicodeStream 
	*/
	public InputStream getUnicodeStream ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getUnicodeStream" );
		
		return null;
	}
	

	/**
	*	Deprecated. 	
	*/
	public InputStream getUnicodeStream ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "getUnicodeStream" );
		
		return null;
	}
	

	/**
	*	Returns the first warning reported by calls on this ResultSet object. 
	*/
	public SQLWarning getWarnings ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		// TODO: Implement properly
		
		return null;
	}
	

	/**
	*	Inserts the contents of the insert row into this ResultSet object and into the database. 
	*/
	public void insertRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		validateEditable ( );
		
		if ( !m_bCursorOnInsertedRow )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::insertRow " +
								"Cursor is not on the insert row" );
		if ( m_rCache == null || m_rCache. size ( ) < m_sqlCommand. getSELECTColumnCount ( ) )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::insertRow " +
								"New values must be provided for each and every column in ResultSet" );
		
		// Currently I insert a row only in to the main FORM table. JOINed tables are left unchanged.
		
		/**
		*		1. Generate necessary opDataArray and opFieldArray objects
		*		2. Create the new record
		*			2.1 Select all records and put them in a set.
		*			2.2 Perform opProcess. CreateRecord call
		*			2.3 Select all records again and put them in a set.
		*			2.4 Calculate the difference between two sets and this obtain a new set
		*				with only one record in it - newly created one.
		*			2.5 Extract and save the mRecordNumber of the new record.
		*		3. Cache newly created row
		*		4. Update necessary internal variables to retain proper cursor navigability
		*/
		
		// 1.
		opFieldArray		opfArray = m_rCache. getopFields ( m_sqltDBMD. getDBName ( ) );
		opDataArray			opdArray = m_rCache. getopData ( m_sqltDBMD. getDBName ( ) );

		/*
		*	opProcess. CreateRecord ( ) method changes current selection and thus I need
		*	to use unique access lock so that anybody ( including this resultset ) would
		*	have to restore current selection before doing anything else.
		*/
		Object		objLock = new Object ( );
		/*
			There is no need to restore selection because I'm going to change it anyway,
			that's why the thrid parameter is false.
		*/
		registerTables ( objLock, false, false ); // registers tables but does not restore their selections
			synchronized ( m_objDataAccessLock )
			{
				// 2. 
String		strSemaphoreName = "jdbc_rs_ins_smphr_" + m_sqltDBMD. getNumber ( );
				try
				{
/*
	Use 4d semaphores to make record creation an atomic operation.
	Create a global unique semaphore name for a given table so that
	nobody else can create new record until I'm done creating.
*/
					boolean			bUseOptimization = ( m_connectionImpl != null ) ? m_connectionImpl. isUsingOptimization ( ) : false;
				if ( !bUseOptimization )
				{
					m_opTableHelper. mTableNumber = m_sqltDBMD. getNumber ( );
					
			if ( !m_fdExtensions. waitForSemaphore ( strSemaphoreName, 30 ) )
				throw new SQLException (
									"com.fourd.jdbc.ResultSetImpl::insertRow: Insert semaphore timed out." +
									"\nUnable to insert new row.");
			
					// 2.1
					opSelection		opsHelper = m_opProcess. AllRecords ( m_opTableHelper );
					// m_opProcess. RecordsInSelection ( opsHelper );
// System. out. println ( "Records in selection : " + opsHelper. mRecordsInSelection );
					opSet			opSetBeforeCreate = new opSet ( );
					opSetBeforeCreate. mTableNumber = m_sqltDBMD. getNumber ( );
					opSetBeforeCreate. mSetName = m_connectionImpl. generateSetName ( );
					opSetBeforeCreate. mRecordNumber = 9999;
					m_opProcess. CreateSet ( opSetBeforeCreate );

					// 2.2
					m_opProcess. CreateRecord ( opfArray, opdArray );

					// 2.3
					opsHelper = m_opProcess. AllRecords ( m_opTableHelper );
					// m_opProcess. RecordsInSelection ( opsHelper );
// System. out. println ( "Records in selection : " + opsHelper. mRecordsInSelection );
					opSet			opSetAfterCreate = new opSet ( );
					opSetAfterCreate. mTableNumber = m_sqltDBMD. getNumber ( );
					opSetAfterCreate. mSetName = m_connectionImpl. generateSetName ( );
					opSetAfterCreate. mRecordNumber = 9999;
					m_opProcess. CreateSet ( opSetAfterCreate );
					
					// 2.4
					opSet			opSetNewRecord = new opSet ( );
					opSetNewRecord. mTableNumber = m_sqltDBMD. getNumber ( );
					opSetNewRecord. mSetName = m_connectionImpl. generateSetName ( );
					m_opProcess. DifferenceSet (
											opSetAfterCreate,
											opSetBeforeCreate,
											opSetNewRecord );
											
					// 2.5
					m_opProcess. UseSet ( opSetNewRecord );
					opsHelper. mRecordNumber = -1;
					/*
						A call to RecordsInSelection not only returns amount of rows in 
						a selection (which is 1 currently) but also returns mRecordNumber 
						for current (just created) row in selection.
						I need mRecordNumber in order to be able to access it later because 
						this value is used in opProcess. DeleteRecord ( ). Unfortunately, 
						mRecordNumber has nothing to do with row order number. 
						mRecordNumber is just some number assigned to a row.
					*/
					m_opProcess. RecordsInSelection ( opsHelper );
					
			m_fdExtensions. clearSemaphore ( strSemaphoreName );
			
					m_vctrNewRows. addElement ( String. valueOf ( opsHelper. mRecordNumber ) );
// System. out. println ( "New record number : " + opsHelper. mRecordNumber );

					// Clear used sets
					m_opProcess. ClearSet ( opSetBeforeCreate );
					m_opProcess. ClearSet ( opSetAfterCreate );
					m_opProcess. ClearSet ( opSetNewRecord );
					
					m_nRowCount++;

					/*
						// the following call is essential because it brings the record
						// that I will update inside the current selection
					m_opProcess. AllRecords ( m_opTableHelper );
					
					m_opTableHelper. mRecordNumber = opsHelper. mRecordNumber;
					m_opTableHelper. mTableNumber = m_sqltDBMD. getNumber ( );
					m_opProcess. GotoRecord ( m_opTableHelper );
					opdArray. mDataArray [ 0 ]. mString = "completely new value";
					m_opProcess. ModifyRecord ( opfArray, opdArray );
					*/
				}
				else
				{
					// Lets use optimized insert
					int			nNewRecordNumber = m_opProcess. CreateRecord2004 ( opfArray, opdArray );
					m_vctrNewRows. addElement ( String. valueOf ( nNewRecordNumber ) );

					m_nRowCount++;
				}
				}
				catch ( opException opE )
				{
					m_fdExtensions. clearSemaphore ( strSemaphoreName );
					unregisterTables ( objLock );
					
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::insertRow: " + opE. GetErrorMsg ( ) +
										"\nUnable to insert new row.");
				}
			}
		unregisterTables ( objLock );
		
		// 3.
		/*
			TODO:
				Do I need to refreshRow ( ); ? So that any "hidden" JOINed columns are also 
				prefetched? Probably not (at least not now).
				
		*/
		m_rsDataCache. insert ( m_nRowCount, m_rCache );
		
		deleteRowCache ( );
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Indicates whether the cursor is after the last row in this ResultSet object. 
	*/
	public boolean isAfterLast ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		// throwUnimplementedException ( "" );
		return m_nCurrentRow == POSITION_AFTER_LAST;
	}
	

	/**
	*	Indicates whether the cursor is before the first row in this ResultSet object. 
	*/
	public boolean isBeforeFirst ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		// throwUnimplementedException ( "" );
		return m_nCurrentRow == POSITION_BEFORE_FIRST;
	}
	

	/**
	*	Indicates whether the cursor is on the first row of this ResultSet object. 
	*/
	public boolean isFirst ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		if ( m_bCursorOnDeletedRow || m_bCursorOnInsertedRow )
			return false;
			
		if ( m_nRowCount > 0 && m_nCurrentRow == 1 )
			return true;
		else
			return false;

		//throwUnimplementedException ( "" );
	}
	

	/**
	*	Indicates whether the cursor is on the last row of this ResultSet object. 
	*/
	public boolean isLast ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		if ( m_bCursorOnDeletedRow || m_bCursorOnInsertedRow )
			return false;
			
		if ( m_nRowCount > 0 && m_nCurrentRow == m_nRowCount )
			return true;
		else
			return false;
			
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Moves the cursor to the last row in this ResultSet object. 
	*/
	public boolean last ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		resetAllEdits ( );
		
		if ( m_nRowCount == 0 )
			return false;
			
		m_nCurrentRow = m_nRowCount;
		
		// throwUnimplementedException ( "" );
		return true;
	}
	

	/**
	*	Moves the cursor to the remembered cursor position, usually the current row. 
	*/
	public void moveToCurrentRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		m_bCursorOnInsertedRow = false;
		deleteRowCache ( );
	}
	

	/**
	*	Moves the cursor to the insert row. 
	*/
	public void moveToInsertRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );

		resetAllEdits ( );
		
		m_bCursorOnInsertedRow = true;
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Moves the cursor down one row from its current position. 
	*/
	public boolean next ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		if ( m_bCursorOnInsertedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::next " +
									"Cursor is on the insert row" );
			
		if ( m_bCursorOnDeletedRow )
		{
			m_bCursorOnDeletedRow = false;
			
			// There may or may not be rows after the current (deleted) row
			if ( m_nCurrentRow == m_nRowCount + 1 ) // deleteRow changes m_nRowCount but does not change m_nCurrentRow
			{
				m_nCurrentRow = POSITION_AFTER_LAST;
				return false;
			}
			else
				// There is no need for m_nRowCount++
				return true;
		}
		
		if ( m_nRowCount == 0 || m_nCurrentRow == m_nRowCount ) // Nowhere to move
		{
			// throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::next: Nowhere to move" );
			m_nCurrentRow = POSITION_AFTER_LAST;
			
			return false;
		}
			
		if ( m_nCurrentRow == POSITION_BEFORE_FIRST )
		{
			m_nCurrentRow = 1; // row index is 1-based
			
			/*
				There is no need to do anything else because 4D DB autmatically makes the first row
				in a selection the current one.
			*/
			return true;
		}
		
		// Lets move one row forward
		
		/*
		// There is no need to move server-side cursor because RSDataCache uses opProcess. RangeSelectionToArray
		// which fetches rows by index, not bu current row.
		
		registerTables ( );
			synchronized ( m_objDataAccessLock )
			{
				try
				{
					m_opSelection. mCurrentRecord = m_nCurrentRow; // opSelection. mCurrentRecord is 0-based.
					m_opProcess. GotoSelectedRecord ( m_opSelection );
				}
				catch ( opException opE )
				{
					unregisterTables ( );
					
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::next: " + opE. GetErrorMsg ( ) +
										"/nUnable to move to the next row.");
				}
			}
		unregisterTables ( );
		*/
		
		m_nCurrentRow++;
		
		return true;
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Moves the cursor to the previous row in this ResultSet object. 
	*/
	public boolean previous ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		if ( m_bCursorOnInsertedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::previous " +
									"Cursor is on the insert row" );
		if ( m_bCursorOnDeletedRow )
		{
			m_bCursorOnDeletedRow = false;
			
			// There may or may not be rows before the current (deleted) row
			if ( m_nCurrentRow == 1 ) // deleteRow changes m_nRowCount but does not change m_nCurrentRow
			{
				m_nCurrentRow = POSITION_BEFORE_FIRST;
				return false;
			}
			else
			{
				m_nCurrentRow--;
				return true;
			}
		}
		
		if ( m_nRowCount == 0 || m_nCurrentRow == 1 ) // Nowhere to move
		{
			// throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::next: Nowhere to move" );
			m_nCurrentRow = POSITION_BEFORE_FIRST;
			
			return false;
		}
			
		if ( m_nCurrentRow == POSITION_AFTER_LAST )
		{
			m_nCurrentRow = m_nRowCount; // row index is 1-based
			
			return true;
		}
		
		m_nCurrentRow--;
		
		return true;
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Refreshes the current row with its most recent value in the database. 
	*/
	public void refreshRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );

		if ( m_bCursorOnInsertedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::refreshRow " +
									"Cursor is on the insert row" );
									
		if ( m_opProcess == null )
			// Disconnected result set
			return;
			
		validateCurrentRow ( );
		
		// TODO: implement refreshing of newly created rows
		if ( m_nCurrentRow > m_nOriginalRowCount )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::refreshRow " +
									"Refresh of new rows is not implemented" );
			
		// Remove current row from cache
		m_rsDataCache. removeRow ( m_nCurrentRow );
		
		// Ask for a value from removed row and thus refetch and recache it.
		int		nOldFetchSize = getFetchSize ( );
		setFetchSize ( 1 ); // so that only one row is fetched
		
		// getObject ( 1 ); // request object from the first column and force fetching
		/*
		*	I can't just use getObject ( 1 ) to prefetch a row, because
		*	a cached value will be returned if the whole row is not cached BUT
		*	row update is in progerss. Thus, I need to ask RSDataCache for a value.
		*/
		m_rsDataCache. getValue ( m_nCurrentRow, 0 );
		
		setFetchSize ( nOldFetchSize ); // restore fetch size

		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Moves the cursor a relative number of rows, either positive or negative. 
	*/
	public boolean relative ( int nRows ) throws SQLException
	{
		/*
			TODO: Test to see if it is going to be enough (performans-wise) to just
			use the following code :
			
				for ( int i = 0; i < nRows; i++ )
					next ( ); // previous ( );
		*/
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		if ( m_bCursorOnInsertedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl::relative " +
									"Cursor is on the insert row" );

		if ( nRows == 0 )
			// Do not move cursor at all
			return	!m_bCursorOnDeletedRow &&
					m_nCurrentRow != POSITION_BEFORE_FIRST &&
					m_nCurrentRow != POSITION_AFTER_LAST;
			
		if ( m_bCursorOnDeletedRow )
		{
			m_bCursorOnDeletedRow = false;
			if ( nRows > 0 )
			{
				if ( m_nCurrentRow + nRows - 1 > m_nRowCount )
				{
					m_nCurrentRow = POSITION_AFTER_LAST;
					return false;
				}
				m_nCurrentRow += nRows - 1;
			}
			else
			{
				if ( m_nCurrentRow + nRows <= 0 )
				{
					m_nCurrentRow = POSITION_BEFORE_FIRST;
					return false;
				}
				m_nCurrentRow += nRows;
			}
			
			return true;
		}
		
		if ( nRows > 0 ) // move forward
		{
			if (
				m_nRowCount == 0 || // already nowhere to move
				m_nCurrentRow == POSITION_AFTER_LAST || // already nowhere to move
				(	m_nCurrentRow == POSITION_BEFORE_FIRST && nRows > m_nRowCount || // too little rows left to move this far
					m_nCurrentRow != POSITION_BEFORE_FIRST && m_nCurrentRow + nRows > m_nRowCount ) ) // too little rows left to move this far
			{
				m_nCurrentRow = POSITION_AFTER_LAST;
				return false;
			}
			
			// If I get here then cursor is on the valid row and there is space to move.
			if ( m_nCurrentRow == POSITION_BEFORE_FIRST )
				m_nCurrentRow = nRows;
			else
				m_nCurrentRow += nRows;

				return true;
		}
		else // move backward
		{
			if (
				m_nRowCount == 0 || // already nowhere to move
				m_nCurrentRow == POSITION_BEFORE_FIRST || // already nowhere to move
				(	m_nCurrentRow == POSITION_AFTER_LAST && nRows > m_nRowCount || // too little rows left to move this far
					m_nCurrentRow != POSITION_AFTER_LAST && m_nCurrentRow + nRows <= 0 ) ) // too little rows left to move this far
			{
				m_nCurrentRow = POSITION_BEFORE_FIRST;
				return false;
			}
			
			// If I get here then cursor is on the valid row and there is space to move.
			if ( m_nCurrentRow == POSITION_AFTER_LAST )
				m_nCurrentRow = m_nRowCount + nRows + 1;
			else
				m_nCurrentRow += nRows;

				return true;
		}
		
		// throwUnimplementedException ( "" );
	} // end of public boolean relative ( int nRows ) throws SQLException
	

	/**
	*	Indicates whether a row has been deleted. 
	*/
	public boolean rowDeleted ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "rowDeleted" );
		
		return false;
	}
	

	/**
	*	Indicates whether the current row has had an insertion. 
	*/
	public boolean rowInserted ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "rowInserted" );
		
		return false;
	}
	

	/**
	*	Indicates whether the current row has been updated. 
	*/
	public boolean rowUpdated ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "rowUpdated" );
		
		return false;
	}
	

	/**
	*	Gives a hint as to the direction in which the rows in this ResultSet object will be processed. 
	*/
	public void setFetchDirection ( int nDirection ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "setFetchDirection" );
	}
	

	/**
	*	Gives the JDBC driver a hint as to the number of rows that should be fetched from the database when more rows are needed for this ResultSet object. 
	*/
	public void setFetchSize ( int nRows ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		if ( m_nFetchSize < 0 )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:setFetchSize: fetch value should be greater than 0" );
			
		if ( m_nFetchSize == 0 )
		{
			m_nFetchSize = DEFAULT_FETCH_SIZE;
			
			return;
		}
		
		m_nFetchSize = nRows;
	}
	

	/**
	*	Updates the designated column with an ascii stream value. 
	*/
	public void updateAsciiStream ( int nColumnIndex, InputStream is, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with an ascii stream value. 
	*/
	public void updateAsciiStream ( String strColumnName, InputStream is, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a java.math.BigDecimal value. 
	*/
	public void updateBigDecimal ( int nColumnIndex, BigDecimal bd ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a java.sql.BigDecimal value. 
	*/
	public void updateBigDecimal ( String strColumnName, BigDecimal bd ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a binary stream value. 
	*/
	public void updateBinaryStream ( int nColumnIndex, InputStream is, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a binary stream value. 
	*/
	public void updateBinaryStream ( String strColumnName, InputStream is, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a boolean value. 
	*/
	public void updateBoolean ( int nColumnIndex, boolean bNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Boolean ( bNewValue ) );
	}
	

	/**
	*	Updates the designated column with a boolean value. 
	*/
	public void updateBoolean ( String strColumnName, boolean bNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Boolean ( bNewValue ) );
	}
	

	/**
	*	Updates the designated column with a byte value. 
	*/
	public void updateByte ( int nColumnIndex, byte btNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Byte ( btNewValue ) );
	}
	

	/**
	*	Updates the designated column with a byte value. 
	*/
	public void updateByte ( String strColumnName, byte btNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Byte ( btNewValue ) );
	}
	

	/**
	*	Updates the designated column with a byte array value. 
	*/
	public void updateBytes ( int nColumnIndex, byte [ ] szbtNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, szbtNewValue );
	}
	

	/**
	*	Updates the designated column with a boolean value. 
	*/
	public void updateBytes ( String strColumnName, byte [ ] szbtNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, szbtNewValue );
	}
	

	/**
	*	Updates the designated column with a character stream value. 
	*/
	public void updateCharacterStream ( int nColumnIndex, Reader rdr, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a character stream value. 
	*/
	public void updateCharacterStream ( String strColumnName, Reader rdr, int nLength ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a java.sql.Date value. 
	*/
	public void updateDate ( int nColumnIndex, java.sql.Date dtNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, dtNewValue );
	}
	

	/**
	*	Updates the designated column with a java.sql.Date value. 
	*/
	public void updateDate ( String strColumnName, java.sql.Date dtNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, dtNewValue );
	}
	

	/**
	*	Updates the designated column with a double value. 
	*/
	public void updateDouble ( int nColumnIndex, double dblNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Double ( dblNewValue ) );
	}
	

	/**
	*	Updates the designated column with a double value. 
	*/
	public void updateDouble ( String strColumnName, double dblNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Double ( dblNewValue ) );
	}
	

	/**
	*	Updates the designated column with a float value. 
	*/
	public void updateFloat ( int nColumnIndex, float fltNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Float ( fltNewValue ) );
	}
	

	/**
	*	Updates the designated column with a float value. 
	*/
	public void updateFloat ( String strColumnName, float fltNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Float ( fltNewValue ) );
	}
	

	/**
	*	Updates the designated column with an int value. 
	*/
	public void updateInt ( int nColumnIndex, int nNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Integer ( nNewValue ) );
	}
	

	/**
	*	Updates the designated column with an int value. 
	*/
	public void updateInt ( String strColumnName, int nNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Integer ( nNewValue ) );
	}
	

	/**
	*	Updates the designated column with a long value. 
	*/
	public void updateLong ( int nColumnIndex, long lNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Long ( lNewValue ) );
	}
	

	/**
	*	Updates the designated column with a long value. 
	*/
	public void updateLong ( String strColumnName, long lNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Long ( lNewValue ) );
	}
	

	/**
	*	Gives a nullable column a null value. 
	*/
	public void updateNull ( int nColumnIndex ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a null value. 
	*/
	public void updateNull ( String strColumnName ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with an Object value. 
	*/
	public void updateObject ( int nColumnIndex, Object objNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );
		prepareForColumnUpdate ( nColumnIndex, objNewValue );
		m_rCache. put ( nColumnIndex, objNewValue );
	}
	

	/**
	*	Updates the designated column with an Object value. 
	*/
	public void updateObject ( int nColumnIndex, Object x, int	scale) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with an Object value. 
	*/
	public void updateObject ( String strColumnName, Object objNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );
		int		nColumnQueryIndex = m_sqlCommand. getSELECTColumnIndex ( strColumnName ) + 1;
		// prepareForColumnUpdate ( nColumnQueryIndex, objNewValue ); // this call will be done in the method below
		updateObject ( nColumnQueryIndex, objNewValue );
	}
	

	/**
	*	Updates the designated column with an Object value. 
	*/
	public void updateObject ( String strColumnName, Object x, int	scale) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the underlying database with the new contents of the current row of this ResultSet object. 
	*/
	public void updateRow ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		validateEditable ( );
		validateCurrentRow ( );
		
		/*
		*		1. Make sure that at least one change was made to the current row.
		*		2. Create opFieldArray and opDataArray to perform opProcess. ModifyRecord method call.
		*		3. Request necessary tables for exclusive use.
		*		4. Perform update.
		*		5. Restore selection if necessary.
		*		6. Release previously requested tables.
		*		7. Synchronize RSDataCache with DB.
		*/
		
		// 1.
		if ( m_bCursorOnInsertedRow )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::updateRow: " +
								"updateRow can not be called while cursor is on the insert row." );

		if ( m_rCache == null )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::updateRow: " +
								"No changes were made to the current row." );
		
		/**
		*	Currently I support updates only to main FROM table.
		*/
		// 2.
		opFieldArray		opfArray = m_rCache. getopFields ( m_sqltDBMD. getDBName ( ) );
		if ( opfArray. mSize == 0 )
		{
			/*
			*	If the row is not yet cached but JOINed table is being updated
			*	then new updated value will be replaced with the old one when
			*	the JOINed values are actually fetched and cached. This is wrong.
			*	The new value value should stick.
			*	On the other hand, if the row is already prefetched and cached then
			*	the new value will stay there and will not be replaced.
			*	WORKAROUND: If columns from JOINed tables are being updated and current row
			*	is not yet cached then I should prefetch it first and only then perform an update.
			*/
			if ( m_rsDataCache. rowIsCached ( m_nCurrentRow ) ) // also true for newly created rows
				m_rsDataCache. update ( m_nCurrentRow, m_rCache );
			else
			{
				if ( m_rCache. size ( ) == 0 )
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::updateRow: " +
										"No changes were made to the current row." );
										
				/*
					Changes were in fact made but only to JOINed columns, which I currenly 
					do not save in the database.
				*/
				refreshRow ( );
				m_rsDataCache. update ( m_nCurrentRow, m_rCache );
			}
		
			deleteRowCache ( );
			
			return;
		}
		
		opDataArray			opdArray = m_rCache. getopData ( m_sqltDBMD. getDBName ( ) );

		// 3.
		registerTables ( );
			synchronized ( m_objDataAccessLock )
			{
				// 4.
				try
				{
					if ( m_nCurrentRow <= m_nOriginalRowCount )
					{ // update original row that was present when result set was created
						m_opSelection. mCurrentRecord = mapRowToSelection ( m_nCurrentRow ) - 1; // m_nCurrentRow - 1;
						m_opProcess. GotoSelectedRecord ( m_opSelection );
						m_opProcess. ModifyRecord ( opfArray, opdArray );
					}
					else
					{
						/*
							This will make sure that the record I'm going to update,
							is within the current selection and thus is available for update.
						*/
						/*
							TODO: optimization. There is no need to restore original selection for 
									this 'else' branch.
						*/
						m_opTableHelper. mTableNumber = m_sqltDBMD. getNumber ( );
						m_opProcess. AllRecords ( m_opTableHelper );
					
						int			nNewRecordIndex = m_nCurrentRow - m_nOriginalRowCount - 1;
						m_opTableHelper. mRecordNumber = Integer. valueOf (
															( String ) m_vctrNewRows. elementAt (
																nNewRecordIndex ) ). intValue ( );
						
						m_opProcess. GotoRecord ( m_opTableHelper );
						m_opProcess. ModifyRecord ( opfArray, opdArray );
					}
				}
				catch ( opException opE )
				{
					unregisterTables ( ); // 6.
					
					throw new SQLException (
										"com.fourd.jdbc.ResultSetImpl::updateRow: " + opE. GetErrorMsg ( ) +
										"\nUnable to update current row.");
				}
			}
		// 6.
		unregisterTables ( );
		
		// 7.
		/*
			Optimization.
			If the row is not cached then do nothing. If the row is cached then update local cache
			and do not refetch the whole row.
		*/
		/*
		*	If the row is not yet cached but JOINed table is being updated
		*	then new updated value will be replaced with the old one when
		*	the JOINed values are actually fetched and cached.
		*	On the other hand, if the row is already prefetched and cached then
		*	the new value will stay there and will not be replaced.
		*	WORKAROUND: If columns from JOINed tables are being updated and current row
		*	is not yet cached then I should prefetch it first and only then perform an update.
		*/
		if ( m_rsDataCache. rowIsCached ( m_nCurrentRow ) ) // also true for newly created rows
			m_rsDataCache. update ( m_nCurrentRow, m_rCache );
		else
		{
			if ( m_rCache. size ( ) - opdArray. mSize > 0 )
			{
				/*
					Some changes were made to JOINed columns, which I currenly 
					do not save in the database.
				*/
				refreshRow ( );
				m_rsDataCache. update ( m_nCurrentRow, m_rCache );
			}
		}
		
		deleteRowCache ( );
		// throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a short value. 
	*/
	public void updateShort ( int nColumnIndex, short sNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, new Short ( sNewValue ) );
	}
	

	/**
	*	Updates the designated column with a short value. 
	*/
	public void updateShort ( String strColumnName, short sNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, new Short ( sNewValue ) );
	}
	

	/**
	*	Updates the designated column with a String value. 
	*/
	public void updateString ( int nColumnIndex, String strNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
		
		updateObject ( nColumnIndex, strNewValue );
	}

	/**
	*	Updates the designated column with a String value. 
	*/
	public void updateString ( String strColumnName, String strNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, strNewValue );
	}
	

	/**
	*	Updates the designated column with a java.sql.Time value. 
	*/
	public void updateTime ( int nColumnIndex, Time tmNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( nColumnIndex, tmNewValue );
	}
	
	/**
	*	Updates the designated column with a java.sql.Time value. 
	*/
	public void updateTime ( String strColumnName, Time tmNewValue ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		updateObject ( strColumnName, tmNewValue );
	}
	

	/**
	*	Updates the designated column with a java.sql.Timestamp value. 
	*/
	public void updateTimestamp ( int nColumnIndex, Timestamp x ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Updates the designated column with a java.sql.Timestamp value. 
	*/
	public void updateTimestamp ( String strColumnName, Timestamp x ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		throwUnimplementedException ( "" );
	}
	

	/**
	*	Reports whether the last column read had a value of SQL NULL .
	*/
	public boolean wasNull ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: ResultSet is closed" );
			
		return false;
	}
	

// !!!!!!!!!!!!!!!!!!!! ------------------ Methods that are not part of the ResultSet interface ------------------ !!!!!!!!!!!!!!!!!!!! 

	private void s_e_p_a_r_a_t_o_r__________________________1 ( )
	{
		;
	}
	
	/**
		Locks all tables used in this ResultSet for exclusive use.
		Also, restores original selection and positiones cursors to their original positions
		in any of the tables that were used by somebody else since this ResultSet accessed them last time.
		
		TODO: Should I move this method up to the StatementImpl class?
	*/
	void registerTables ( ) throws SQLException
	{
		registerTables ( m_statementImpl );
	}

	void registerTables ( Object objLock ) throws SQLException
	{
		registerTables ( objLock, true );
	}

	void registerTables ( Object objLock, boolean bDoRecordsInSelection ) throws SQLException
	{
		registerTables ( objLock, bDoRecordsInSelection, true );
	}
	
	void registerTables ( Object objLock, boolean bDoRecordsInSelection, boolean bReselect ) throws SQLException
	{
		String [ ]		szstrTablesToReselect = m_connectionImpl. registerTables (
																			m_szstrTableNames,
																			objLock );
		if ( bReselect )
			reselectTables ( szstrTablesToReselect, bDoRecordsInSelection );
	}

	/**
		Unlocks all tables used in this ResultSet and thus makes them available for everybody else to use.
		
		TODO: Should I move this method up to the StatementImpl class?
	*/
	void unregisterTables ( ) throws SQLException
	{
		unregisterTables ( m_statementImpl );
	}

	void unregisterTables ( Object objLock ) throws SQLException
	{
		// Currently we support only one table to be SELECTed
		String			strFROMTable = m_sqltDBMD. getDBName ( );
		m_connectionImpl. unregisterTables ( m_szstrTableNames, objLock );
	}
	
	/**
		Restores all tables in szstrTableNames to their last state (from the point of view of this
		ResultSet object) since they may have been changed by somebody else.
	*/
	private void reselectTables (
								String [ ] szstrTableNames,
								boolean bDoRecordsInSelection ) throws SQLException
	{
		if ( szstrTableNames. length == 0 )
			return;
			
		// TODO: instead of reselecting all tables used in the sql command, reselect only those from szstrTableNames
		m_opSelection = m_statementImpl. restoreSelection (
														m_sqlCommand,
														m_nCurrentRow,
														bDoRecordsInSelection );
		
		// throw new SQLException ( "com.fourd.jdbc.ResultSet::reselectTables: Method is not implemented" );
	}

	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.ResultSetImpl: Method " +
							strMethodName + " is not implemented" );
	}
	
	/**
	*	This method should be called at the very beginning of every getXXX method.
	*	It does nothing if m_nCurrentRow is within proper limits and throws exception otherwise.
	*/
	private void validateCurrentRow ( ) throws SQLException
	{
		if (  m_bCursorOnInsertedRow )
			return;
			
		if ( m_nRowCount == 0 )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: There are no rows" );
			
		if ( m_nCurrentRow == POSITION_BEFORE_FIRST )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: Cursor is before the first row" );
			
		if ( m_nCurrentRow == POSITION_AFTER_LAST )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: Cursor is after the last row" );
			
		if ( m_bCursorOnDeletedRow )
			throw new SQLException ( "com.fourd.jdbc.ResultSetImpl:: Cursor is on deleted row" );
			
		// POSITION_AT_INSERT case should be handled separately
	}
	
	int getRowCount ( )
	{
		return m_nRowCount;
	}
	
	private void prepareHelpers ( )
	{
		int			nFROMTableCount = m_sqlCommand. getFROMTableCount ( );
		int			nJOINTableCount = m_sqlCommand. getJOINTableCount ( );
		
		m_szstrTableNames = new String [ nFROMTableCount + nJOINTableCount ];
									
		SQLTableMetaData [ ]		szsqltMD = m_sqlCommand. getFROMTables ( );
		for ( int i = 0; i < nFROMTableCount; i++ )
			m_szstrTableNames [ i ] = szsqltMD [ i ]. getDBName ( );
			
		szsqltMD = m_sqlCommand. getJOINTables ( );
		for ( int i = 0; i < nJOINTableCount; i++ )
			m_szstrTableNames [ i + nFROMTableCount ] = szsqltMD [ i ]. getDBName ( );
	}
	
	/**
	*	Just a helper method, really. Creates a RowCache object and stores it in m_rCache if
	*	m_rCache is null.
	*/
	private void ensureRowCache ( ) throws SQLException
	{
		if ( m_rCache == null )
			m_rCache = new RowCache ( m_sqlCommand, m_dbMD );
	}
	
	/**
	*	Should be called whenever cursor changes position.
	*/
	private void deleteRowCache ( )
	{
		m_rCache = null;
	}
	
	int mapRowToSelection ( int nCurrentRow )
	{
		/**
		*	I need to calculate current's row index in the original selection. It may be different
		*	because deleted rows do not change selection. Lets say, there were two rows : #1 and #2.
		*	I deleted #1. Thus, #2 becomes #1. However, I still need to refer to it as #2 if 
		*	I want to do any operations on it (like modify, or delete). That's what this
		*	method is about.
		*	To perform the task I need to walk through the m_vctrDeletedRows in reverse order and 
		*	"restore" the original index of nCurrentRow. If index from m_vctrDeletedRows is greater
		*	then nCurrentRow then its deletion could not have affected nCurrentRow. If index from
		*	m_vctrDeletedRows is less then or equal to nCurrentRow then I should increase
		*	nCurrentRow by one.
		*/
		
		int			nResult = nCurrentRow;
		int			nDeletedCount = m_vctrDeletedRows. size ( );
		String		strDeletedIndex = null;
		int			nDeletedIndex = -1;
		for ( int i = nDeletedCount - 1; i >= 0; i-- )
		{
			strDeletedIndex = ( String ) m_vctrDeletedRows. elementAt ( i );
			nDeletedIndex = Integer. parseInt ( strDeletedIndex );
			if ( nResult >= nDeletedIndex )
				nResult++;
		}
		
		return nResult;
	}
	
	private void storeDeletedRowIndex ( int nRow )
	{
		/* Note: I also should keep newly created rows in mind because they do not affect 
				current selection.
		*/
		m_vctrDeletedRows. addElement ( String. valueOf ( nRow ) );
	} // private void storeDeletedRowIndex ( int nRow )
	
	private void validateUpdateXXXInput ( Object objValue ) throws SQLException
	{
		if ( objValue == null )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl: " +
								"Input value can not be null for updateXXX methods." );
	}
	
	/**
	*	nColumn is 1-based
	*/
	private void prepareForColumnUpdate ( int nColumn, Object objValue ) throws SQLException
	{
		validateCurrentRow ( );
		
		if ( nColumn > m_sqlCommand. getSELECTColumnCount ( ) || nColumn < 1 )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::prepareForColumnUpdate: " +
								"Column index " + nColumn + " is out of bounds" );
			
		validateUpdateXXXInput ( objValue );
		ensureRowCache ( );
	}
	
	private void resetAllEdits ( )
	{
		m_bCursorOnDeletedRow = false;
		m_bCursorOnInsertedRow = false;
		deleteRowCache ( );
	}
	
	int getOriginalRowCount ( )
	{
		return m_nOriginalRowCount;
	}
	
	private void validateEditable ( ) throws SQLException
	{
		if ( m_nConcurrency == CONCUR_READ_ONLY )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetImpl::validateEditable: " +
								"Result set is not editable" );
	}
}