/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;

/**
*	Stores information about table column.
*	Theoretically this class does not need any real functionality.
*/
class SQLColumnMetaData
{
	// Stores real name of the column in the database.
	private String			m_strDBName = null;
	
	// Stores real name of the table in the database to which this column belongs.
	private String			m_strDBTableName = null;
	
	/** 
	*	Stores alias name of the table to which this column belongs.
	*	Alias of the table name comes from
	*		SELECT DBT1.DBColumnName AS ColumnAlias
	*		FROM DBTableName AS DBT1
	*
	*	In the example above, DBT1 is an alias.
	*/
	private String			m_strTableAlias = null;
	
	/**
	*	Stores alias for a column. ColumnAlias is an alias for a column DBColumnName in the example for m_strTableAlias
	*/
	private String			m_strAlias = null;
	
	/**
	*	Unique way to identify column (field) in 4D database.
	*/
	private short			m_nNumber = -1;
	
	static final int			ORDER_BY_ASC = 0;
	static final int			ORDER_BY_DESC = 1;
	private int			m_nOrderBy = ORDER_BY_ASC; // default
	
	// Contains opData type of the column
	private int			m_nType = -1; // undefined
	
	SQLColumnMetaData ( )
	{
		;
	}
	
	SQLColumnMetaData ( String strDBName ) throws IllegalArgumentException
	{
		setDBName ( strDBName );
	}
	
	void setDBName ( String strDBName ) throws IllegalArgumentException
	{
		if ( strDBName == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLColumnMetaData.setDBName: Column name can not be null" );
						
		m_strDBName = strDBName;
	}
	
	String getDBName ( )
	{
		return m_strDBName;
	}
	
	/**
	*	Determines if strColumnName represents a valid 4D column name.
	*	Column name may be enclosed in square brackets.
	*	Returns true if the column name is valid, false - otherwise.
	*/
	static boolean isValidColumnName ( String strColumnName )
	{	
		// TODO: find out what are the restrictions for 4D column names
		
		if ( strColumnName == null || SQLCommand. isSQLKeyword ( strColumnName ) )
			return false;
		
		if ( strColumnName. length ( ) == 1 )
			return Character. isLetter ( strColumnName. charAt ( 0 ) );

		// At this point strColumnName contains at least 2 characters.
		
		// Remove enclosing characters
		strColumnName = SQLTools. removeEnclosers ( strColumnName, '[', ']' );
		
		// Now lets make sure that the first character is a letter and all other are letters, digits or underscore.
		/*if ( !Character. isLetter ( strColumnName. charAt ( 0 ) ) )
			return false;
			
		for ( int i = 1; i < strColumnName. length ( ); i++ )
			if (
				!Character. isLetterOrDigit ( strColumnName. charAt ( i ) ) &&
				strColumnName. charAt ( i ) != '_' )
				return false;
		*/

		return true;
	}
	
	/**
	*	Sets the name of the table in the database to which this column belongs
	*/
	void setTableName ( String strDBTableName ) throws IllegalArgumentException
	{
		if ( strDBTableName == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLColumnMetaData.setTableName: Table name can not be null" );
						
		m_strDBTableName = strDBTableName;
	}

	String getDBTableName ( )
	{	
		if ( m_strDBTableName != null )
			return m_strDBTableName;
			
		return getTableAlias ( );
	}

	/**
	*	Sets the alias of the table in the query to which this column belongs.
	*/
	void setTableAlias ( String strTableAlias ) throws IllegalArgumentException
	{
		if ( strTableAlias == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLColumnMetaData.setTableAlias: Table alias can not be null" );
						
		m_strTableAlias = strTableAlias;
	}
	
	String getTableAlias ( )
	{
		return m_strTableAlias;
	}

	void setAlias ( String strAlias ) throws IllegalArgumentException
	{
		if ( strAlias == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLColumnMetaData.setAlias: Column alias can not be null" );
						
		m_strAlias = strAlias;
	}
	
	String getAlias ( )
	{
		if ( m_strAlias == null )
			return m_strDBName;
			
		return m_strAlias;
	}
	
	void setNumber ( short nNumber )
	{
		m_nNumber = nNumber;
	}
	
	short getNumber ( )
	{
		return m_nNumber;
	}
	
	int getopType ( )
	{
		return m_nType;
	}
	
	void setopType ( int nType )
	{
		m_nType = nType;
	}
	
	int getOrderBy ( )
	{
		return m_nOrderBy;
	}
	
	void setOrderBy ( int nOrderBy ) throws SQLException
	{
		if ( nOrderBy != ORDER_BY_ASC && nOrderBy != ORDER_BY_DESC )
			throw new SQLException ( "com.fourd.jdbc.SQLColumnMetaData::setOrderBy: Unknown ORDER BY value" );
			
		m_nOrderBy = nOrderBy; 
	}
}