/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.util.*;
import java.sql.*;

import j4d.open.*;

/**
*	Stores all the data associated with SQL query in a neat organized way.
*	
*	An attempt to represent every single SQL command as one class. 
*	It would be nice to have everything in one place instead of a separate
*	class for each command.
*/
class SQLCommand
{
	static final String			SQL_KEYWORD_SELECT = "SELECT";
	static final String			SQL_KEYWORD_ALL = "ALL";
	static final String			SQL_KEYWORD_DISTINCT = "DISTINCT";
	static final String			SQL_KEYWORD_AS = "AS";
	static final String			SQL_KEYWORD_FROM = "FROM";
	static final String			SQL_KEYWORD_ORDER = "ORDER";
	static final String			SQL_KEYWORD_BY = "BY";
	static final String			SQL_KEYWORD_ASC = "ASC";
	static final String			SQL_KEYWORD_DESC = "DESC";
	static final String			SQL_KEYWORD_WHERE = "WHERE";
	static final String			SQL_KEYWORD_WHERE4D = "WHERE4D";
	static final String			SQL_KEYWORD_AND = "AND";
	static final String			SQL_KEYWORD_OR = "OR";
	static final String			SQL_KEYWORD_TRUE = "TRUE";
	static final String			SQL_KEYWORD_FALSE = "FALSE";
	static final String			SQL_KEYWORD_LEFT = "LEFT";
	static final String			SQL_KEYWORD_JOIN = "JOIN";
	static final String			SQL_KEYWORD_ON = "ON";
	static final String			SQL_KEYWORD_DELETE = "DELETE";
	static final String			SQL_KEYWORD_INSERT = "INSERT";
	static final String			SQL_KEYWORD_INTO = "INTO";
	static final String			SQL_KEYWORD_VALUES = "VALUES";
	static final String			SQL_KEYWORD_UPDATE = "UPDATE";
	static final String			SQL_KEYWORD_SET = "SET";
	static final String			SQL_KEYWORD_CALL = "CALL";
	static final String			SQL_KEYWORD_GO = "GO";
	
	static final private String [ ]	SZ_SQL_KEYWORDS = new String [ ] {
									SQL_KEYWORD_SELECT,
									SQL_KEYWORD_ALL,
									SQL_KEYWORD_DISTINCT,
									SQL_KEYWORD_AS,
									SQL_KEYWORD_FROM,
									SQL_KEYWORD_ORDER,
									SQL_KEYWORD_BY,
									SQL_KEYWORD_ASC,
									SQL_KEYWORD_DESC,
									SQL_KEYWORD_WHERE,
									SQL_KEYWORD_WHERE4D,
									SQL_KEYWORD_AND,
									SQL_KEYWORD_OR,
									SQL_KEYWORD_TRUE,
									SQL_KEYWORD_FALSE,
									SQL_KEYWORD_LEFT,
									SQL_KEYWORD_JOIN,
									SQL_KEYWORD_ON,
									SQL_KEYWORD_DELETE,
									SQL_KEYWORD_INSERT,
									SQL_KEYWORD_INTO,
									SQL_KEYWORD_VALUES,
									SQL_KEYWORD_UPDATE,
									SQL_KEYWORD_SET,
									SQL_KEYWORD_CALL,
									SQL_KEYWORD_GO };

	static final int				CMD_TYPE_SELECT = 0;
	static final int				CMD_TYPE_DELETE = 1;
	static final int				CMD_TYPE_INSERT = 2;
	static final int				CMD_TYPE_UPDATE = 3;
	static final int				CMD_TYPE_STORED_PROCEDURE = 4;
	
	private int				m_nCommandType = -1; // undefined
	private String				m_strWHEREType = null; // undefined, possible values are SQL_KEYWORD_WHERE and SQL_KEYWORD_WHERE4D
	
	private Vector				m_vctrSELECTColumns = new Vector ( );
	private Vector				m_vctrFROMTables = new Vector ( );
	private Vector				m_vctrORDERBYColumns = new Vector ( );
	private Vector				m_vctrWHEREExpressions = new Vector ( );
	private Vector				m_vctrJOINTables = new Vector ( ); // each element is a table that is being joined on some other table
	private Vector				m_vctrValues = new Vector ( );
	
	private String				m_strStoredProcedureName = null;
	// Contains parameter values for each '?' in a query
	private Vector				m_vctrParameters = new Vector ( );
	
	SQLCommand ( int nCommandType ) throws IllegalArgumentException
	{
		if (	nCommandType != CMD_TYPE_SELECT &&
				nCommandType != CMD_TYPE_DELETE &&
				nCommandType != CMD_TYPE_INSERT &&
				nCommandType != CMD_TYPE_UPDATE &&
				nCommandType != CMD_TYPE_STORED_PROCEDURE )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLCommand.SQLCommand: Unknown command type - " +
						nCommandType );
						
		m_nCommandType = nCommandType;
	}
	
	int getType ( )
	{
		return m_nCommandType;
	}
	
	String getWHEREType ( )
	{
		return m_strWHEREType;
	}
	
	void setWHEREType ( String strWHEREType ) throws SQLException
	{
		if (	strWHEREType == null )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::setWHEREType: WHERE type can not be null" );
		
		if (	!strWHEREType. equals ( SQL_KEYWORD_WHERE ) &&
			!strWHEREType. equals ( SQL_KEYWORD_WHERE4D ) )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::setWHEREType: Unknown WHERE type" );
			
		m_strWHEREType = strWHEREType;
	}
	
	void addSELECTColumn ( SQLColumnMetaData sqlcMT )
	{
		if ( sqlcMT == null )
			// Should never happen but PARANOIA IS GOOD
			return;
		
		/*
			TODO: at this point or after the parsing is done I should make sure that
			no duplicate columns are being selected.
			COMMENT: actually, lets allow to select duplicate columns
		*/
		m_vctrSELECTColumns. addElement ( sqlcMT );
	}
	
	/**
	*	Returns the amount of columns to be returned for SELECT quiery
	*/
	int getSELECTColumnCount ( )
	{
		return m_vctrSELECTColumns. size ( );
	}
	
	/**
	*	Returns amount of columns returned in SELECT command from a table with name or alias of strTableAlias
	*	Throws an exception if anything goes wrong.
	*/
	int getSELECTColumnCount ( String strTableAlias ) throws SQLException
	{
		if ( strTableAlias == null )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumnCount: Table alias can not be null" );
		
		// At this point we do not know if strTableAlias contains table alias or table name. Lets find out.
		String					strTableName = getTable ( strTableAlias ). getDBName ( );
		strTableAlias = getTable ( strTableAlias ). getAlias ( );
		
		int						nResult = 0;
		SQLColumnMetaData		sqlcMD = null;
		// First, lets see if strTableAlias contains FROM table name or alias
		if (	FROMTableNameExists ( strTableName ) ||
			strTableAlias != null &&
			FROMTableAliasExists ( strTableAlias ) )
			// Yes, strTableAlias is FROM table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if (	sqlcMD. getDBTableName ( ) == null ||
					sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
					nResult++;
			}
		else
			// strTableAlias is JOIN table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if ( sqlcMD. getDBTableName ( ) == null )
					continue;
					
				if (	sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
					nResult++;
			}
			
		return nResult;
	}
	
	SQLColumnMetaData [ ] getSELECTColumns ( )
	{
		return ( SQLColumnMetaData [ ] ) m_vctrSELECTColumns. toArray (
					new SQLColumnMetaData [ m_vctrSELECTColumns. size ( ) ] );
	}
	
	/**
	*	Returns array of columns returned in SELECT command from a table with name or alias of strTableAlias
	*	Throws an exception if anything goes wrong.
	*/
	SQLColumnMetaData [ ] getSELECTColumns ( String strTableAlias ) throws SQLException
	{
		if ( strTableAlias == null )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumns: Table alias can not be null" );
		
		// At this point we do not know if strTableAlias contains table alias or table name. Lets find out.
		String					strTableName = getTable ( strTableAlias ). getDBName ( );
		strTableAlias = getTable ( strTableAlias ). getAlias ( );
		
		Vector					vctrResult = new Vector ( );
		SQLColumnMetaData		sqlcMD = null;
		// First, lets see if strTableName contains FROM table name or alias
		if (	FROMTableNameExists ( strTableName ) ||
			FROMTableAliasExists ( strTableName ) )
			// Yes, strTableAlias is FROM table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if (	sqlcMD. getDBTableName ( ) == null ||
					sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
					vctrResult. addElement ( sqlcMD );
			}
		else
			// strTableAlias is JOIN table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if ( sqlcMD. getDBTableName ( ) == null )
					continue;
					
				if (	sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
					vctrResult. addElement ( sqlcMD );
			}
			
		return ( SQLColumnMetaData [ ] ) vctrResult. toArray ( new SQLColumnMetaData [ vctrResult. size ( ) ] );
	}
	
	/**
	*	Returns array of 0-based indexes of columns in SELECT clause 
	*	that belong to a table with name or alias of strTableAlias.
	*/
	int [ ] getSELECTColumnIndexes ( String strTableAlias ) throws SQLException
	{
		if ( strTableAlias == null )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumnIndexes: Table alias can not be null" );
		
		// At this point we do not know if strTableAlias contains table alias or table name. Lets find out.
		String					strTableName = getTable ( strTableAlias ). getDBName ( );
		strTableAlias = getTable ( strTableAlias ). getAlias ( );
		
		int [ ]					sznResult = new int [ getSELECTColumnCount ( strTableName ) ];
		int						nCurrent = 0;
		SQLColumnMetaData		sqlcMD = null;
		// First, lets see if strTableAlias contains FROM table name or alias
		if (	FROMTableNameExists ( strTableName ) ||
			strTableAlias != null &&
			FROMTableAliasExists ( strTableAlias ) )
			// Yes, strTableAlias is FROM table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if (	sqlcMD. getDBTableName ( ) == null ||
					sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
				{
					sznResult [ nCurrent ] = i;
					nCurrent++;
				}
			}
		else
			// strTableAlias is JOIN table
			for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
			{
				sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
				if ( sqlcMD. getDBTableName ( ) == null )
					continue;
					
				if (	sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) ||
					(	strTableAlias != null &&
						sqlcMD. getDBTableName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ) )
				{
					sznResult [ nCurrent ] = i;
					nCurrent++;
				}
			}
			
		return sznResult;
	}

	SQLColumnMetaData getSELECTColumn ( int nIndex ) throws SQLException
	{
		if ( nIndex < 0 || nIndex >= m_vctrSELECTColumns. size ( ) )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumn: Column index out of range" );
			
		return ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( nIndex );
	}
	
	/**
	*	Returns 0-based index of a given column in the SELECT part of SQL query.
	*	strColumnAlias is column's alias in the SELECT part of the query. If no alias were provided
	*	then column's name is column's alias.
	*
	*	Throws SQLException if anything goes wrong.
	*/
	int getSELECTColumnIndex ( String strColumnAlias ) throws SQLException
	{
		if ( strColumnAlias == null )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumnIndex: Column alias must not be null" );
			
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < m_vctrSELECTColumns. size ( ); i++ )
		{
			sqlcMD = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( i );
			if ( sqlcMD. getAlias ( ). toUpperCase ( ). equals ( strColumnAlias. toUpperCase ( ) ) )
				return i;
		}
		
		throw new SQLException ( "com.fourd.jdbc.SQLCommand::getSELECTColumnIndex: Unknown colunm alias " + strColumnAlias );
	}
	
	/**
	*	Returns index of the given column within SELECT clause. Throws exception if given colunm does not 
	*	exist or if anything else goes wrong.
	*/
	int getSELECTColumnIndex ( String strTableName, String strColumnName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getSELECTColumnIndex: " +
								"Table name can not be null" );
		if ( strColumnName == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getSELECTColumnIndex: " +
								"Column name must not be null" );

		SQLColumnMetaData [ ]		szsqlcMD = getSELECTColumns ( );
		if ( FROMTableNameExists ( strTableName ) )
		{
			// I'm dealing with FROM table
			
			SQLTableMetaData		sqltMD = getFROMTable ( strTableName );
			String					strtblName = sqltMD. getDBName ( );
			String					strtblAlias = sqltMD. getAlias ( );
			
			for ( int i = 0; i < szsqlcMD. length; i++ )
			{
				if ( szsqlcMD [ i ]. getDBTableName ( ) == null )
				{
					if ( szsqlcMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
						return i;
				}
				else if (	(	szsqlcMD [ i ]. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) || // szsqlcMD [ i ]. getDBTableName ( ) may return table alias
								strtblAlias != null &&
								szsqlcMD [ i ]. getDBTableName ( ). toUpperCase ( ). equals ( strtblAlias. toUpperCase ( ) ) ) &&
							szsqlcMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
					return i;
			}
		}
		else
			for ( int i = 0; i < szsqlcMD. length; i++ )
			{
				if ( szsqlcMD [ i ]. getDBTableName ( ) == null )
					continue;
					
				if (	szsqlcMD [ i ]. getDBTableName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) &&
						szsqlcMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
					return i;
			}
		
		throw new SQLException (
							"com.fourd.jdbc.SQLCommand::getSELECTColumnIndex: " +
							"Column " + strTableName + "." + strColumnName +
							" is not used in the SELECT clause" );
	} // int getSELECTColumnIndex ( String strTableName, String strColumnName ) throws SQLException
	
	boolean SELECTColumnExists ( String strTableName, String strColumnName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::SELECTColumnExists: " +
								"Table name can not be null" );
		if ( strColumnName == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::SELECTColumnExists: " +
								"Column name must not be null" );
		
		SQLColumnMetaData [ ]			szsqlcMD = getSELECTColumns ( strTableName );
		for ( int i = 0; i < szsqlcMD. length; i++ )
			if ( strColumnName. toUpperCase ( ). equals ( szsqlcMD [ i ]. getDBName ( ). toUpperCase ( ) ) )
				return true;
			
		return false;
	}
	
	void addFROMTable ( SQLTableMetaData sqltMT ) throws Exception
	{
		if ( sqltMT == null )
			// Should never happen but PARANOIA IS GOOD
			return;
		
		/*
			Before accepting a table I should make shure that:
				1. There should not be duplicate table names.
				2. There should not be duplicate table aliases.
				3. Name of the table being added should not be the same as any of the existing aliases
				4. Alias of the table being added should not be the same as any of the existing names
		*/
		if ( FROMTableNameExists ( sqltMT. getDBName ( ) ) )
			throw new Exception (
							"com.fourd.jdbc.SQLCommand.addFROMTable: Duplicate table names in the FROM part - " +
							sqltMT. getDBName ( ) );
		if ( FROMTableNameExists ( sqltMT. getAlias ( ) ) )
			throw new Exception (
							"com.fourd.jdbc.SQLCommand.addFROMTable: Alias of the table is the same as the table name in the FROM part - " +
							sqltMT. getAlias ( ) );
		if ( FROMTableAliasExists ( sqltMT. getDBName ( ) ) )
			throw new Exception (
							"com.fourd.jdbc.SQLCommand.addFROMTable: Name of the table is the same as the table alias in the FROM part - " +
							sqltMT. getDBName ( ) );
		if ( FROMTableAliasExists ( sqltMT. getAlias ( ) ) )
			throw new Exception (
							"com.fourd.jdbc.SQLCommand.addFROMTable: Duplicate table aliases in the FROM part - " +
							sqltMT. getAlias ( ) );
			
		m_vctrFROMTables. addElement ( sqltMT );
	}
	
	/**
	*	Returns the amount of tables to be used in SELECT quiery
	*/
	int getFROMTableCount ( )
	{
		return m_vctrFROMTables. size ( );
	}
	
	SQLTableMetaData [ ] getFROMTables ( )
	{
		return ( SQLTableMetaData [ ] ) m_vctrFROMTables. toArray (
					new SQLTableMetaData [ m_vctrFROMTables. size ( ) ] );
	}
	
	/**
	*	Returns table meta data for a table with a given alias or table name.
	*	strTableAlias contains either table alias or table name.
	*/
	SQLTableMetaData getFROMTable ( String strTableAlias ) throws SQLException
	{
		if ( strTableAlias == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getFROMTable: " +
								"Table alias can not be null" );
								
		SQLTableMetaData [ ]		szsqltMD = getFROMTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
			if (	szsqltMD [ i ]. getAlias ( ) != null &&
				szsqltMD [ i ]. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ||
				szsqltMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				return szsqltMD [ i ];
				
		return null;
	}
	
	/**
	*	Determines presence of a FROM table with a given name.
	*/
	private boolean FROMTableNameExists ( String strTableName )
	{
		if ( strTableName == null )
			return false;
		
		SQLTableMetaData		sqltMD = null;	
		for ( int i = 0; i < m_vctrFROMTables. size ( ); i++ )
		{
			sqltMD = ( SQLTableMetaData ) m_vctrFROMTables. elementAt ( i );
			if ( sqltMD. getDBName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	/**
	*	Determines presence of a FROM table with a given alias.
	*/
	private boolean FROMTableAliasExists ( String strTableAlias )
	{
		if ( strTableAlias == null )
			return false;
		
		SQLTableMetaData		sqltMD = null;	
		for ( int i = 0; i < m_vctrFROMTables. size ( ); i++ )
		{
			sqltMD = ( SQLTableMetaData ) m_vctrFROMTables. elementAt ( i );
			if (	sqltMD. getAlias ( ) != null &&
				sqltMD. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	void addORDERBYColumn ( SQLColumnMetaData sqlcMT ) throws SQLException
	{
		if ( sqlcMT == null )
			// Should never happen but PARANOIA IS GOOD
			return;
		
		/*
			Note: Should I make sure that all ORDER BY columns are mentioned in the SELECT clause?
			I guess not, I just have to make sure that column being sorted is present in the table.
			I should be able to execute a SELECT command like the following:
				SELECT LastName, FirstName
				FROM MyTable
				ORDER BY id	// "id" is not present in the SELECT clause
		*/
		
		/*
			Lets make sure that no duplicate columns are being ordered by.
		*/
		if ( ORDERBYColumnExists ( sqlcMT. getDBName ( ) ) )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::addORDERBYColumn: " +
								"Column " + sqlcMT. getDBName ( ) + " is being sorted twice" );
								
		m_vctrORDERBYColumns. addElement ( sqlcMT );
	}
	
	SQLColumnMetaData [ ] getORDERBYColumns ( )
	{
		return ( SQLColumnMetaData [ ] ) m_vctrORDERBYColumns. toArray (
					new SQLColumnMetaData [ m_vctrORDERBYColumns. size ( ) ] );
	}
	
	private boolean ORDERBYColumnExists ( String strColumnName )
	{
		if ( strColumnName == null )
			return false;
			
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < m_vctrORDERBYColumns. size ( ); i++ )
		{
			sqlcMD = ( SQLColumnMetaData ) m_vctrORDERBYColumns. elementAt ( i );
			if ( sqlcMD. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	void addWHEREExpression ( BooleanExpression bExpression ) throws SQLException
	{
		if ( bExpression == null )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::addWHEREExpression: " +
							"Boolean expression can not be null" );
							
		m_vctrWHEREExpressions. addElement ( bExpression );
	}

	BooleanExpression [ ] getWHEREExpressions ( )
	{
		return ( BooleanExpression [ ] ) m_vctrWHEREExpressions. toArray (
					new BooleanExpression [ m_vctrWHEREExpressions. size ( ) ] );
	}
	
	void addJOINTable ( SQLTableMetaData sqltMD ) throws SQLException
	{
		if ( sqltMD == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::addJOINTable: " +
								"Table can not be null" );
		
		m_vctrJOINTables. addElement ( sqltMD );
	}
	
	SQLTableMetaData [ ] getJOINTables ( )
	{
		return ( SQLTableMetaData [ ] ) m_vctrJOINTables. toArray (
					new SQLTableMetaData [ m_vctrJOINTables. size ( ) ] );
	}

	int getJOINTableCount ( )
	{
		return m_vctrJOINTables. size ( );
	}

	/**
	*	Returns table meta data for a JOIN table with a given alias or table name.
	*	strTableAlias contains either table alias or table name.
	*/
	SQLTableMetaData getJOINTable ( String strTableAlias ) throws SQLException
	{
		if ( strTableAlias == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getJOINTable: " +
								"Table alias can not be null" );
								
		SQLTableMetaData [ ]		szsqltMD = getJOINTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
			if (	szsqltMD [ i ]. getAlias ( ) != null &&
				szsqltMD [ i ]. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) ||
				szsqltMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				return szsqltMD [ i ];
				
		return null;
	}
	
	/**
	*	Returns array of columns from table with name strTableName that are joined to by other tables or
	*	are being joined to other columns from other tables.
	*	
	*	String strTableName - Contains DB name of the table
	*	boolean bIncludeSELECTedColumns - Flag that indicates whether or not to return columns
	*		that are mentioned in the SELECT clause.
	*
	*	Returned column names are upperased.
	*/
	String [ ] getColumnsJOINed (
							String strTableName,
							boolean bIncludeSELECTedColumns ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getColumnsJOINedOn: " +
								"Table name can not be null" );
								
		SQLTableMetaData			sqltMD = getTable ( strTableName );
		String					strTableAlias = sqltMD. getAlias ( ); // can be null
		if ( sqltMD == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::getColumnsJOINedOn: " +
								"Table " + strTableName + " is not used in the query" );
			
		Vector					vctrResult = new Vector ( );
		
		SQLTableMetaData	 [ ]		szsqltMD = getJOINTables ( );
		JoinCondition [ ]			szjConditions = null;
		SQLColumnMetaData		sqlcMDLeft = null;
		String					strTableNameLeft = null;
		SQLColumnMetaData		sqlcMDRight = null;
		String					strTableNameRight = null;
		if ( FROMTableNameExists ( strTableName ) )
			// We are dealing with FROM table
			for ( int i = 0; i < szsqltMD. length; i++ )
			{
				szjConditions = szsqltMD [ i ]. getJOINConditions ( );
				for ( int j = 0; j < szjConditions. length; j++ )
				{
					sqlcMDLeft = szjConditions [ j ]. getLeftColumn ( );
					strTableNameLeft = sqlcMDLeft. getDBTableName ( );
					
					if (	bIncludeSELECTedColumns ||
						!SELECTColumnExists ( strTableName, sqlcMDLeft. getDBName ( ) ) )
					{
						if ( strTableNameLeft == null ) // FROM table
						{
							if ( !vctrResult. contains ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) );
						}
						else if (	strTableName. toUpperCase ( ). equals ( strTableNameLeft. toUpperCase ( ) ) ||
							(	strTableAlias != null &&
								strTableAlias. toUpperCase ( ). equals ( strTableNameLeft. toUpperCase ( ) ) ) )
							if ( !vctrResult. contains ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) );
					}

					sqlcMDRight = szjConditions [ j ]. getRightColumn ( );
					strTableNameRight = sqlcMDRight. getDBTableName ( );
					
					if (	bIncludeSELECTedColumns ||
						!SELECTColumnExists ( strTableName, sqlcMDRight. getDBName ( ) ) )
					{
						if ( strTableNameRight == null ) // FROM table
						{
							if ( !vctrResult. contains ( sqlcMDRight. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDRight. getDBName ( ). toUpperCase ( ) );
						}
						else if (	strTableName. toUpperCase ( ). equals ( strTableNameRight. toUpperCase ( ) ) ||
							(	strTableAlias != null &&
								strTableAlias. toUpperCase ( ). equals ( strTableNameRight. toUpperCase ( ) ) ) )
							if ( !vctrResult. contains ( sqlcMDRight. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDRight. getDBName ( ). toUpperCase ( ) );
					}
				}
			} // end of for ( int i = 0; i < szsqltMD. length; i++ )
		else
			for ( int i = 0; i < szsqltMD. length; i++ )
			{
				szjConditions = szsqltMD [ i ]. getJOINConditions ( );
				for ( int j = 0; j < szjConditions. length; j++ )
				{
					sqlcMDLeft = szjConditions [ j ]. getLeftColumn ( );
					strTableNameLeft = sqlcMDLeft. getDBTableName ( );

					if (	bIncludeSELECTedColumns ||
						!SELECTColumnExists ( strTableName, sqlcMDLeft. getDBName ( ) ) )
					{
						if ( strTableNameLeft != null )
						{
							if (	strTableName. toUpperCase ( ). equals ( strTableNameLeft. toUpperCase ( ) ) ||
								(	strTableAlias != null &&
									strTableAlias. toUpperCase ( ). equals ( strTableNameLeft. toUpperCase ( ) ) ) )
								if ( !vctrResult. contains ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDLeft. getDBName ( ). toUpperCase ( ) );
						}
					}

					sqlcMDRight = szjConditions [ j ]. getRightColumn ( );
					strTableNameRight = sqlcMDRight. getDBTableName ( );

					if (	bIncludeSELECTedColumns ||
						!SELECTColumnExists ( strTableName, sqlcMDRight. getDBName ( ) ) )
					{
						if ( strTableNameRight != null )
						{
							if (	strTableName. toUpperCase ( ). equals ( strTableNameRight. toUpperCase ( ) ) ||
								(	strTableAlias != null &&
									strTableAlias. toUpperCase ( ). equals ( strTableNameRight. toUpperCase ( ) ) ) )
								if ( !vctrResult. contains ( sqlcMDRight. getDBName ( ). toUpperCase ( ) ) )
								vctrResult. addElement ( sqlcMDRight. getDBName ( ). toUpperCase ( ) );
						}
					}
				}
			} // end of for ( int i = 0; i < szsqltMD. length; i++ )
			
		return ( String [ ] ) vctrResult. toArray ( new String [ vctrResult. size ( ) ] );
	} // end of SQLColumnMetaData [ ] getColumnsJOINed (
	
	void addValue ( Value val ) throws SQLException
	{
		if ( val == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::addValue: " +
								"New value can not be null");
								
		m_vctrValues. addElement ( val );
	}
	
	void setValue ( int nIndex, String strValue ) throws SQLException
	{
		if ( nIndex < 0 || nIndex > m_vctrValues. size ( ) )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::setValue: " +
								"Invalid index " + nIndex );
								
		if ( nIndex < m_vctrValues. size ( ) )
			( ( Value ) m_vctrValues. elementAt ( nIndex ) ). setValue ( strValue );
		else
		{
			Value		val = new Value ( );
			val. setValue ( strValue );
			m_vctrValues. addElement ( val );
		}
	}
	
	void setValue ( int nIndex, Parameter paramValue ) throws SQLException
	{
		if ( nIndex < 0 || nIndex > m_vctrValues. size ( ) )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::setValue: " +
								"Invalid index " + nIndex );
								
		if ( nIndex < m_vctrValues. size ( ) )
			( ( Value ) m_vctrValues. elementAt ( nIndex ) ). setParameter ( paramValue );
		else
		{
			Value		val = new Value ( );
			val. setParameter ( paramValue );
			m_vctrValues. addElement ( val );
		}
	}
	
	Value [ ] getValues ( )
	{
		return ( Value [ ] ) m_vctrValues. toArray ( new Value [ m_vctrValues. size ( ) ] );
	}
	
	static boolean isSQLKeyword ( String strData )
	{
		if ( strData == null )
			return false;
			
		for ( int i = 0; i < SZ_SQL_KEYWORDS. length; i++ )
			if ( SZ_SQL_KEYWORDS [ i ]. equals ( strData. toUpperCase ( ) ) )
				return true;
				
		return false;
	}
	
	/**
	*	Makes sure that it represents a valid SQL command.
	*/
	void validate ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		if ( m_nCommandType == CMD_TYPE_SELECT )
			validateSELECT ( dbMD );
		else if ( m_nCommandType == CMD_TYPE_DELETE )
			validateDELETE ( dbMD );
		else if ( m_nCommandType == CMD_TYPE_INSERT )
			validateINSERT ( dbMD );
		else if ( m_nCommandType == CMD_TYPE_UPDATE )
			validateUPDATE ( dbMD );
		else if ( m_nCommandType == CMD_TYPE_STORED_PROCEDURE )
			validateStoredProcedure ( dbMD );
		else
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validate: " +
								"Unknown command type : " + m_nCommandType );
	}
	
	private void validateSELECT ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		// Lets make sure there is at least one table to select from
		if ( getFROMTableCount ( ) <= 0 )
			throw new SQLException ( "com.fourd.jdbc.SQLCommand::validateSELECT: There are no tables to select from" );
			
		// Note: should allow no more than 1 FROM table
		
		// Now lets make sure that all columns are being selected	 from listed tables
		SQLColumnMetaData [ ] 	szsqlcMD = getSELECTColumns ( );
		String					strFROMTableName = ( ( SQLTableMetaData ) m_vctrFROMTables. elementAt ( 0 ) ).
																		getDBName ( ); // There is only one FROM table
		for ( int i = 0; i < szsqlcMD. length; i++ )
		{
			if (	szsqlcMD [ i ]. getDBTableName ( ) == null ||
				FROMTableNameExists ( szsqlcMD [ i ]. getDBTableName ( ) ) ||
				FROMTableAliasExists ( szsqlcMD [ i ]. getDBTableName ( ) ) )
			{
				// We are dealing with column from FROM table
				// Lets make sure that the column exists in the table
				if ( !dbMD. columnExists ( strFROMTableName, szsqlcMD [ i ]. getDBName ( ) ) )
					throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateSELECT: " +
									"Column " + szsqlcMD [ i ]. getDBName ( ) +
									" does not exist in the table " + strFROMTableName );
			}
			else if (	JOINTableNameExists ( szsqlcMD [ i ]. getDBTableName ( ) ) ||
					JOINTableAliasExists ( szsqlcMD [ i ]. getDBTableName ( ) ) )
			{
				// We are dealing with column from JOIN table
				// Lets make sure that the column exists in the table
				if ( !dbMD. columnExists (
									getJOINTable ( szsqlcMD [ i ]. getDBTableName ( ) ). getDBName ( ),
									szsqlcMD [ i ]. getDBName ( ) ) )
					throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateSELECT: " +
									"Column " + szsqlcMD [ i ]. getDBName ( ) +
									" does not exist in the table " + szsqlcMD [ i ]. getDBTableName ( ) );
			}
			else
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateSELECT: " +
									"Table " + szsqlcMD [ i ]. getDBTableName ( ) +
									" is not declared in the FROM or JOIN clause. " +
									"Usage of column " + szsqlcMD [ i ]. getDBName ( ) + " is not allowed." );
		}
		
		validateWHERE ( dbMD );
		validateJOIN ( dbMD );
	} // end of private void validateSELECT ( ) throws SQLException
	
	private void validateDELETE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		/*
			Lets make sure that there is exactly one table to delete from.
			As a matter of fact, parsing class will not allow anything else, but
			lets double-check.
		*/
		if ( getFROMTableCount ( ) != 1 )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateSELECT: " +
								"There should be exactly one table to delete from" );
								
		// Make sure that the table exists in the database.
		String			strFROMTable = getFROMTables ( ) [ 0 ]. getDBName ( );
		if ( !dbMD. tableExists ( strFROMTable ) )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::validateSELECT: " +
							"Table " + strFROMTable + " does not exist in the database" );
							
		validateWHERE ( dbMD );
	} // end of private void validateDELETE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	private void validateINSERT ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		/*
			Lets make sure that there is exactly one table to insert to.
			As a matter of fact, parsing class will not allow anything else, but
			lets double-check.
		*/
		if ( getFROMTableCount ( ) != 1 ) // INSERT table is stored as FROM table
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateINSERT: " +
								"There should be exactly one table to insert to" );
								
		// Make sure that the table exists in the database.
		String			strFROMTable = getFROMTables ( ) [ 0 ]. getDBName ( );
		if ( !dbMD. tableExists ( strFROMTable ) )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::validateINSERT: " +
							"Table " + strFROMTable + " does not exist in the database" );
		
		/*
			Now lets make sure that there are as many column values as there are
			column names (if column names are present at all).
			Column names should be either all nulls or all non-nulls.
			Column values should be all non-nulls or parameters.
		*/
		Value [ ]		szValues = getValues ( );
		if ( szValues. length == 0 )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::validateINSERT: " +
							"No values to insert" );
								
		boolean			bFirstColumnNameIsNull = szValues [ 0 ]. getColumnName ( ) == null;
		String			strColumnName = null;
		for ( int i = 0; i < szValues. length; i++ )
		{
			if ( szValues [ i ]. getValue ( ) == null && !szValues [ i ]. isParameter ( ) )
				throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateINSERT: " +
								"Value count is less then column count" );
								
			strColumnName = szValues [ i ]. getColumnName ( );
			if (	bFirstColumnNameIsNull && strColumnName != null ||
					!bFirstColumnNameIsNull && strColumnName == null )
				throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateINSERT: " +
								"Value count is different from column count" );
		}
		
		SQLTableMetaData		sqltMD = dbMD. getTable ( strFROMTable );
		/*
			Now lets see that the column count in INSERT statement does not exceed
			table's total column count.
		*/
		if ( sqltMD. getColumnCount ( ) < szValues. length )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateINSERT: " +
								"Attempt to insert more columns than there is in the table" );

		if ( !bFirstColumnNameIsNull )
		{
			// Now lets see that there are no duplicate columns in the INSERT statement.
			for ( int i = 0; i < szValues. length - 1; i++ )
			{
				strColumnName = szValues [ i ]. getColumnName ( ). toUpperCase ( );
				for ( int j = i + 1; j < szValues. length; j++ )
					if ( szValues [ j ]. getColumnName ( ). toUpperCase ( ). equals ( strColumnName ) )
						throw new SQLException (
										"com.fourd.jdbc.SQLCommand::validateINSERT: " +
										"Duplicate insert column : " + strColumnName );
			}
			
			// Lets verify that all mentioned columns exist in the table
			for ( int i = 0; i < szValues. length; i++ )
				if ( !sqltMD. columnExists ( szValues [ i ]. getColumnName ( ) ) )
					throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateINSERT: " +
									"Column " + szValues [ i ]. getColumnName ( ) +
									" does not exist in the table " + strFROMTable );
		}
		else
		{
			// Set all missing column names
			SQLColumnMetaData [ ]		szsqlcMD = sqltMD. getColumns ( );
			for ( int i = 0; i < szValues. length; i++ )
				( ( Value ) m_vctrValues . elementAt ( i ) ). setColumnName (
										szsqlcMD [ i ]. getDBName ( ) );
		}
	} // end of private void validateINSERT ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	private void validateUPDATE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		/*
			Lets make sure that there is exactly one table to update.
			As a matter of fact, parsing class will not allow anything else, but
			lets double-check.
		*/
		if ( getFROMTableCount ( ) != 1 )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateUPDATE: " +
								"There should be exactly one table to update" );
								
		// Make sure that the table exists in the database.
		String			strFROMTable = getFROMTables ( ) [ 0 ]. getDBName ( );
		if ( !dbMD. tableExists ( strFROMTable ) )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::validateUPDATE: " +
							"Table " + strFROMTable + " does not exist in the database" );
							
		SQLTableMetaData		sqltMD = dbMD. getTable ( strFROMTable );

		/*
			Validate SET part of the command. Namely, make sure that :
				1. All mentioned columns are from the INSERT table.
				2. There are no duplicate columns.
				3. There is no need to check that the amount of values is no more
					than column count because of 1. & 2. 
		*/
		Value [ ]			szValues = getValues ( );
		if ( szValues. length == 0 )
			throw new SQLException (
							"com.fourd.jdbc.SQLCommand::validateUPDATE: " +
							"No values to update" );
		
		// 2. Now lets see that there are no duplicate columns in the UPDATE statement.
		String				strColumnName = null;
		for ( int i = 0; i < szValues. length - 1; i++ )
		{
			strColumnName = szValues [ i ]. getColumnName ( ). toUpperCase ( );
			for ( int j = i + 1; j < szValues. length; j++ )
				if ( szValues [ j ]. getColumnName ( ). toUpperCase ( ). equals ( strColumnName ) )
					throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateUPDATE: " +
									"Duplicate update column : " + strColumnName );
		}
			
		// 1. Lets verify that all mentioned columns exist in the table
		for ( int i = 0; i < szValues. length; i++ )
			if ( !sqltMD. columnExists ( szValues [ i ]. getColumnName ( ) ) )
				throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateUPDATE: " +
								"Column " + szValues [ i ]. getColumnName ( ) +
								" does not exist in the table " + strFROMTable );
							
		validateWHERE ( dbMD );
	} // end of private void validateUPDATE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	private void validateStoredProcedure ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		/*
			Currently, there is no way to find out if the stored procedure really exists 
			on the server or what parameters it has. So I just need to make sure that 
			there is a non-null parameter value for each '?' in the query or stored procedure.
		*/
		for ( int i = 0; i < m_vctrParameters. size ( ); i++ )
			if ( m_vctrParameters. elementAt ( i ) == null )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateStoredProcedure: " +
									"Parameter #" + ( i + 1 ) + " is null" );
	} // end of private void validateStoredProcedure ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	void expandAsteriskInSELECT ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		if ( m_nCommandType != CMD_TYPE_SELECT )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::expandAsteriskInSELECT: " +
								"Command type is not SELECT" );
		if ( dbMD == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::expandAsteriskInSELECT: " +
								"SQLDataBaseMetaData must not be null" );

		if ( m_vctrSELECTColumns. size ( ) < 1 )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::expandAsteriskInSELECT: " +
								"There should be at least one column to SELECT" );
								
		SQLColumnMetaData	sqlcMDAsterisk = ( SQLColumnMetaData ) m_vctrSELECTColumns. elementAt ( 0 );
		if ( !sqlcMDAsterisk. getDBName ( ). equals ( "*" ) )
			// Explisit list of columns instead of just *
			return;
			
		if ( m_vctrSELECTColumns. size ( ) != 1 )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::expandAsteriskInSELECT: " +
								"Expected 1 but found " + m_vctrSELECTColumns. size ( ) + " columns" );
								
		// Remove "*" column
		m_vctrSELECTColumns. clear ( );
		
		// For each SELECT table add all its columns to m_vctrSELECTColumns
		
		SQLTableMetaData			sqltMDFROM = null; // Table metadata from SQL query
		SQLTableMetaData			sqltMDDB = null; // Table metadata from DB
		SQLColumnMetaData [ ]		szsqlcMDDB = null; // Columns meta data from DB
		SQLColumnMetaData		sqlcMDSELECT = null; // New column meta data for SELECT query
		for ( int i = 0; i < m_vctrFROMTables. size ( ); i++ )
		{
			sqltMDFROM = ( SQLTableMetaData ) m_vctrFROMTables. elementAt ( i );
			sqltMDDB = dbMD. getTable ( sqltMDFROM. getDBName ( ) );
			if ( sqltMDDB == null ) // Table does not exist
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::expandAsteriskInSELECT: " +
									"Table " + sqltMDFROM. getDBName ( ) + " does not exist in the database" );
									
			szsqlcMDDB = sqltMDDB. getColumns ( );
			
			/*
				Instead of using column meta data from DB directly, lets duplicate them so that later we 
				can change them without affecting original metadata from DB.
			*/
			for ( int j = 0; j < szsqlcMDDB. length; j++ )
			{
				sqlcMDSELECT = new SQLColumnMetaData ( szsqlcMDDB [ j ]. getDBName ( ) );
				sqlcMDSELECT. setAlias ( szsqlcMDDB [ j ]. getDBName ( ) );
				sqlcMDSELECT. setTableName ( sqltMDFROM. getDBName ( ) );
				
				addSELECTColumn ( sqlcMDSELECT );
			}
		}
	} // end of void expandAsteriskInSELECT ( SQLDataBaseMetaData dbMD )
	
	/**
	*	Returns a "pattern" that represents logical operators in the WHERE clause.
	*	A block of OR is represented as single 'o', a block of AND is represented by single 'a'.
	*	For example, ... OR ... OR ... AND ... AND ... AND ... OR has a pattern "oao".
	*	A single boolean expression is represented by empty string "".
	*/
	private String getWHEREPattern ( )
	{
		BooleanExpression [ ]		szbe = getWHEREExpressions ( );
		String						strResult = "";
		for ( int i = 0; i < szbe. length - 1; i++ )
		{
			if ( strResult. equals ( "" ) ) // The very first boolean operator
			{
				if ( szbe [ i ]. getLogicalOperator ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) )
					strResult = "a";
				else
					strResult = "o";
					
				continue;
			}		
			
			if (	strResult. charAt ( strResult. length ( ) - 1 ) == 'a' &&
					szbe [ i ]. getLogicalOperator ( ). equals ( SQLCommand. SQL_KEYWORD_OR ) )
					strResult += "o";
			else if (	strResult. charAt ( strResult. length ( ) - 1 ) == 'o' &&
						szbe [ i ]. getLogicalOperator ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) )
					strResult += "a";
		}
		
		return strResult;
	} // end of private String getWHEREPattern ( )
	
	/**
	*	For WHERE4D does nothing, for WHERE accepts only the following patterns:
	*	"", "o", "a", "oa", "ao", "oao". If unsupported pattern is present, throws an exception.
	*	Does nothing if the pattern is supported.
	*/
	private void validateWHERE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		// First of all lets make sure that all columns are valid
		BooleanExpression [ ]		szbe = getWHEREExpressions ( );
		if ( szbe. length == 0 )
			// There is nothing to validate
			return;
			
		SQLColumnMetaData			sqlcMD = null;
		for ( int i = 0; i < szbe. length; i++ )
		{
			sqlcMD = szbe [ i ]. getColumnMetaData ( );
			// Make sure that tables from WHERE clause are declared in the query
			if (
				sqlcMD. getDBTableName ( ) != null &&
				!FROMTableNameExists ( sqlcMD. getDBTableName ( ) ) &&
				!FROMTableAliasExists ( sqlcMD. getDBTableName ( ) ) )
					throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateWHERE: " +
									"Table " + sqlcMD. getDBTableName ( ) +
									" is not declared in the FROM part. " +
									"Usage of column " + sqlcMD. getDBName ( ) + " is not allowed." );
									
			/*
				There is no need to check that the table is present in the database, because right now there is only
				FROM table and its existance was verified before (in other methods).
			*/
			
			if (	sqlcMD. getDBTableName ( ) != null &&
					!dbMD. columnExists (
								getFROMTable ( sqlcMD. getDBTableName ( ) ). getDBName ( ),
								sqlcMD. getDBName ( ) ) )
				throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateWHERE: " +
								"Column " + sqlcMD. getDBName ( ) +
								" does not exist in the table " + sqlcMD. getDBTableName ( ) );
		}

		// Now lets check that the combination of AND and OR is acceptable.
		if ( getWHEREType ( ) == null || getWHEREType ( ). equals ( SQL_KEYWORD_WHERE4D ) )
			// With 4D WHERE we support any kind of AND and OR combinations
			return;
			
		String		strPattern = getWHEREPattern ( );
		if (	!strPattern. equals ( "" ) &&
			!strPattern. equals ( "a" ) &&
			!strPattern. equals ( "o" ) &&
			!strPattern. equals ( "ao" ) &&
			!strPattern. equals ( "oa" ) &&
			!strPattern. equals ( "oao" ) )			
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateWHERE: " +
								"Unsupported WHERE configuration. " +
								"Can not have OR-AND-OR pattern more than once, " +
								"can not have AND-OR-AND pattern." );
								
		// Rearrange boolean statements so that AND block always goes before OR
		if ( 	!strPattern. equals ( "oa" ) &&
			!strPattern. equals ( "oao" ) )
			// AND and OR are already properly ordered
			return;
			
		Vector					vctrAND = new Vector ( );
		Vector					vctrOR = new Vector ( );
		
		// First lets extract the leading "o" section
		int						nCurrentExpression = 0;
		int						nLastOR = -1;
		while ( szbe [ nCurrentExpression ]. getLogicalOperator ( ). equals ( SQLCommand. SQL_KEYWORD_OR ) )
		{
			vctrOR. addElement ( szbe [ nCurrentExpression ] );
			nCurrentExpression++;
		}
		
		// Now nCurrentExpression contains index of the first boolean expression in the "a" section
		nLastOR = nCurrentExpression - 1;
		// Lets extract the "a" section
		while (	szbe [ nCurrentExpression ]. getLogicalOperator ( ) != null &&
				szbe [ nCurrentExpression ]. getLogicalOperator ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) )
		{
			vctrAND. addElement ( szbe [ nCurrentExpression ] );
			nCurrentExpression++;
		}
		
		// Now nCurrentExpression contains index of the last expression in the "a" section
		vctrAND. addElement ( szbe [ nCurrentExpression ] );
		szbe [ nCurrentExpression ]. setLogicalOperator ( SQLCommand. SQL_KEYWORD_OR );
		
		if ( nCurrentExpression < szbe. length - 1 ) // There is trailing "o" section
		{
			for ( int i = nCurrentExpression + 1; i < szbe. length; i++ )
				vctrOR. addElement ( szbe [ i ] );
				
			szbe [ szbe. length - 1 ]. setLogicalOperator ( null );
		}
		else
			szbe [ nLastOR ]. setLogicalOperator ( null );
			
		// Now lets save reordered boolean expressions
		m_vctrWHEREExpressions. clear ( );
		for ( int i = 0; i < vctrAND. size ( ); i++ )
			m_vctrWHEREExpressions. addElement ( vctrAND. elementAt ( i ) );
		for ( int i = 0; i < vctrOR. size ( ); i++ )
			m_vctrWHEREExpressions. addElement ( vctrOR. elementAt ( i ) );
			
// DEBUG
/*for ( int i = 0; i < m_vctrWHEREExpressions. size ( ); i++ )
	System. out. println (
					( ( BooleanExpression ) m_vctrWHEREExpressions. elementAt ( i ) ). getComparisonOperator ( ) +
					" " +
					( ( BooleanExpression ) m_vctrWHEREExpressions. elementAt ( i ) ). getLogicalOperator ( ) + " " );
*/
// DEBUG
	} // end of private void validateWHERE ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	private void validateJOIN ( DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		// TODO: break up this method into smaller once for extra clarity.
		
		/*
			To validate:
				1. Table aliases used in JOIN are not table names from JOIN or from FROM.					DONE
				2. Table names used in JOIN are not table aliases used in FROM.							DONE
				3. Table aliases are unique.															DONE
				4. All tables and columns from JOIN actually exist in the database.							DONE
				5. Tables are joined to right of the main FROM table or to the right of the table				DONE
					that is properly joined with main FROM table, directly or indirectly.						DONE
				6. The same table should not be joined twice											DONE
				7. Columns from the JOIN condition should refer to the table being joined and target table		DONE
				8. Columns in join condition should be of the same type									DONE
				9. FROM table is not being JOINed.													DONE
				
			To validate? :
				?1? Each column in the JOIN comdition should be mentioned only once?
		*/
		
		// 9.
		String					strFROMTableName = getFROMTables ( ) [ 0 ]. getDBName ( );
		if ( JOINTableNameExists ( strFROMTableName ) )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateJOIN: " +
								"Table " + strFROMTableName + " from FROM clause " +
								"can not be joined in JOIN clause." );
			

		SQLTableMetaData [ ]		szsqltMDJoin = getJOINTables ( );
		JoinCondition [ ]			szjConditions = null;
		SQLColumnMetaData		sqlcMDLeft = null;
		SQLColumnMetaData		sqlcMDRight = null;
		SQLTableMetaData			sqltMDLeft = null;
		SQLTableMetaData			sqltMDRight = null;
		HashSet					hsTableNames = null;
		int						nColumnTypeLeft = -1;
		int						nColumnTypeRight = -1;
		for ( int i = 0; i < szsqltMDJoin. length; i++ )
		{
			// 2.
			if ( FROMTableAliasExists ( szsqltMDJoin [ i ]. getDBName ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"A table in FROM clause " +
									"uses alias " + szsqltMDJoin [ i ]. getDBName ( ) +
									" which is a name of a table from JOIN clause. Choose different alias." );
			
			// 6.
			if ( !JOINTableNameIsUnique ( szsqltMDJoin [ i ]. getDBName ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"A table can be joined only once, but " + szsqltMDJoin [ i ]. getDBName ( ) +
									" is being joined more than once." );
									
			// 4.
			if ( !dbMD. tableExists ( szsqltMDJoin [ i ]. getDBName ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"Table " + szsqltMDJoin [ i ]. getDBName ( ) +
									" does not exist in the database" );
									
			// 7.&4. Lets make sure that tables and columns in the left and right part of the join condition exist 
			szjConditions = szsqltMDJoin [ i ]. getJOINConditions ( );
			/*
				Columns on the left side of the join condition should be from the same table.
				The same is true for columns from right side but the check is performed automatically.
			*/
			hsTableNames = new HashSet ( );
			for ( int k = 0; k < szjConditions. length; k++ )
			{
				// Lets validate left part of the join condition
				sqlcMDLeft = szjConditions [ k ]. getLeftColumn ( );
				if ( sqlcMDLeft. getTableAlias ( ) == null )
				{
					// Just a column name without a table name/alias and thus this column name is from the single FROM table
					if ( !dbMD. columnExists ( strFROMTableName, sqlcMDLeft. getDBName ( ) ) )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Column " + sqlcMDLeft. getDBName ( ) +
											" does not exist in the table " + strFROMTableName );
											
					// 8. 
					nColumnTypeLeft = dbMD. getTable ( strFROMTableName ). getColumnMetaData ( sqlcMDLeft. getDBName ( ) ). getopType ( );
					
					hsTableNames. add ( strFROMTableName );
				}
				else
				{
					// Lets make sure that the table from left part of the condition is either a valid table in the database or a valid table alias
					sqltMDLeft = getFROMTable ( sqlcMDLeft. getTableAlias ( ) );
					if ( sqltMDLeft == null )
						sqltMDLeft = getJOINTable ( sqlcMDLeft. getTableAlias ( ) );
					if ( sqltMDLeft == null )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Table/alias " + sqlcMDLeft. getTableAlias ( ) +
											" does not exist in the database/query" );
											
					// Now we have to make sure that the column from the left part actually exists in the table
					if ( !dbMD. columnExists ( sqltMDLeft. getDBName ( ), sqlcMDLeft. getDBName ( ) ) )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Column " + sqlcMDLeft. getDBName ( ) +
											" does not exist in the table " + sqltMDLeft. getDBName ( ) );
											
					// 8. 
					nColumnTypeLeft = dbMD. getTable ( sqltMDLeft. getDBName ( ) ). getColumnMetaData ( sqlcMDLeft. getDBName ( ) ). getopType ( );

					hsTableNames. add ( sqltMDLeft. getDBName ( ) );
				}
				
				if ( hsTableNames. size ( ) > 1 )
					throw new SQLException (
										"com.fourd.jdbc.SQLCommand::validateJOIN: " +
										"Left side of join conditions should contain columns from only one table" );

				// Lets validate right part of the join condition
				sqlcMDRight = szjConditions [ k ]. getRightColumn ( );
				if ( sqlcMDRight. getTableAlias ( ) == null )
				{
					// Just a column name without a table name/alias and thus this column name is from the table that is being joined
					if ( !dbMD. columnExists ( szsqltMDJoin [ i ]. getDBName ( ), sqlcMDRight. getDBName ( ) ) )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Column " + sqlcMDRight. getDBName ( ) +
											" does not exist in the table " + szsqltMDJoin [ i ]. getDBName ( ) );

					// 8. 
					nColumnTypeRight = dbMD. getTable ( szsqltMDJoin [ i ]. getDBName ( ) ). getColumnMetaData ( sqlcMDRight. getDBName ( ) ). getopType ( );
				}
				else
				{
					// Lets make sure that the table from right part of the condition is a table name or alias of the table being joined
					if (	!szsqltMDJoin [ i ]. getDBName ( ). toUpperCase ( ). equals ( sqlcMDRight. getTableAlias ( ). toUpperCase ( ) ) &&
						!(	szsqltMDJoin [ i ]. getAlias ( ) != null &&
							szsqltMDJoin [ i ]. getAlias ( ). toUpperCase ( ). equals ( sqlcMDRight. getTableAlias ( ). toUpperCase ( ) ) ) )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Column " + sqlcMDRight. getDBName ( ) + " on the right side of join condition is from table " +
											sqlcMDRight. getTableAlias ( ) + " but should be from table " + szsqltMDJoin [ i ]. getDBName ( ) );
											
					// Now we have to make sure that the column from the right part actually exists in the table
					if ( !dbMD. columnExists ( szsqltMDJoin [ i ]. getDBName ( ), sqlcMDRight. getDBName ( ) ) )
						throw new SQLException (
											"com.fourd.jdbc.SQLCommand::validateJOIN: " +
											"Column " + sqlcMDRight. getDBName ( ) +
											" does not exist in the table " + szsqltMDJoin [ i ]. getDBName ( ) );

					// 8. 
					nColumnTypeRight = dbMD. getTable ( szsqltMDJoin [ i ]. getDBName ( ) ). getColumnMetaData ( sqlcMDRight. getDBName ( ) ). getopType ( );
				}
				
				// 8.
				if ( nColumnTypeLeft != nColumnTypeRight )
					throw new SQLException (
										"com.fourd.jdbc.SQLCommand::validateJOIN: " +
										"Joined columns should be of the same type, but columns " +
										 sqlcMDLeft. getDBName ( ) + " and " + sqlcMDRight. getDBName ( ) +
										" are of different types." );
			} // end of for ( int k = 0; k < szjConditions. length; k++ )
									
			
			if ( szsqltMDJoin [ i ]. getAlias ( ) == null )
				continue;
				
			// 1.
			if ( FROMTableNameExists ( szsqltMDJoin [ i ]. getAlias ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"A table with name " +  szsqltMDJoin [ i ]. getDBName ( ) + " in JOIN clause " +
									"uses alias " + szsqltMDJoin [ i ]. getAlias ( ) +
									" which is a name of the table from FROM clause. Choose different alias." );

			// 1.
			if ( JOINTableNameExists ( szsqltMDJoin [ i ]. getAlias ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"A table with name " +  szsqltMDJoin [ i ]. getDBName ( ) + " in JOIN clause " +
									"uses alias " + szsqltMDJoin [ i ]. getAlias ( ) +
									" which is a name of the table from JOIN clause. Choose different alias." );

			// 3.
			if ( !tableAliasIsUnique ( szsqltMDJoin [ i ]. getAlias ( ) ) )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOIN: " +
									"Table alias " + szsqltMDJoin [ i ]. getAlias ( ) +
									" is used for two different tables or is declared twice for the same table." );
		} // end of for ( int i = 0; i < szsqltMDJoin. length; i++ )
		
		// 5.
		validateJOINTablesIntegrality ( );
	} // end of private void validateJOIN ( DatabaseMetaDataImpl dbMD ) throws SQLException
	
	/**
	*	Validates the fact that all tables are joined either to the main FROM table or to a 
	*	table joined to the main FROM table (either directly or indirectly via other tables).
	*	Does nothing if tables are joined properly. Throws an exception otherwise.
	*
	*	A proper way to join a table:
	*		Main_FROM_Table <-- ( 0..n <-- JOINed_Table ) <-- JOINed_Table
	*/
	private void validateJOINTablesIntegrality ( ) throws SQLException
	{
		/*
			Here is how this method works.
			1. Add a name of the main FROM table to the vector and all JOIN tables to a container.
			2. Find all tables in the container that are JOINed to the tables in the vector.
			3. If no tables are found and container is not empty than tables are not joined properly. THROW AN EXCEPTION
			4. If some tables were found, remove them from original container, add them to the vector.
			5. If container is empty, return else GOTO 2.
		*/
		
		Vector		vctrJOINTables = new Vector ( );
		vctrJOINTables. addAll ( m_vctrJOINTables );
		
		Vector		vctrValidTables = new Vector ( );
		String		strFROMTableName = ( ( SQLTableMetaData ) m_vctrFROMTables. elementAt ( 0 ) ).
																	getDBName ( ). toUpperCase ( );
		vctrValidTables. add ( strFROMTableName );
																	
		SQLTableMetaData			sqltMD = null;
		SQLTableMetaData			sqltMDJoinedTo = null;
		boolean					bAdded = false;
		while ( vctrJOINTables. size ( ) > 0 )
		{
			bAdded = false;
			for ( int i = 0; i < vctrJOINTables. size ( ); i++ )
			{
				sqltMD = ( SQLTableMetaData ) vctrJOINTables. elementAt ( i );
// System. out. println ( "Checking " + sqltMD. getDBName ( ) + " table" );
				sqltMDJoinedTo = getTable ( sqltMD. getTableAliasJoinedTo ( strFROMTableName ) );
// System. out. println ( "\t" + sqltMD. getDBName ( ) + " table is joined to " + sqltMDJoinedTo. getDBName ( ) );
				if ( !vctrValidTables. contains ( sqltMDJoinedTo. getDBName ( ). toUpperCase ( ) ) )
				{
// System. out. println ( "\t" + sqltMDJoinedTo. getDBName ( ) + " table is not a valid joined table" );
					continue;
				}
// System. out. println ( "\t" + sqltMDJoinedTo. getDBName ( ) + " table is a valid joined table" );
					
				bAdded = true;
				vctrValidTables. addElement ( sqltMD. getDBName ( ). toUpperCase ( ) );
				vctrJOINTables. removeElement ( sqltMD );
				
				break;
			}
			
			if ( !bAdded )
				throw new SQLException (
									"com.fourd.jdbc.SQLCommand::validateJOINTablesIntegrality: " +
									"Tables are not properly joined." );

		}
	} // end of private void validateJOINTablesIntegrality ( ) throws SQLException
	
	/**
	*	Determines presence of a JOIN table with a given name.
	*/
	private boolean JOINTableNameExists ( String strTableName )
	{
		if ( strTableName == null )
			return false;
		
		SQLTableMetaData		sqltMD = null;	
		for ( int i = 0; i < m_vctrJOINTables. size ( ); i++ )
		{
			sqltMD = ( SQLTableMetaData ) m_vctrJOINTables. elementAt ( i );
			if ( sqltMD. getDBName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	/**
	*	Determines presence of a JOIN table with a given alias.
	*/
	private boolean JOINTableAliasExists ( String strTableAlias )
	{
		if ( strTableAlias == null )
			return false;
		
		SQLTableMetaData		sqltMD = null;
		for ( int i = 0; i < m_vctrJOINTables. size ( ); i++ )
		{
			sqltMD = ( SQLTableMetaData ) m_vctrJOINTables. elementAt ( i );
			if (	sqltMD. getAlias ( ) != null &&
				sqltMD. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	/**
	*	Returns true if strTableAlias is used no more than once in the query,
	*	return false otherwise.
	*/
	private boolean tableAliasIsUnique ( String strTableAlias )
	{
		if ( strTableAlias == null )
			return false; // null is incorrect alias. TODO: should I throw an exception
			
		int							nCounter = 0;
		SQLTableMetaData [ ]			szsqltMD = getFROMTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
		{
			if ( szsqltMD [ i ]. getAlias ( ) == null )
				continue;
				
			if ( szsqltMD [ i ]. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				nCounter++;
		}

		szsqltMD = getJOINTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
		{
			if ( szsqltMD [ i ]. getAlias ( ) == null )
				continue;
				
			if ( szsqltMD [ i ]. getAlias ( ). toUpperCase ( ). equals ( strTableAlias. toUpperCase ( ) ) )
				nCounter++;
		}
		
		return nCounter <= 1;
	} // end of private boolean tableAliasIsUnique ( String strTableAlias )

	/**
	*	Returns true if strTableName is joined no more than once,
	*	return false otherwise.
	*/
	private boolean JOINTableNameIsUnique ( String strTableName )
	{
		if ( strTableName == null )
			return false; // null is incorrect alias. TODO: should I throw an exception
			
		int							nCounter = 0;
		SQLTableMetaData [ ]			szsqltMD = getJOINTables ( );
		for ( int i = 0; i < szsqltMD. length; i++ )
		{
			if ( szsqltMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) )
				nCounter++;
		}

		return nCounter <= 1;
	} // end of private boolean JOINTableNameIsUnique ( String strTableName )
	
	/**
	*	Returns meta data object for a table with an alias or name of strAlias
	*	Throws an exception if such a table does not exist in the query.
	*/
	SQLTableMetaData getTable ( String strAlias ) throws SQLException
	{
		SQLTableMetaData		sqltMDResult = getFROMTable ( strAlias );
		if ( sqltMDResult == null )
			sqltMDResult = getJOINTable ( strAlias );
		if ( sqltMDResult == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::validateJOIN: " +
								"Table/alias " + strAlias +
								" does not exist in the query" );
								
		return sqltMDResult;
	}
	
	void setStoredProcedureName ( String strStoredProcedureName )
	{
		m_strStoredProcedureName = strStoredProcedureName;
	}
	
	String getStoredProcedureName ( )
	{
		return m_strStoredProcedureName;
	}
	
	/**
	*	When sql parser parses query or stored procedure command,
	*	it adds null for each '?'. Later I need to make sure that 
	*	user registers valid values for each null in the vector
	*	before actually trying to execute stored procedure call.
	*
	*	TODO: Object objParameter, should probably be replaces
	*	by opData opParamter.
	*/
	void addParameter ( Object objParameter )
	{
		// Note: I accept nulls
		m_vctrParameters. addElement ( objParameter );
	}
	
	/**
	*	int nIndex is 1-based.
	*
	*	This method just stores a value for a parameter with a given index.
	*	It does not store the actual Parameter objects. Parameter objects
	*	are stored elsewhere : in BooleanExpression objects as part of WHERE,
	*	in values as part of SET (UPDATE command) or VALUES (INSERT command), etc...
	*/
	void setParameter ( int nIndex, opData opdParameter ) throws SQLException
	{
		if ( nIndex < 1 || nIndex > m_vctrParameters. size ( ) )
			throw new SQLException (
								"com.fourd.jdbc.SQLCommand::setParameter: "+
								"Invalid parameter index - " + nIndex );
								
		m_vctrParameters. setElementAt ( opdParameter, nIndex - 1 );
		
		// Lets also set parameter values
		Value [ ]		szValues = getValues ( );
		for ( int i = 0; i < szValues. length; i++ )
			if ( szValues [ i ]. isParameter ( ) && szValues [ i ]. getParameter ( ). getIndex ( ) == nIndex - 1 )
			{
				szValues [ i ]. setopValue ( opdParameter );
				
				break;
			}
	}
	
	opData [ ] getParameters ( )
	{
		return ( opData [ ] ) m_vctrParameters. toArray ( new opData [ m_vctrParameters. size ( ) ] );
	}
}