/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.util.*;

/**
*	SQLTokenizer class is used to divide an sql string into tokens.
*/
class SQLTokenizer
{
	private StringBuffer			m_strbSQL = null;
	
	/*	Characters of the m_strbSQL are traversed from left to right during
		tokenization process. m_nCurrentPosition contains index of the character
		that was processed last. In the very beginning, m_nCurrentPosition point
		right before the very first character.
		
		Position pointer is the easiest way for bidirectional lookups. It also saves time and
		memory. It is not as pure object-oriented as String or InputStream but it should be
		enough in current situation.
	*/
	private int					m_nCurrentPosition = -1;
	
	private static final char [ ]		SZ_CH_WHITE_SPACES = new char [ ] { ' ', '\t', '\n', '\r' };
	private static final char [ ]		SZ_CH_ONE_CHAR_TOKENS = new char [ ] { '*', '.', ',', '?', '=', '(', ')', '{', '}' };
	
	private static final char [ ] [ ]	SZ_CH_TOKEN_ENCLOSERS = new char [ ] [ ] {
																new char [ ] { '"', '\'', '[', '\'' },
																new char [ ] { '"', '\'', ']', '\'' } };
																
	/*	For each token extracted using getNextToken method the start of the token
		(index of its first character) is saved in m_vctrTokenLocations. This will allow
		to roll back.
	*/
	private Vector					m_vctrTokenLocations = new Vector ( );
	
	/**
	*	SQLTokenizer does not modify strbSQL input parameter.
	*/
	SQLTokenizer ( StringBuffer strbSQL ) throws IllegalArgumentException
	{
		if ( strbSQL == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLTokenizer.SQLTokenizer::SQL string must not be null" );
			
		m_strbSQL = strbSQL;
		
		// Skip leading white spaces and advance to the first token
		skipChars ( SZ_CH_WHITE_SPACES );
	}

	/**
	*	Returns true if there are more potential tokens.
	*/	
	boolean hasMoreTokens ( ) throws Exception
	{
		if ( m_strbSQL == null )
			throw new Exception ( "com.fourd.jdbc.SQLTokenizer.hasMoreTokens::SQL string must not be null" );
			
		return m_nCurrentPosition < m_strbSQL. length ( ) - 1; // also takes care of the 0-length sql string
	}
	
	/**
	*	Returns next token in the sql string or throws and exception if there was an error.
	*	An error may happen if either the sql string is null or there are no more tokens or
	*	incorrect token was found. An opening quote without a closing quote would be an 
	*	example of incorrect token.
	*/
	String getNextToken ( ) throws Exception
	{
		/*
			m_nCurrentPosition will point either to the very last character in the sql string
			or to the very first whitespace after the most recent token.
		*/
		if ( m_strbSQL == null )
			throw new Exception ( "com.fourd.jdbc.SQLTokenizer.getNextToken::SQL string must not be null" );
			
		if ( !hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLTokenizer.getNextToken::There are no more tokens" );
			
		/*
			SQLTokenizer recognizes the following tokens:
		DONE	1. Words - start with character and contain characters, digits, and underscores '_' only.
		DONE	2. Question mark - '?'. Missing parameter in callable statements.
		DONE	3. String in double quotes.
		DONE	4. String in singles quotes.
		DONE	5. String in square brakets.
		DONE	6. Arithmetic comparison signs - '<', '>', '=', '!=', '<=', '>='. 
		DONE	7. Asteric '*', comma ',', period '.'
		*/
		
		// Here we can safely assume that there are still some characters left
		// because a call to hasMoreTokens above returned true.
		
		// Lets save the index of the first character in the next tooken
		m_vctrTokenLocations. addElement ( new Integer ( m_nCurrentPosition + 1 ) );
		
		String		strNextToken = null;
		char			chNext = m_strbSQL. charAt ( m_nCurrentPosition + 1 );
		if ( charIsInArray (
					m_strbSQL. charAt ( m_nCurrentPosition + 1 ),
					SZ_CH_TOKEN_ENCLOSERS [ 0 ] ) ) // Beginning of an enclosed string
		{
			// Advance to the opening enclosing character
			m_nCurrentPosition++;
			
			/*	Lets remember the location of the opening character
				just in case if we need to display exactly which string does
				not have a closing character
			*/
			int		nTokenBeginning = m_nCurrentPosition;

			char		chClosingChar = getClosingChar ( m_strbSQL. charAt ( m_nCurrentPosition ) );
			// Read until closing character (inclusive) or end of the sql string
			strNextToken = moveToChar ( chClosingChar );
			if ( strNextToken == null )
			{
				throw new Exception (
								"com.fourd.jdbc.SQLTokenizer.getNextToken::A closing " +
								chClosingChar +
								" is missing : " +
								m_strbSQL. substring ( nTokenBeginning ) );
			}
				
			// Save the opening double-quote
			strNextToken = String. valueOf ( chNext ) + strNextToken;
				
			// Enclosed string was successfully read.
		} // end of if ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '"' )
		else if ( Character. isLetter ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) ) // Beginning of a word
		{
			// Advance to the beginning of the word
			m_nCurrentPosition++;
			
			// Advance to the end of the word.
			strNextToken = moveToEndOfWord ( );
		} // end of String getNextToken ( ) throws Exception
		else if (	Character. isDigit ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) || 
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '-' ||
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '+' )
		{
			strNextToken = "";
			if ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '-' ||
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '+' )
			{
				if ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '-' )
					// I do not need '+' at the beginning of a number
					strNextToken = String. valueOf ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) );
				
				// Advance to the +/-
				m_nCurrentPosition++;
			}
			
			if (
				m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1  && // There is still at least one character to read
				Character. isDigit ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) )
			{
				// Advance to the digit
				m_nCurrentPosition++;
				
				strNextToken += moveToEndOfDigits ( );
				if (
					m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1  && // There is still at least one character to read
					m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '.' ) // Fractional part of a number should follow
				{
					// Advance to the .
					m_nCurrentPosition++;
					if (
						m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1  && // There is still at least one character to read
						Character. isDigit ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) )
					{
						// Advance to the digit
						m_nCurrentPosition++;
				
						strNextToken += "." + moveToEndOfDigits ( ); // read fractional part
					}
				}
			}
		}
		else if ( charIsInArray ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ), SZ_CH_ONE_CHAR_TOKENS ) )
		{
			// Advance to a one-char token
			m_nCurrentPosition++;
			
			strNextToken = String. valueOf ( m_strbSQL. charAt ( m_nCurrentPosition ) );
		}
		else if ( chNext == '<' || chNext == '>' ) // one or two character arithmetic comparison token
		{
			// Advance one character
			m_nCurrentPosition++;
			
			if (
				m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 && // There is still at least one character to read
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '=' ) // <= or >= token
			{
				// Advance to '=' character
				m_nCurrentPosition++;
				
				strNextToken = String. valueOf ( chNext ) + '=';
			}
			else // < or > token or no more characters to read
				strNextToken = String. valueOf ( chNext );
		}
		else if ( chNext == '!' ) // != token
		{
			// Advance one character to '!'
			m_nCurrentPosition++;

			if (
				m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 && // There is still at least one character to read
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) == '=' ) // != token
			{
				// Advance to '=' character
				m_nCurrentPosition++;
				
				strNextToken = "!=";
			}
			else
				throw new Exception (
								"Unrecognizable start of token : " +
								m_strbSQL. substring ( m_nCurrentPosition ) );
		}
		else
			throw new Exception (
							"Unrecognizable start of token : " +
							m_strbSQL. substring ( m_nCurrentPosition + 1 ) );
		
// DEBUG
// System. out. println ( "Next token is : " + strNextToken );
// DEBUG

		// Lets see if there are any white spaces to be skipped.
		// But first lets make sure that there is anything left read.
		if (
			m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 &&
			charIsInArray ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ), SZ_CH_WHITE_SPACES ) )
		{
			// Advance m_nCurrentPosition to the first white space after the token
			m_nCurrentPosition++;
			
			skipChars ( SZ_CH_WHITE_SPACES );
		}
		
		return strNextToken;
	}
	
	/**
	*	skipChars advances m_nCurrentPosition from its current location to
	*	the character which is right before the first character not in the szchToSkip array.
	*	skipChars does nothing if m_nCurrentPosition initially points to a character 
	*	not in the szchToSkip array.
	*/
	private void skipChars ( char [ ] szchToSkip )
	{
		if ( m_strbSQL. length ( ) == 0 )
			// There is nothing to skip
			return;
			
		if ( m_nCurrentPosition == -1 ) // special case: m_nCurrentPosition points before the first character
		{
			if ( !charIsInArray ( m_strbSQL. charAt ( 0 ), szchToSkip ) )
				// First character should not be skipped
				return;
			else
			{
				// First character should be skipped
				m_nCurrentPosition++;
// DEBUG
// System. out. println ( "Character skipped : '" + m_strbSQL. charAt ( 0 ) + "'" );
// DEBUG
			}
		}
		else if ( !charIsInArray ( m_strbSQL. charAt ( m_nCurrentPosition ), szchToSkip ) )
			// m_nCurrentPosition does not initially point to a char that should be skipped
			return;
			
		while (
				m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 &&
				charIsInArray ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ), szchToSkip ) )
		{
			m_nCurrentPosition++;
// DEBUG
// System. out. println ( "Character skipped : '" + m_strbSQL. charAt ( m_nCurrentPosition ) + "'" );
// DEBUG
		} // end of private void skipChars ( char [ ] szchToSkip )

		return;
	}

	/**
	*	Returns true if character ch is one of the elements of the array szchArray.
	*	Returns false otherwise.
	*/	
	private boolean charIsInArray ( char ch, char [ ] szchArray )
	{
		for ( int i = 0; i < szchArray. length; i++ )
			if ( ch == szchArray [ i ] )
				return true;
				
		return false;
	}
	
	/**
	*	Moves m_nCurrentPosition to point to the first occurence of chTarget character
	*	to the right of the current m_nCurrentPosition.
	*	Returns a string starting from initial m_nCurrentPosition (inclusive) to
	*	the next occurence of target character (also inclusive) if the target character was found.
	*	Returns null otherwise. If target character was not found, then m_nCurrentPosition will
	*	point to the last character in the sql string.
	*/
	private String moveToChar ( char chTarget )
	{
		StringBuffer	strbResult = new StringBuffer ( "" );
		while ( m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 )
		{
			m_nCurrentPosition++;
			strbResult. append ( m_strbSQL. charAt ( m_nCurrentPosition ) );
			if ( m_strbSQL. charAt ( m_nCurrentPosition ) == chTarget )
				if ( m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 )
				{
					// Lets see if chTarget is just being escaped
					if ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) != chTarget )
						// No it is not being escaped
						return strbResult. toString ( );
						
					// OK, chTarget is in fact being escaped
					m_nCurrentPosition++;
				}
				else
					return strbResult. toString ( );
		}
		
		return null;
	}
	
	/**
	*	Moves to the end of the word so that m_nCurrentPosition points to the last character
	*	of the word. Returns a word. Returns null if initial location of m_nCurrentPosition
	*	does not point at the valid beginning of a word ( alpha character or digit ).
	*/
	private String moveToEndOfWord ( )
	{
		if ( !Character. isLetterOrDigit ( m_strbSQL. charAt ( m_nCurrentPosition ) ) )
			// First character of a word should be an alpha character or a digit.
			return null;
			
		StringBuffer		strbResult = new StringBuffer ( "" );
		strbResult. append ( m_strbSQL. charAt ( m_nCurrentPosition ) );
			
		while ( m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 )
		{
			if (
				!Character. isLetterOrDigit ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) &&
				m_strbSQL. charAt ( m_nCurrentPosition + 1 ) != '_' )
				return strbResult. toString ( );
				
			strbResult. append ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) );
			m_nCurrentPosition++;
		}
		
		return strbResult. toString ( );
	} // end of private String moveToEndOfWord ( )	
	
	/**
	*	Moves to the end of the sequence of digits so that m_nCurrentPosition points to the last digit
	*	of the sequence. Returns all digits that were read as a String.
	*	Returns null if initial location of m_nCurrentPosition does not point at a digit.
	*/
	private String moveToEndOfDigits ( )
	{
		if ( !Character. isDigit ( m_strbSQL. charAt ( m_nCurrentPosition ) ) )
			// First character should be a digit.
			return null;
			
		StringBuffer		strbResult = new StringBuffer ( "" );
		strbResult. append ( m_strbSQL. charAt ( m_nCurrentPosition ) );
			
		while ( m_nCurrentPosition + 1 <= m_strbSQL. length ( ) - 1 )
		{
			if ( !Character. isDigit ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) ) )
				return strbResult. toString ( );
				
			strbResult. append ( m_strbSQL. charAt ( m_nCurrentPosition + 1 ) );
			m_nCurrentPosition++;
		}
		
		return strbResult. toString ( );
	} // end of private String moveToEndOfWord ( )	

	/**
	*	Returns closing character for a given opening one.
	*	Strings or tokens may be enclosed in pairs of characters. For examle - strings in square brackets.
	*	In this case for input '[' the ']' will be returned.
	*/
	private char getClosingChar ( char chOpening )
	{
		for ( int i = 0; i < SZ_CH_TOKEN_ENCLOSERS [ 0 ]. length; i++ )
			if ( SZ_CH_TOKEN_ENCLOSERS [ 0 ] [ i ] == chOpening )
				return SZ_CH_TOKEN_ENCLOSERS [ 1 ] [ i ];
				
		return '0';
	}
	
	/**
	*	Rolls back nTokens times. Does nothing if nTokens is less then or equal to zero.
	*	Rolls back to the beginning of sql string if nTokens is greater then the amount of
	*	currently retrieved tokens (using getNextToken method).
	*/
	void rollBack ( int nTokens )
	{
		if ( nTokens <= 0 )
			return;
			
		// I can roll back only to the beginning of the sql string
		nTokens = Math.min ( nTokens, m_vctrTokenLocations. size ( ) );
		
		m_nCurrentPosition = ( ( Integer ) m_vctrTokenLocations. elementAt (
											m_vctrTokenLocations. size ( ) - nTokens ) ).
													intValue ( ) - 1;
	}
	
	/**
		Returns next token without actually advancing past it.
		Convinience method. It simply reads the next token and then rolls back once.
		Throws exactly the same exceptions as getNextToken method.
	*/
	String peekAtNextToken ( ) throws Exception
	{
		String	strResult = getNextToken ( );
		rollBack ( 1 );
		
		return strResult;
	}
	
	/**
	*	Throws an exception if there are no more tokens left.
	*	Does nothing if there is at least one token left.
	*	Does not advance to the next token, just peeks at it.
	*	
	*	Helper method.
	*/
	void validateNextToken ( ) throws Exception
	{
		if ( !hasMoreTokens ( ) )
			throw new Exception (
							"com.fourd.jdbc.SQLTokenizer::validateNextToken: " +
							"Query is incomplete, expected a token but found end of query." );
	}

	/**
	*	Makes sure that there is at least one token left and that the next token is strExpectedToken.
	*	Does nothing if the above condition is true, throws exception otherwise.
	*	Does not advance to the next token, just peeks at it.
	*	
	*	Helper method, really.
	*/
	void validateNextToken ( String strExpectedToken ) throws Exception
	{
		if ( strExpectedToken == null )
			throw new Exception (
							"com.fourd.jdbc.SQLTokenizer::validateNextToken: " +
							"Expected token can not be null." );
							
		if ( !hasMoreTokens ( ) )
			throw new Exception (
							"com.fourd.jdbc.SQLTokenizer::validateNextToken: " +
							"Query is incomplete, expected " + strExpectedToken +
							" but found end of query." );
							
		if ( !peekAtNextToken ( ). toUpperCase ( ). equals ( strExpectedToken. toUpperCase ( ) ) )
			throw new Exception (
							"com.fourd.jdbc.SQLTokenizer::validateNextToken: " +
							"Unexpected token - " + peekAtNextToken ( ) +
							" expected " + strExpectedToken + " instead" );
	}
}