/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

class StatementImpl implements Statement
{
	// A reference to the connection that generated instance of this class
	private ConnectionImpl				m_connectionImpl = null;
	// A connection to 4D DB via 4D Open For Java
	private opConnection				m_opConnection = null;
	
	// An object that will handle all communication with the DB.
	protected opProcess					m_opProcess = null;
	
	/**
		Holds a reference to the current result set produced by this statement.
	*/
	private ResultSetImpl				m_resultSetImpl = null;
	
	/*
		Represents records currently selected by SELECT command.
		I may need to use an array of selections instead of just one
		to be able to deal with joins.
	*/
	private opSelection					m_opSELECTSelection = null;
	
	protected DatabaseMetaDataImpl		m_dbMD = null;
	
	/**
		Should always synchronize on m_objDataAccessLock before accessing m_opProcess.
		Can't synchronize on m_opProcess itself.
	*/
	private Object						m_objDataAccessLock = null;
	
	/**
		"Statement Is Closed" flag. Once m_bClosed is set to true,
		all methods will throw "Statement is closed" exception.
	*/
	protected boolean						m_bClosed = false;
	
	// Helpers
	/*
		I need to use an array of "fields" to be able to call opProcess. OrderBy ( ... )
		Instead of creating this array over and over again each time I need to restore selection, 
		I'll just create it once as a class member and use it as I need.
	*/
	private opFieldArray				m_opfaOrderBy = null;
	private opSearchArray				m_opsaWhere = null;
	
	private int							m_nSelectionID = 0;
	
	/**
	*	Stores results of SQL queries. Resutls are either ResulSets or Integers.
	*/
	protected Vector					m_vctrResults = new Vector ( );
	
	private int							m_nFetchSize = 1000;

	/**
	*	Dummy constructor for PreparedStatementImpl
	*/
	StatementImpl ( )
	{
		;
	}
	
	StatementImpl (
				ConnectionImpl connImpl,
				opConnection opC,
				opProcess opP,
				DatabaseMetaDataImpl dbMD,
				Object objDataAccessLock )
	{
		// Instances of this class will only be created by connectionImpl and so there is no need to check for valid parameters
		
		m_connectionImpl = connImpl;
		m_opConnection = opC;
		m_opProcess = opP;
		m_dbMD = dbMD;
		m_objDataAccessLock = objDataAccessLock;
	}
	
	/**
	*	Adds an SQL command to the current batch of commmands for this Statement object. 
	*/
	public void addBatch ( String strSQL ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "addBatch" );
	}

	/**
	*	Cancels this Statement object if both the DBMS and driver support aborting an SQL statement. 
	*/
	public void cancel ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "cancel" );
	}

	/**
	*	Makes the set of commands in the current batch empty. 
	*/
	public void clearBatch ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "clearBatch" );
	}

	/**
	*	Clears all the warnings reported on this Statement object. 
	*/
	public void clearWarnings ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "clearWarnings" );
	}

	/**
	*	Releases this Statement object's database and JDBC resources immediately instead of waiting for this to happen when it is automatically closed. 
	*/
	public void close ( ) throws SQLException
	{
		if ( m_bClosed )
			return;
			// throw new SQLException ( "com.fourd.jdbc.StatementImpl:: Statement is closed" );
			
		if ( m_resultSetImpl != null )
			m_resultSetImpl. close ( );
		
		m_bClosed = true;
			
		// throwUnimplementedException ( "" );
	}

	/**
	*	Executes an SQL statement that may return multiple results. 
	*/
	public boolean execute ( String strSQL ) throws SQLException
	{
		validateOpened ( );
		
		SQLCommand		sqlCommand = prepareForExecute ( strSQL );
		
		if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_SELECT )
		{
			m_resultSetImpl = executeSELECTCommand ( sqlCommand );
			m_vctrResults. addElement ( m_resultSetImpl );
			
			return true;
		}
		else if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_DELETE )
			m_vctrResults. addElement ( new Integer ( executeDELETECommand ( sqlCommand ) ) );
		else if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_INSERT )
			m_vctrResults. addElement ( new Integer ( executeINSERTCommand ( sqlCommand ) ) );
		else if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_UPDATE )
			m_vctrResults. addElement ( new Integer ( executeUPDATECommand ( sqlCommand ) ) );
		else
			throw new SQLException (
								"com.fourd.jdbc.StatementImpl::execute: " +
								"Unsupported query" );

		return false;
	}


	/**
	*	Submits a batch of commands to the database for execution and if all commands execute successfully, returns an array of update counts. 
	*/
	public int [ ] executeBatch ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "executeBatch" );
		
		return null;
	}

	/**
	*	Executes an SQL statement that returns a single ResultSet object. 
	*/
	public ResultSet executeQuery ( String strSQL ) throws SQLException
	{
		validateOpened ( );
			
		SQLCommand		sqlCommand = prepareForExecute ( strSQL );
		
		if ( sqlCommand. getType ( ) != SQLCommand. CMD_TYPE_SELECT )
			// execute Query should only be able to perform SELECT queries in the current implementation
			throw new SQLException ( "com.fourd.jdbc.StatementImpl::executeQuery: Unsupported query" );
		
		m_resultSetImpl = executeSELECTCommand ( sqlCommand );
		return m_resultSetImpl;
	}

	/**
	*	Executes an SQL INSERT, UPDATE or DELETE statement. 
	*/
	public int executeUpdate ( String strSQL ) throws SQLException
	{
		validateOpened ( );
			
		SQLCommand		sqlCommand = prepareForExecute ( strSQL );
		
		/*
			executeUpdate should only be able to perform INSERT, UPDATE or DELETE commands.
			Currently, it supports only DELETE and INSERT.
		*/
		if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_DELETE )
			return executeDELETECommand ( sqlCommand );
		else if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_INSERT )
			return executeINSERTCommand ( sqlCommand );
		else if ( sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_UPDATE )
			return executeUPDATECommand ( sqlCommand );
		else
			throw new SQLException (
								"com.fourd.jdbc.StatementImpl::executeUpdate: " +
								"Unsupported query" );
	}

	/**
	*	Returns the Connection object that produced this Statement object. 
	*/
	public Connection getConnection ( ) throws SQLException
	{
		validateOpened ( );
			
		// throwUnimplementedException ( "" );
		
		return m_connectionImpl;
	}

	/**
	*	Retrieves the direction for fetching rows from database tables that is the default for result sets generated from this Statement object.
	*/
	public int getFetchDirection ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getFetchDirection" );
		
		return 0;
	}

	/**
	*	Retrieves the number of result set rows that is the default fetch size for result sets generated from this Statement object. 
	*/
	public int getFetchSize ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getFetchSize" );
		
		return 0;
	}

	/**
	*	Returns the maximum number of bytes allowed for any column value. 
	*/
	public int getMaxFieldSize ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getMaxFieldSize" );
		
		return 0;
	}

	/**
	*	Retrieves the maximum number of rows that a ResultSet object can contain. 
	*/
	public int getMaxRows ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getMaxRows" );
		
		return 0;
	}

	/**
	*	Moves to a Statement object's next result. 
	*/
	public boolean getMoreResults ( ) throws SQLException
	{
		validateOpened ( );
			
		//TODO: implement properly
		
		// throwUnimplementedException ( "getMoreResults" );
		
		return false;
	}

	/**
	*	Retrieves the number of seconds the driver will wait for a Statement object to execute. 
	*/
	public int getQueryTimeout ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getQueryTimeout" );
		
		return 0;
	}

	/**
	*	Returns the current result as a ResultSet object. 
	*/
	public ResultSet getResultSet ( ) throws SQLException
	{
		validateOpened ( );
			
		// throwUnimplementedException ( "" );
		
		return m_resultSetImpl;
	}

	/**
	*	Retrieves the result set concurrency for ResultSet objects generated by this Statement object. 
	*/
	public int getResultSetConcurrency ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getResultSetConcurrency" );
		
		return 0;
	}

	/**
	*	Retrieves the result set type for ResultSet objects generated by this Statement object. 
	*/
	public int getResultSetType ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "getResultSetType" );
		
		return 0;
	}

	/**
	*	Returns the current result as an update count; if the result is a ResultSet object or there are no more results, -1 is returned. 
	*/
	public int getUpdateCount ( ) throws SQLException
	{
		validateOpened ( );
		
		if ( m_vctrResults. size ( ) == 0 )
			return -1;
			
		Object		objResult = m_vctrResults. elementAt ( 0 );
		m_vctrResults. remove ( 0 );
		if ( objResult instanceof Integer )
			return ( ( Integer ) objResult ). intValue ( );
		else if ( objResult instanceof ResultSetImpl )
			return -1;

		return -1;
	}

	/**
	*	Retrieves the first warning reported by calls on this Statement object. 
	*/
	public SQLWarning getWarnings ( ) throws SQLException
	{
		validateOpened ( );
		
		return null;
	}

	/**
	*	Defines the SQL cursor name that will be used by subsequent Statement object execute methods. 
	*/
	public void setCursorName ( String strName ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setCursorName" );
	}

	/**
	*	Sets escape processing on or off. 
	*/
	public void setEscapeProcessing ( boolean bEnable ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setEscapeProcessing" );
	}

	/**
	*	Gives the driver a hint as to the direction in which the rows in a result set will be processed. 
	*/
	public void setFetchDirection ( int nDirection ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setFetchDirection" );
	}

	/**
	*	Gives the JDBC driver a hint as to the number of rows that should be fetched from the database when more rows are needed. 
	*/
	public void setFetchSize ( int nRows ) throws SQLException
	{
		validateOpened ( );

		m_nFetchSize = nRows;
	}

	/**
	*	Sets the limit for the maximum number of bytes in a column to the given number of bytes. 
	*/
	public void setMaxFieldSize ( int nMax ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setMaxFieldSize" );
	}

	/**
	*	Sets the limit for the maximum number of rows that any ResultSet object can contain to the given number. 
	*	TODO: Currently the method does nothing. Should be properly implemented.
	*/
	public void setMaxRows ( int nMax ) throws SQLException
	{
		validateOpened ( );
	}

	/**
	*	Sets the number of seconds the driver will wait for a Statement object to execute to the given number of seconds. 
	*	TODO: Currently does nothing. Should be properly implemented.
	*/
	public void setQueryTimeout ( int nSeconds ) throws SQLException
	{
		validateOpened ( );
	}

// !!!!!!!!!!!!!!!!!!!!!!!!!!--------========== Methods not in the Statement interface ==========--------!!!!!!!!!!!!!!!!!!!!!!!!!!

	private void s_e_p_a_r_a_t_o_r__________________________1 ( )
	{
		;
	}
	
	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException ( "com.fourd.jdbc.StatementImpl: Method " + strMethodName + " is not implemented" );
	}
	
	protected ResultSetImpl executeSELECTCommand ( SQLCommand sqlCommand ) throws SQLException
	{
		validateOpened ( );
			
		// This method is called by executeQuery method and thus there is no need to check for valid input parameters
		
		sqlCommand. validate ( m_dbMD );
		
		sqlCommand. expandAsteriskInSELECT ( m_dbMD );
		/*
			Here is what I currently support in SELECT command:
				SELECT from one table only
		*/
		if ( sqlCommand. getFROMTableCount ( ) > 1 )
			throw new SQLException ( "com.fourd.jdbc.StatementImpl::executeSELECTCommand: SELECT from more than one table is not supported" );
			
		SQLTableMetaData		sqltMDFrom = sqlCommand. getFROMTables ( ) [ 0 ];
		String				strFROMTable = sqltMDFrom. getDBName ( );
		
		// TODO: move all verification to SQLCommand.validate ( ... )

		// It is time to make sure that all tables and columns mentioned in SELECT command really do exist in the Database.
		if ( !m_dbMD. tableExists ( strFROMTable ) )
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::executeSELECTCommand: " +
							"Table " + strFROMTable + " does not exist in the database" );
		
		// TODO: make sure that all ORDER BY columns exist and are from single FROM table
		SQLColumnMetaData [ ]			szsqlcMD = sqlCommand. getORDERBYColumns ( );
		for ( int i = 0; i < szsqlcMD. length; i++ )
		{
			// ORDER BY column is from FROM table
			if (
				szsqlcMD [ i ]. getDBTableName ( ) != null &&
				!szsqlcMD [ i ]. getDBTableName ( ). toUpperCase ( ). equals ( strFROMTable. toUpperCase ( ) ) &&
				!szsqlcMD [ i ]. getDBTableName ( ). toUpperCase ( ). equals ( sqltMDFrom. getAlias ( ). toUpperCase ( ) ) )
					throw new SQLException (
										"com.fourd.jdbc.StatementImpl::executeSELECTCommand: " +
										"table " + szsqlcMD [ i ]. getDBTableName ( ) +
										" is not specified in the FROM clause" );
										
			// ORDER BY column actually exists in the table in database
			if ( !m_dbMD. columnExists ( strFROMTable, szsqlcMD [ i ]. getDBName ( ) ) )
				throw new SQLException (
								"com.fourd.jdbc.StatementImpl::executeSELECTCommand: " +
								"Column " + szsqlcMD [ i ]. getDBName ( ) +
								" does not exist in the table " + strFROMTable );
		}

		// Now we are finally ready to talk to 4D DB
		
		/*
			sqltDBMDFrom represents meta data about FROM table in the database and thus it is
			different from sqltMDFrom which represents meta data about FROM table in SELECT query
		*/
		SQLTableMetaData		sqltDBMDFrom = m_dbMD. getTable ( strFROMTable );
		
		// Lets find table's number. "Table number" identifies table in a 4D DB.
		short		nFROMTableNumber = sqltDBMDFrom. getNumber ( );
		
		ResultSetImpl		rsResult = null;
		try
		{
			// Select all records in the table
        
			// Lets prepare ORDER BY for execution, if required
			szsqlcMD = sqlCommand. getORDERBYColumns ( );
			if ( szsqlcMD. length > 0 )
			{
				m_opfaOrderBy = new opFieldArray ( szsqlcMD. length );
				m_opfaOrderBy. mTargetTable = nFROMTableNumber;
				for ( int i = 0; i < szsqlcMD. length; i++ )
				{
					m_opfaOrderBy. mFieldArray [ i ] = new opField (
																nFROMTableNumber,
																sqltDBMDFrom. getColumnNumber (
																				szsqlcMD [ i ]. getDBName ( ) ) );
					if ( szsqlcMD [ i ]. getOrderBy ( ) == SQLColumnMetaData. ORDER_BY_ASC )
						m_opfaOrderBy. mFieldArray [ i ]. mOrdering = opConstants. ORDER_BY_ASC;
					else
						m_opfaOrderBy. mFieldArray [ i ]. mOrdering = opConstants. ORDER_BY_DESC;
				}
			}
			
			prepareWHEREHelper ( nFROMTableNumber, sqlCommand, sqltDBMDFrom );
								
			// Request a table for exclusive use
			m_connectionImpl. registerTables ( new String [ ] { strFROMTable }, this );
                
			// Select all records in the table
			/*
				Before actually doing anything with opProcess, we have to aquire a sync lock on it
				because there may be multiple processes trying to access opProcess object simultaneously.
			*/
			opSelection		opsHelper = null;
			synchronized ( m_objDataAccessLock )
			{
				opsHelper = m_opProcess. AllRecords ( new opTable ( nFROMTableNumber ) );
				if ( m_opsaWhere. mSize > 0 )
					opsHelper = m_opProcess. Search ( m_opsaWhere );
				if ( m_opfaOrderBy != null )
					m_opProcess. OrderBy ( m_opfaOrderBy );
					
				m_opSELECTSelection = new opSelection ( );
				m_opSELECTSelection. mTableNumber = nFROMTableNumber;
				m_opSELECTSelection. mSelectionName = m_connectionImpl. generateSelectionName ( );
				m_opProcess. CopyNamedSelection ( m_opSELECTSelection );
				m_opProcess. UseNamedSelection ( m_opSELECTSelection );
				m_opProcess. RecordsInSelection ( m_opSELECTSelection );
			}

			// Release the table
			m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
		
			rsResult = new ResultSetImpl (
								sqlCommand,
								sqltDBMDFrom,
								m_opSELECTSelection,
								this,
								m_opProcess,
								m_connectionImpl,
								m_objDataAccessLock,
								m_dbMD );
			rsResult. setFetchSize ( m_nFetchSize );
		}
		catch ( opException opE )
		{
			m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
			
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::executeSELECTCommand: Failed to execute SELECT command.\n" +
							opE. GetErrorMsg ( ) );
		}
		
		return rsResult;
	}
	
	/**
	*	Restores selection required for sqlCommand. This method should be used when somebody else
	*	changed our selection. restoreSelection also restores current row.
	*
	*	This method should be used as a starting point for table access and thus it DOES NOT register tables that it is using.
	*	Registration should be done by this method's caller.
	*
	*	TODO: Should this be in the ResultSetImpl class?
	*/
	opSelection restoreSelectionOld ( SQLCommand sqlCommand, int nCurrentRow ) throws SQLException
	{
		validateOpened ( );
			
		/*
			There is no need to validate sqlCommand because at this point it already should be fine
			but still we should check if it is null.
		*/
		if ( sqlCommand == null )
			throw new SQLException ( "com.fourd.jdbc.StatementImpl::restoreSelection: SQLCommand must not be null" );
			
		SQLTableMetaData		sqltMDFrom = sqlCommand. getFROMTables ( ) [ 0 ];
		String				strFROMTable = sqltMDFrom. getDBName ( );
		
		// Now we are ready to talk to 4D DB
		
		/*
			sqltDBMDFrom represents meta data about FROM table in the database and thus it is
			different from sqltMDFrom which represents meta data about FROM table in SELECT query
		*/
		SQLTableMetaData		sqltDBMDFrom = m_dbMD. getTable ( strFROMTable );
		
		// Lets find table's number. "Table number" identifies table in a 4D DB.
		short		nFROMTableNumber = sqltDBMDFrom. getNumber ( );
		
		try
		{
			// Select all records in the table
        
			// Request a table for exclusive use
			// m_connectionImpl. registerTables ( new String [ ] { strFROMTable }, this );

			// Select all records in the table
			/*
				Before actually doing anything with opProcess, we have to aquire a sync lock on it
				because there may be multiple processes trying to access opProcess object simultaneously.
			*/
			synchronized ( m_objDataAccessLock )
			{
				m_opSELECTSelection = m_opProcess. AllRecords ( new opTable ( nFROMTableNumber ) );
				if ( m_opsaWhere. mSearchArray. length > 0 )	
					m_opSELECTSelection = m_opProcess. Search ( m_opsaWhere );
				if ( m_opfaOrderBy != null )
					m_opProcess. OrderBy ( m_opfaOrderBy );
				
				/*
					// No need to restore current row.	
				opS. mCurrentRecord = nCurrentRow - 1; // 4D rows are 1-based BUT opSelection. mCurrentRecord is, in fact, 0-based, DUH
				m_opProcess. GotoSelectedRecord ( opS );
				*/
			}

			// Release the table
			// m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
		}
		catch ( opException opE )
		{
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::restoreSelection: Failed to restore selection.\n" +
							opE. GetErrorMsg ( ) );
		}
		
		return m_opSELECTSelection;
	} // end of opSelection restoreSelectionOld ( SQLCommand sqlCommand, int nCurrentRow ) throws SQLException

	/**
	*	Restores selection required for sqlCommand. This method should be used
	*	when somebody else has changed our selection. DOES NOT restore current row.
	*
	*	This method should be used as a starting point for table access and thus
	*	it DOES NOT register tables that it is using. Registration should be done
	*	by this method's caller.
	*
	*	TODO: Should this be in the ResultSetImpl class?
	*/
	opSelection restoreSelection (
							SQLCommand sqlCommand,
							int nCurrentRow,
							boolean bDoRecordsInSelection ) throws SQLException
	{
		validateOpened ( );
			
		/*
			There is no need to validate sqlCommand because at this point it already should be fine
			but still we should check if it is null.
		*/
		if ( sqlCommand == null )
			throw new SQLException ( "com.fourd.jdbc.StatementImpl::restoreSelection: SQLCommand must not be null" );
			
		try
		{
			/*
				Before actually doing anything with opProcess, we have to aquire a sync lock on it
				because there may be multiple processes trying to access opProcess object simultaneously.
			*/
			synchronized ( m_objDataAccessLock )
			{
				m_opProcess. UseNamedSelection ( m_opSELECTSelection );
				if ( bDoRecordsInSelection )
					m_opProcess. RecordsInSelection ( m_opSELECTSelection );
			}
		}
		catch ( opException opE )
		{
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::restoreSelection: Failed to restore selection.\n" +
							opE. GetErrorMsg ( ) );
		}
		
		return m_opSELECTSelection;
	} // end of opSelection restoreSelection ( SQLCommand sqlCommand, int nCurrentRow ) throws SQLException
	
	protected int executeDELETECommand ( SQLCommand sqlCommand ) throws SQLException
	{
		sqlCommand. validate ( m_dbMD );
		
		SQLTableMetaData		sqltMDFrom = sqlCommand. getFROMTables ( ) [ 0 ];
		String					strFROMTable = sqltMDFrom. getDBName ( );
		
		/*
			sqltDBMDFrom represents meta data about FROM table in the database and thus it is
			different from sqltMDFrom which represents meta data about FROM table in SELECT query
		*/
		SQLTableMetaData		sqltDBMDFrom = m_dbMD. getTable ( strFROMTable );
		// Lets find table's number. "Table number" identifies table in a 4D DB.
		short		nFROMTableNumber = sqltDBMDFrom. getNumber ( );
		
		prepareWHEREHelper ( nFROMTableNumber, sqlCommand, sqltDBMDFrom );
		
		// Request a table for exclusive use
		m_connectionImpl. registerTables ( new String [ ] { strFROMTable }, this );
		opSelection		opsHelper = null;
		int				nResult = 0;
		try
		{
			synchronized ( m_objDataAccessLock )
			{
				opsHelper = m_opProcess. AllRecords ( new opTable ( nFROMTableNumber ) );
				if ( m_opsaWhere. mSize > 0 )
					opsHelper = m_opProcess. Search ( m_opsaWhere );
			
				m_opProcess. RecordsInSelection ( opsHelper );
				nResult = opsHelper. mRecordsInSelection;
// System. out. println ( "About to delete " + nResult + " records" );
				if ( nResult > 0 )
					m_opProcess. DeleteSelection ( opsHelper );
			}
		}
		catch ( opException opE )
		{
			m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
			
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::executeDELETECommand: Failed to execute DELETE command.\n" +
							opE. GetErrorMsg ( ) );
		}

		// Release the table
		m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
		
		return nResult;
	} // end of private ResultSetImpl executeDELETECommand ( SQLCommand sqlCommand ) throws SQLException

	protected int executeINSERTCommand ( SQLCommand sqlCommand ) throws SQLException
	{
		sqlCommand. validate ( m_dbMD );
		
		SQLTableMetaData		sqltMDFrom = sqlCommand. getFROMTables ( ) [ 0 ];
		String					strFROMTable = sqltMDFrom. getDBName ( );
		
		/*
			sqltDBMDFrom represents meta data about FROM table in the database and thus it is
			different from sqltMDFrom which represents meta data about FROM table in SELECT query
		*/
		SQLTableMetaData		sqltDBMDFrom = m_dbMD. getTable ( strFROMTable );
		short					nFROMTableNumber = sqltDBMDFrom. getNumber ( );
		
		// Create opFieldArray and opDataArray from insert columns and values.
		Value [ ]				szValues = sqlCommand. getValues ( );
		opFieldArray			opfArray = new opFieldArray ( szValues.length );
		opfArray. mTargetTable = nFROMTableNumber;
		opDataArray				opdArray = new opDataArray ( szValues.length );
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < szValues. length; i++ )
		{
			sqlcMD = sqltDBMDFrom. getColumnMetaData ( szValues [ i ]. getColumnName ( ) );
			
			opfArray. mFieldArray [ i ]. mFieldNumber = sqlcMD. getNumber ( );
			opfArray. mFieldArray [ i ]. mFieldType = sqlcMD. getopType ( );

			if ( szValues [ i ]. getValue ( ) != null )
				opdArray. mDataArray [ i ] = SQLTools. createopData (
																szValues [ i ]. getValue ( ),
																sqlcMD. getopType ( ) );
			else if ( szValues [ i ]. getopValue ( ) != null )
				opdArray. mDataArray [ i ] = szValues [ i ]. getopValue ( );
			else
			{
				if ( szValues [ i ]. getParameter ( ) != null )
					throw new SQLException (
									"com.fourd.jdbc.StatementImpl::executeINSERTCommand: " +
									"Missing value for parameter #" +
									( szValues [ i ]. getParameter ( ). getIndex ( ) + 1 ) );
									
				throw new SQLException (
								"com.fourd.jdbc.StatementImpl::executeINSERTCommand: " +
								"Missing value" );
			}
		}
		
		// Request a table for exclusive use
		m_connectionImpl. registerTables ( new String [ ] { strFROMTable }, this );
		try
		{
			synchronized ( m_objDataAccessLock )
			{
				m_opProcess. CreateRecord ( opfArray, opdArray );
			}
		}
		catch ( opException opE )
		{
			m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
			
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::executeINSERTCommand: Failed to execute INSERT command.\n" +
							opE. GetErrorMsg ( ) );
		}

		// Release the table
		m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
		
		return 1;
	} // end of private ResultSetImpl executeINSERTCommand ( SQLCommand sqlCommand ) throws SQLException

	protected int executeUPDATECommand ( SQLCommand sqlCommand ) throws SQLException
	{
		sqlCommand. validate ( m_dbMD );
		
		SQLTableMetaData		sqltMDFrom = sqlCommand. getFROMTables ( ) [ 0 ];
		String					strFROMTable = sqltMDFrom. getDBName ( );
		
		/*
			sqltDBMDFrom represents meta data about FROM table in the database and thus it is
			different from sqltMDFrom which represents meta data about FROM table in SELECT query
		*/
		SQLTableMetaData		sqltDBMDFrom = m_dbMD. getTable ( strFROMTable );
		short					nFROMTableNumber = sqltDBMDFrom. getNumber ( );
		
		// Create opFieldArray from update columns.
		Value [ ]				szValues = sqlCommand. getValues ( );
		opFieldArray			opfArray = new opFieldArray ( szValues.length );
		opfArray. mTargetTable = nFROMTableNumber;
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < szValues. length; i++ )
		{
			sqlcMD = sqltDBMDFrom. getColumnMetaData ( szValues [ i ]. getColumnName ( ) );
			
			opfArray. mFieldArray [ i ]. mFieldNumber = sqlcMD. getNumber ( );
			opfArray. mFieldArray [ i ]. mFieldType = sqlcMD. getopType ( );
		}
		
		opDataArray [ ]			szopdArray = new opDataArray [ szValues.length ];
		prepareWHEREHelper ( nFROMTableNumber, sqlCommand, sqltDBMDFrom );

		int		nResult = 0;
		
		// Request a table for exclusive use
		m_connectionImpl. registerTables ( new String [ ] { strFROMTable }, this );
		try
		{
			synchronized ( m_objDataAccessLock )
			{
				opSelection		opsHelper = m_opProcess. AllRecords (
														new opTable ( nFROMTableNumber ) );
				if ( m_opsaWhere. mSize > 0 )
					opsHelper = m_opProcess. Search ( m_opsaWhere );
					
				m_opProcess. RecordsInSelection ( opsHelper );
				nResult = opsHelper. mRecordsInSelection;
				if ( nResult > 0 )
				{
					// Create opDataArray from update values.
					for ( int i = 0; i < szValues. length; i++ )
					{
						sqlcMD = sqltDBMDFrom. getColumnMetaData (
														szValues [ i ]. getColumnName ( ) );
						szopdArray [ i ] = new opDataArray ( nResult );
						/*
							TODO: possible optimization : create new opData only once per each i
							and then reuse it for each j.
						*/
						for ( int j = 0; j < nResult; j++ )
						{
							if ( szValues [ i ]. getValue ( ) != null )
								szopdArray [ i ]. mDataArray [ j ] = SQLTools. createopData (
																				szValues [ i ]. getValue ( ),
																				sqlcMD. getopType ( ) );
							else if ( szValues [ i ]. getopValue ( ) != null )
								szopdArray [ i ]. mDataArray [ j ] = szValues [ i ]. getopValue ( );
							else
							{
								if ( szValues [ i ]. getParameter ( ) != null )
									throw new SQLException (
													"com.fourd.jdbc.StatementImpl::executeUPDATECommand: " +
													"Missing value for parameter #" +
													( szValues [ i ]. getParameter ( ). getIndex ( ) + 1 ) );
									
								throw new SQLException (
												"com.fourd.jdbc.StatementImpl::executeUPDATECommand: " +
												"Missing value" );
							}
						}
					} // end of for ( int i = 0; i < szValues. length; i++ )
					
					m_opProcess. ArrayToSelection ( szopdArray, opfArray );
				} // end of if ( nResult > 0 )
			} // end of synchronized ( m_objDataAccessLock )
		} // end of try
		catch ( opException opE )
		{
			m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
			
			throw new SQLException (
							"com.fourd.jdbc.StatementImpl::executeUPDATECommand: Failed to execute UPDATE command.\n" +
							opE. GetErrorMsg ( ) );
		}

		// Release the table
		m_connectionImpl. unregisterTables ( new String [ ] { strFROMTable }, this );
		
		return nResult;
	} // end of private int executeUPDATECommand ( SQLCommand sqlCommand ) throws SQLException
	
	private SQLCommand prepareForExecute ( String strSQL ) throws SQLException
	{
		// Close previously opened ResultSet (if any)
		if ( m_resultSetImpl != null )
			m_resultSetImpl. close ( );
		
		if ( strSQL == null )
			throw new SQLException (
								"com.fourd.jdbc.StatementImpl::prepareForExecute: " +
								"SQL string can not be null" );
			
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{
			e. printStackTrace ( );
			
			throw new SQLException ( e. getMessage ( ) );
		}
		
		return sqlCommand;
	} // private void prepareForExecute ( String strSQL ) throws SQLException
	
	private void prepareWHEREHelper (
								int nFROMTableNumber,
								SQLCommand sqlCommand,
								SQLTableMetaData sqltDBMDFrom ) throws SQLException
	{
		BooleanExpression [ ]		szbe = sqlCommand. getWHEREExpressions ( );
		m_opsaWhere = new opSearchArray ( szbe. length );
		byte						btLogicalOperator = opConstants. NONE;
		opData [ ]					szopdParameters = sqlCommand. getParameters ( );
		for ( int i = 0; i < szbe. length; i++ )
		{
			m_opsaWhere. mSearchArray [ i ] = new opSearch (
													btLogicalOperator,
													
													nFROMTableNumber,
													
													sqltDBMDFrom. getColumnNumber (
														szbe [ i ]. getColumnMetaData ( ). getDBName ( ) ),
														
													szbe [ i ]. getopComparisonOperator ( ),
													
													szbe [ i ]. getopValue (
														sqltDBMDFrom. getColumnMetaData (
															szbe [ i ]. getColumnMetaData ( ). 
																	getDBName ( ) ). getopType ( ),
														 szopdParameters ) );
			btLogicalOperator = szbe [ i ]. getopLogicalOperator ( );
		}
	}
	
	/**
	*	Just an assert type thingy that makes sure that the statement is open.
	*	Throws an exception in case the statement is closed.
	*/
	protected void validateOpened ( ) throws SQLException
	{
		if ( m_bClosed )
			throw new SQLException ( "com.fourd.jdbc.StatementImpl:: Statement is closed" );
	}
}